#!/usr/bin/env python3
#
# CluDisc: clustered-mds two-step data scaling
# (C) Martin Saturka, http://www.bioplexity.org/
# Version 2.0.16, released under the MIT license.
#

import sys, os, time, stat
import getopt
import math, random
import traceback
import csv
import platform

CluDiscVersion = "2.0.16"

def log_err(message):
  sys.stderr.write(str(message) + "\n")

try:
  import ctypes as C
except:
  log_err("ctypes module (foreign function interface) has to be available.")
  exit(1)

digmod = []
hashash = True
try:
  import hashlib
  digmod = hashlib.sha1
except:
  hashash = False
if not hashash:
  try:
    import sha
    digmod = sha.sha
  except:
    log_err("SHA support (hashlib or sha) has to be available.")
    exit(1)

try:
  import numpy as N
except:
  log_err("NumPy module (Numerics for Python) has to be available.")
  exit(1)

try:
  import tables as H
except:
  log_err("PyTables module (HDF5 for Python) has to be available.")
  exit(1)

hastk = False
try:
  import Tkinter as T
  import tkFileDialog as F
  hastk = True
except:
  hastk = False

if not hastk:
  try:
    import tkinter as T
    import tkinter.filedialog as F
  except:
    log_err("Tkinter module (Tk for Python) has to be available.")
    exit(1)

try:
  import matplotlib.pyplot as plt
except:
  log_err("Matplotlib module (matplotlib) has to be available.")
  exit(1)

try:
  from mpl_toolkits.mplot3d import Axes3D
except:
  log_err("Matplotlib module (mpl_toolkits) has to be available.")
  exit(1)

ICON_APPLICATION = "".join([
  "iVBORw0KGgoAAAANSUhEUgAAAEEAAABBCAYAAACO98lFAAAABmJLR0QAAAD/AABH24+SAAAACXBI",
  "WXMAAAsTAAALEwEAmpwYAAAAB3RJTUUH4gUNFTAqs8LZOgAAAAh0RVh0Q29tbWVudAD2zJa/AAAF",
  "IklEQVR42u1av29cRRCeedjOGfFDFAjSIEdIFNAAcYFCQweWKEBgiSQVQWCBCDRISCngHIkUFEZK",
  "dwhDFyRHCGjokNzQAf9C+A8ipMhnOzl/FH4X7S0zu7Pv7T7fSbfS6d29ez92v51v5pvZJZq3eZu3",
  "unHJh+NXAvWI7n+WvWP9ubtAtE9EQzo+St/XuFxfq1IPHv3It0xw1+fhXALvSET008h73iyAUDFW",
  "JkYBu0mycxzfVlVYmTkQ1JGF6BP5PVMg4IdI/2EDgz3sbqAMLlVRC0CCpSin0YFV5Aeh32dxWqXp",
  "RRquPCsg4Mn+kdhzthMfwm3j898XoER+S0Did7bRYWYc4+HWwnYwOkAfuK8ToGgHIqKz/cHy1CpG",
  "fFMrRFcZnvIUoqAax4rRVYpDCqvIjzIqyLI6AYpnU/QDAr60pMbPBgKuOUO2RgKWz7N3u3TrVkYH",
  "WXVGLFYiAzdTkJhGSxB7BcOAEc2tijuzLCDgi3q4DaaHF4khYMSKZnDPf5WJEnnpwDYHGPIDbojU",
  "DArTZglrl68vJaeCirtHJJBAcJb/e/9JgPDb8icHQZ/A+tTyQ8djXPJifkhc+mFz9Xr9/qnTCVCm",
  "z5hOc4e1hNYg4LLS15gvSEgJWQHIFVSftXSQVXYLYNt//PjklY8wsSVUokBa3Q4EDlChQTUESoiE",
  "wceeCAh4X+gTByyC04CJMcd/xMctKFG1Nn/YUmX3PD8lX/GEkBnC6Eq4a0s4+8FgQZRwHLHrBkzj",
  "CCBuFL7fry5A+HNv425QGEkxzhj3Lny+/XBqsBm3FwZ1vzqhQ0jjalZRn+Nnwpb7x9fv3UnJ03KU",
  "4pJBwNvOu0K5gmTXaOZ2QgP23dDFBg6ymSXAkOkI167+PHjQ8vjhfu+ixqyYDOEToUNC7euv/sbQ",
  "8vjnl/dvpCSu1JISSSDg9cg7NECQB2fN9/o1h7cSKVE+gaqna3V7sJjyiE83+wukDDKUXRZ1jOvr",
  "65XIf0Q82JgK327cS+nYbr8/krJxpPQ3Nwg7t2+OgrJYc5hoN02aY+SIEe7t3Bzlp0PM/gJlNd7t",
  "N6LdS5u2+9BSQpuuH73Mt6oeVsQVJHeVSVlp4lebS/tdEEIrUePf0rn9I/7n9wdwJoslVISVYDrH",
  "Bp2Qyc/GilgTR+MWnzQ6wBCovR7xa+1S/lecYktIm7UptERBwIvKRMTcdeaFQ1ZKFDHXdc6gGarG",
  "vWBDYlVIdkjRA4H/24Egbb3RMhcJkzfy2MMaH69SadGYYwz2x5ECFJ4l+LtPzdGhR8Tn85HiFxCG",
  "XjQ4UKKGexxf83dgP0MaHTSz5/I+AUK2qGkzzuUYD5/2tt4gwHlpdemdvDC86cxkaPOGlruFtvio",
  "ICwe3bsUzBNiS8iFmqXyLLmr4Zcbe+l0iBVIS+2YSIgUiMwDyFbnFUHAaUI0awn4AX63jD2c91ap",
  "pModB9zYGUUzVMn2h+5nPxT+OOKrLaW5KjpwyyZM7g4Yad3BqvLNdMBj3i40pqTFFP6wrGu8FNm/",
  "GLIOENFpgRKVqaLBhmpHAW1gsQoO6Adr1yZAuPNo74JJiyrEW/3XVlJv27Y3j9/jO0RrLudv8Znc",
  "JtNzZLJUIHF/C//zle5sYStSbBlGii63HVqllb263n6eiSZmOoC8vYig8E4076nPHe4sdTnQ7zZ3",
  "lpCoGt3WcxwkT4Dgmvop0nerC0e+2rVbJLrmZJaa6fvfD9xrakpUUXuKVTOmiBqWqDDFLJ63eZu3",
  "KWn/AdzVCZJQSb/0AAAAAElFTkSuQmCC",
])

paramvals = {}
paramvals["inidatadir"] = ""
paramvals["initreefile"] = ""
paramvals["kmcycmax"] = 50
paramvals["ldlist"] = ["."]
paramvals["winstype"] = "q"
paramvals["winsvalue"] = 10
paramvals["atenstart"] = 4
paramvals["atenstay"] = 8
paramvals["atencount"] = 200
paramvals["resnorm"] = 5
paramvals["basecount"] = 200
paramvals["project"] = 0
paramvals["vertical"] = 1
paramvals["extend"] = 2
paramvals["tclpath"] = ""
def sourceArgs(arglen, args, params):
  try:
    optlist, args = getopt.getopt(args, 'i:e:y:p:b:r:w:a:n:l:c:')
    for opt, arg in optlist:
      if "-w" == opt:
        value = ""
        type = ""
        if arg.endswith("l"):
          type = "l"
          value = arg[:-1]
        if arg.endswith("q"):
          type = "q"
          value = arg[:-1]
        if (1 == len(type)) and value.isdigit():
          if 10 > len(value):
            ivalue = int(value)
            if (0 < ivalue) and (21 > ivalue):
              params["winstype"] = type
              params["winsvalue"] = ivalue
      if "-a" == opt:
        atenlist = arg.split(":")
        if 3 == len(atenlist):
          intvalues = []
          isok = True
          for avalue in atenlist:
            if 10 < len(avalue):
              isok = False
              break
            if not avalue.isdigit():
              isok = False
              break
            else:
              ivalue = int(avalue)
              if 1 > ivalue:
                isok = False
                break
              else:
                intvalues.append(ivalue)
          if isok:
            params["atenstart"] = intvalues[0]
            params["atenstay"] = intvalues[1]
            params["atencount"] = intvalues[2]
      if "-y" == opt:
        if arg in ["0", "1", "2"]:
          params["vertical"] = int(arg)
      if "-e" == opt:
        if arg in ["1", "2"]:
          params["extend"] = int(arg)
      if "-n" == opt:
        if arg.isdigit():
          if 10 > len(arg):
            iarg = int(arg)
            params["resnorm"] = iarg
      if "-l" == opt:
        params["ldlist"] += [arg]
      if "-i" == opt:
        params["tclpath"] = arg
      if "-c" == opt:
        if arg.isdigit():
          if 10 > len(arg):
            iarg = int(arg)
            if 1 < iarg:
              params["kmcycmax"] = iarg
      if "-b" == opt:
        if arg.isdigit():
          if 10 > len(arg):
            iarg = int(arg)
            if 9 < iarg:
              params["basecount"] = iarg
      if "-p" == opt:
        if arg.isdigit():
          if 10 > len(arg):
            iarg = int(arg)
            params["project"] = iarg
    if 0 < len(args):
      for arg in args:
        auxfsname = arg
        if os.path.isdir(auxfsname):
          params["inidatadir"] = auxfsname
        elif os.path.isfile(auxfsname):
          params["inidatadir"] = os.path.dirname(auxfsname)
          params["initreefile"] = auxfsname
  except:
    log_err("wrong arguments")

if ('__main__' == __name__):
  args = sys.argv[1:]
  optlist = []
  arglen = len(args)
  sourceArgs(arglen, args, paramvals)
  paramvals["ldlist"] += ['/usr/local/lib/cludisc', '/usr/local/lib', '/usr/lib']

  cansvd = False
  try:
    if "linalg" in dir(N):
      if "svd" in dir(N.linalg):
        cansvd = True
  except:
    cansvd = False
  if not cansvd:
    if 0 == paramvals["project"]:
      paramvals["project"] = 200

clibcd = []
cKMSearch = []
cLocAdj = []
isclibcd = False
ldlist = paramvals["ldlist"]
for ldpath in ldlist:
  isclibcd = True
  try:
    libname = "libcludisc"
    clibcd = N.ctypeslib.load_library(libname, ldpath)
  except:
    isclibcd = False
  if isclibcd:
    break

if isclibcd:
  try:
    clibcd.cludiscversion.restype = C.c_int
    clibcd.cludiscversion.argtypes = [C.c_int, C.POINTER(C.c_long)]
    cdverval = N.empty(2, dtype=C.c_long)
    cdverval = N.ascontiguousarray(cdverval)
    cdverlen = clibcd.cludiscversion(2, cdverval.ctypes.data_as(C.POINTER(C.c_long)))
    if not 2 == cdverlen:
      isclibcd = False
    if isclibcd:
      if not 2 == cdverval[0]:
        isclibcd = False
  except:
    isclibcd = False

if isclibcd:
  try:
    clibcd.censearch.restype = C.c_int
    clibcd.censearch.argtypes = [C.c_int, C.c_int, C.c_int,
      C.POINTER(C.c_double), C.POINTER(C.c_double), C.POINTER(C.c_long), C.POINTER(C.c_long), C.POINTER(C.c_long)]

    def _cKMSearch(mrows, cencount, mcols, pdata, centers, cenlive, cranks, orients):
      return clibcd.censearch(mrows, cencount, mcols,
        pdata.ctypes.data_as(C.POINTER(C.c_double)),
        centers.ctypes.data_as(C.POINTER(C.c_double)),
        cenlive.ctypes.data_as(C.POINTER(C.c_long)),
        cranks.ctypes.data_as(C.POINTER(C.c_long)),
        orients.ctypes.data_as(C.POINTER(C.c_long))
      )
    cKMSearch = _cKMSearch

  except:
    isclibcd = False

if isclibcd:
  try:
    clibcd.locadj.argtypes = [C.c_int, C.c_int, C.c_double, 
      C.POINTER(C.c_double), C.POINTER(C.c_double),
      C.POINTER(C.c_double), C.POINTER(C.c_double), C.POINTER(C.c_double),
      C.c_double]

    def _cLocAdj(ecnt, extends, dzero, distdiffs, adjusted,
      curmovex, curmovey, curmovez, curaten):
      return clibcd.locadj(ecnt, extends, dzero,
        distdiffs.ctypes.data_as(C.POINTER(C.c_double)),
        adjusted.ctypes.data_as(C.POINTER(C.c_double)),
        curmovex.ctypes.data_as(C.POINTER(C.c_double)),
        curmovey.ctypes.data_as(C.POINTER(C.c_double)),
        curmovez.ctypes.data_as(C.POINTER(C.c_double)),
        curaten
      )
    cLocAdj = _cLocAdj

  except:
    isclibcd = False

if not isclibcd:
  log_err("No 'libcludisc' library available, slower implementation is used.")

class DataScale(object):
  def __init__(self, data, sels, params, rtparams=[], isclibcd=False):
    self.cspeed = isclibcd

    itrs = 0
    self.resnorm = 5
    self.extends = 2
    self.atenstart = 10
    self.atenstay = 10
    self.atencount = 100
    if 6 <= len(rtparams):
      itrs = rtparams[0]
      self.resnorm = rtparams[1]
      self.extends = rtparams[2]
      self.atenstart = rtparams[3]
      self.atenstay = rtparams[4]
      self.atencount = rtparams[5]

    self.data = data
    self.size = self.data.shape[0]
    self.ecnt = self.data.shape[1]

    self.sels = sels
    self.gsca = params[0]
    self.exct = params[1]
    self.adjs = params[2]
    self.locl = params[3]

    self.maxitrs = 200
    if 0 < itrs:
      self.maxitrs = itrs

    self.eigenvectors = []
    self.projections = []
    self.scales = []
    self.avgorisize = 0.0
    self.avgscasize = 0.0
    self.avgrensize = 0.0
    self.adjusted = []
    self.adjmoved = []

    self.distdiffs = []
    self.intfunc = []
    if 1 == self.extends:
      self.intfunc = lambda x: N.log(1+x)
    else:
      self.intfunc = N.arctan

    self.error = 0

  def releaseAdjusted(self):
    self.adjusted = []

  def updateLocals(self, params):
    self.adjs = params[0]
    self.locl = params[1]

  def getParameters(self):
    return [self.gsca, self.exct, self.adjs, self.locl]

  def getSelections(self):
    return self.sels

  def getAverages(self):
    return [self.avgorisize, self.avgscasize, self.avgrensize]

  def getScales(self):
    return self.scales

  def getEigenVectors(self):
    return self.eigenvectors

  def getProjections(self):
    return self.projections

  def getAdjusted(self):
    return self.adjusted


  def takeDistQuantiles(self):
    if [] == self.distdiffs:
      return

    ldivs = [20, 10, 20, 5, 4, 10, 20, 5, 20, 2]
    lmuls = [1, 1, 3, 1, 1, 3, 7, 2, 9, 1]
    self.quantiles = 10 * [0.0]

    oneline = self.distdiffs[0,1:]
    distlist = list(oneline[oneline > 0.0])
    for ind in range(1, (self.ecnt - 1)):
      oneline = self.distdiffs[ind,(ind+1):]
      distlist += list(oneline[oneline > 0.0])
    distlist.sort()

    curlen = len(distlist)
    if 0 == curlen:
      self.error = 501
      return

    for ind in range(10):
      curdiv = ldivs[ind]
      if 0 == (curlen % curdiv):
        qind = (curlen // curdiv) * lmuls[ind]
        qvalue = distlist[qind-1] + distlist[qind]
        self.quantiles[ind] = qvalue / 2.0
      else:
        qind = int((ind / 20.0) * curlen)
        self.quantiles[ind] = distlist[qind]


  def localAdjust(self):
    idzero = (self.locl // 5) - 1

    if [] == self.distdiffs:
      self.distdiffs = N.empty((self.ecnt, self.ecnt))
      self.distdiffs = N.ascontiguousarray(self.distdiffs)
      for ione in range(self.ecnt - 1):
        for isec in range((ione + 1), self.ecnt):
          oridist = N.sqrt(N.sum(N.square(self.data[:,ione] - self.data[:,isec])))
          self.distdiffs[ione,isec] = oridist
          self.distdiffs[isec,ione] = oridist
      self.takeDistQuantiles()

    #find the unit distance
    dzero = self.quantiles[idzero]

    self.adjusted = N.copy(self.projections)
    self.adjusted = N.ascontiguousarray(self.adjusted)

    curmovex = N.empty(self.ecnt)
    curmovey = N.empty(self.ecnt)
    curmovez = N.empty(self.ecnt)
    curmovex = N.ascontiguousarray(curmovex)
    curmovey = N.ascontiguousarray(curmovey)
    curmovez = N.ascontiguousarray(curmovez)

    #cycle through adjusting
    atenpos = self.atenstart
    atencnt = 0

    for run in range(self.atencount):
      curaten = 0.5 / atenpos
      atencnt += 1
      if atencnt == self.atenstay:
        atencnt = 0
        atenpos += 1

      curmovex[:] = 0.0
      curmovey[:] = 0.0
      curmovez[:] = 0.0

      if self.cspeed:
        cLocAdj(self.ecnt, self.extends, dzero, self.distdiffs, self.adjusted,
          curmovex, curmovey, curmovez, curaten)
        continue

      for ione in range(self.ecnt - 1):
        for isec in range((ione + 1), self.ecnt):
          oridist = self.distdiffs[ione,isec]
          curdiff = self.adjusted[ione,:] - self.adjusted[isec,:]
          curdist = N.sqrt(N.sum(N.square(curdiff)))
          curmoved = 0.0
          if curdist < oridist:
            curmoved = self.intfunc(oridist / dzero) - self.intfunc(curdist / dzero)
            curmoved *= dzero
          if curdist > oridist:
            currat = oridist / dzero
            if 2 == self.extends:
              currat = N.square(currat)
            curmoved = (oridist - curdist) / (1.0 + currat)
          xdiff = curdiff[0]
          ydiff = curdiff[1]
          zdiff = curdiff[2]
          tdiff = N.abs(xdiff) + N.abs(ydiff) + N.abs(zdiff)
          if 0.0 == tdiff:
            xdiff = 1.0
            ydiff = 1.0
            zdiff = 1.0
            tdiff = 3.0
          xtomove = (xdiff / tdiff) * curmoved
          ytomove = (ydiff / tdiff) * curmoved
          ztomove = (zdiff / tdiff) * curmoved
          curmovex[ione] += xtomove
          curmovey[ione] += ytomove
          curmovez[ione] += ztomove
          curmovex[isec] -= xtomove
          curmovey[isec] -= ytomove
          curmovez[isec] -= ztomove

      self.adjusted.T[0] += curmovex * curaten
      self.adjusted.T[1] += curmovey * curaten
      self.adjusted.T[2] += curmovez * curaten

    self.adjmoved = N.empty(self.ecnt)
    for ind in range(self.ecnt):
      self.adjmoved[ind] = N.sqrt(N.sum(N.square(self.adjusted[ind,:] - self.projections[ind,:])))
    self.adjmoved /= self.avgrensize

  def compProjections(self):
    if 0 < self.error:
      return

    try:
      mdata = N.asmatrix(self.data)
      mvecs = N.asmatrix(self.eigenvectors)

      self.projections = mdata.T * mvecs.T

      #compute self.projections/self.data scale ratios
      #take average sizes along with that
      totorisize = 0.0
      totscasize = 0.0
      self.scales = N.empty(self.ecnt)
      for ind in range(self.ecnt):
        orival = N.sqrt(N.sum(N.square(self.data[:,ind])))
        scaval = N.sqrt(N.sum(N.square(self.projections[ind,:])))
        totorisize += orival
        totscasize += scaval
        if 0.0 == orival:
          if 0.0 == scaval:
            self.scales[ind] = N.nan
          else:
            self.scales[ind] = N.inf
        else:
          self.scales[ind] = scaval / orival

      if 0.0 == totscasize:
        self.error = 401
        return

      self.avgorisize = totorisize / (1.0 * self.ecnt)
      self.avgscasize = totscasize / (1.0 * self.ecnt)

      if 0 < self.resnorm:
        self.avgrensize = 1.0 * self.resnorm
        self.projections *= self.avgrensize / self.avgscasize
      else:
        self.avgrensize = self.avgscasize

      dscale = self.avgrensize / self.avgorisize
      self.data *= dscale
      self.scales *= self.avgorisize / self.avgscasize

    except:
      self.projections = []
      self.error = 402

  def compEigenVecs(self):
    self.error = 0
    if 3 > self.size:
      self.error = 1
      return
    if 2 > self.ecnt:
      self.error = 2
      return
    if not N.any(self.data):
      self.error = 3
      return

    self.eigenvectors = N.zeros((3, self.size))

    if 1 == self.exct:
      self.compEigenVecsDecp()
    else:
      self.compEigenVecsIter()

    if 0 < self.error:
      return

    for ivec in range(3):
      for ind in range(self.size):
        curval = self.eigenvectors[ivec,ind]
        if 0 > curval:
          self.eigenvectors[ivec,:] *= -1
        elif 0 < curval:
          break

  def compEigenVecsDecp(self):
    try:
      mcov = N.cov(self.data)
      u,s,vh = N.linalg.svd(mcov)
      for ivec in range(3):
        self.eigenvectors[ivec] = u[:,ivec]
    except:
      self.error = 101

  def oneEigenVec(self, rank, dcovs):

    dsize = self.size
    runevs = N.asmatrix(self.eigenvectors)

    self.error = 0
    eigvec = N.empty(dsize)
    simvec = N.empty(dsize)

    for ind in range(dsize):
      simvec[ind] = N.sum(N.abs(dcovs[ind]))
    simrng = list(range(dsize))
    simrng.sort(key=lambda x: simvec[x])

    maxind = simrng[(dsize - 1)]
    maxval = simvec[maxind]
    if 0.0 == maxval:
      self.error = 201
      return []

    mcovs = N.asmatrix(dcovs)
    mvect = N.asmatrix(eigvec).T

    eigvec[maxind] = 1
    for indrng in range((dsize - 2), -1, -1):
      ind = simrng[indrng]
      cursize = simvec[ind] / maxval
      curvars = 0.0
      for pos in range((indrng + 1), dsize):
        curvars += dcovs[ind][simrng[pos]]
        if 0.0 > curvars:
          cursize *= -1.0
      eigvec[ind] = cursize
    vecsize = N.sqrt(mvect.T * mvect)
    if 0.0 == vecsize:
      self.error = 202
      return []
    mvect /= vecsize

    for itr in range(self.maxitrs):
      mvect = mcovs * mvect
      for iev in range(rank):
        ev = runevs[iev,:]
        mvect -= ((ev * mvect) * ev).T
      vecsize = N.sqrt(mvect.T * mvect)
      if 0.0 == vecsize:
        self.error = 203
        return []
      mvect /= vecsize

    for ind in range(dsize):
      eigvec[ind] = mvect[ind,0]
    return eigvec

  def compEigenVecsIter(self):
    mdata = N.asmatrix(self.data)
    data2 = N.empty((self.size, self.ecnt))

    mcov = N.cov(self.data, bias=1)
    evfirst = self.oneEigenVec(0, mcov)

    if 0 < self.error:
      return
    self.eigenvectors[0] = evfirst
    mvect = N.asmatrix(evfirst)

    for icol in range(self.ecnt):
      scal = mvect * mdata[:,icol]
      for irow in range(self.size):
        data2[irow,icol] = mdata[irow,icol] - scal * mvect[0, irow]

    mcov = N.cov(data2, bias=1)
    evsecond = self.oneEigenVec(1, mcov)

    if 0 < self.error:
      return
    self.eigenvectors[1] = evsecond
    mvect = N.asmatrix(evsecond)
    mdata2 = N.asmatrix(data2)

    for icol in range(self.ecnt):
      scal = mvect * mdata2[:,icol]
      for irow in range(self.size):
        data2[irow,icol] -= scal * mvect[0, irow]

    mcov = N.cov(data2, bias=1)
    evthird = self.oneEigenVec(2, mcov)

    if 0 < self.error:
      return
    self.eigenvectors[2] = evthird

class CanvTHand(object):
  def __init__(self, canv, tlnode):
    self.canv = canv
    self.tlnode = tlnode

  def inn(self, event):
    self.canv.itemconfigure(self.tlnode, fill="#000000")

  def out(self, event):
    self.canv.itemconfigure(self.tlnode, fill="#909090")

class CanvCB(object):
  def __init__(self, rank, tree):
    self.rank = rank
    self.tree = tree

  def treecb(self, event):
    self.tree.cclust(self.rank)

  def treetg(self, event):
    self.tree.tclust(self.rank)

  def treelf(self, event):
    self.tree.lclust(self.rank)

  def treecl(self, event):
    self.tree.eclust(self.rank)

class CanvSec(object):
  def __init__(self, rank, tree):
    self.rank = rank
    self.tree = tree

  def printOne(self, event):
    self.tree.printCanvasOne(self.rank, True, True)

  def printOneNN(self, event):
    self.tree.printCanvasOne(self.rank, True, False)

  def printSim(self, event):
    self.tree.printCanvasOne(self.rank, False, True)

  def printSimNN(self, event):
    self.tree.printCanvasOne(self.rank, False, False)

class TreeData(object):
  def __init__(self, tnodes, lcount, ecount, gsizes):
    self.tnodes = tnodes
    self.optview = True

    self.lcount = lcount
    self.ecount = ecount
    self.tcount = (2 * lcount) - 1
    self.icount = self.tcount - self.lcount
    self.amounts = gsizes
    self.variats = self.tcount * [0.0,]

    self.tstruc = N.empty((self.icount,2), dtype=int)

    self.totsize = 0
    for ind in range(self.lcount):
      self.variats[ind] = N.var(self.tnodes[ind])
      self.totsize += self.amounts[ind]

    self.states = self.tcount * [0,]

    #prepare topology structures
    self.nlower = self.tcount * [[],]
    self.nupper = self.tcount * [[],]

    self.lsizes = self.tcount * [0,]
    self.rsizes = self.tcount * [0,]

    self.lchild = self.tcount * [-1,]
    self.rchild = self.tcount * [-1,]

    self.posits = self.tcount * [-1,]

    self.nodelists = []

  def optTreeView(self):
    tsibl = N.empty(self.icount, dtype=int)
    tside = N.empty(self.icount, dtype=int)

    for ind in range((self.icount - 1), -1, -1):
      dfirst = abs(self.tstruc[ind][0]) - 1
      ifirst = dfirst - self.lcount
      dsecond = abs(self.tstruc[ind][1]) - 1
      isecond = dsecond - self.lcount
      if 0 <= ifirst:
        tsibl[ifirst] = dsecond
      if 0 <= isecond:
        tsibl[isecond] = dfirst

    vfirst = self.tstruc[self.icount - 1][0]
    ifirst = abs(vfirst) - 1
    vsecond = self.tstruc[self.icount - 1][1]
    isecond = abs(vsecond) - 1

    var1 = self.variats[ifirst]
    var2 = self.variats[isecond]
    mul1 = N.sqrt(self.amounts[ifirst])
    mul2 = N.sqrt(self.amounts[isecond])

    nleft, nright = [ifirst, isecond]
    rat1 = var1 * mul1
    if 1 < self.amounts[ifirst]:
      rat1 = (rat1 - 1) / (mul1 - 1)
    rat2 = var2 * mul2
    if 1 < self.amounts[isecond]:
      rat2 = (rat2 - 1) / (mul2 - 1)
    if rat1 < rat2:
      nleft, nright = [isecond, ifirst]
      self.tstruc[self.icount - 1][0] = vsecond
      self.tstruc[self.icount - 1][1] = vfirst

    ileft = nleft - self.lcount
    iright = nright - self.lcount
    if 0 <= ileft:
      tside[ileft] = 0
    if 0 <= iright:
      tside[iright] = 1

    for ind in range((self.icount - 2), -1, -1):
      vfirst = self.tstruc[ind][0]
      ifirst = abs(vfirst) - 1
      vsecond = self.tstruc[ind][1]
      isecond = abs(vsecond) - 1
      isibl = tsibl[ind]
      iside = tside[ind]

      nleft, nright = [ifirst, isecond]

      corarr1 = N.corrcoef(self.tnodes[ifirst], self.tnodes[isibl], bias=1)
      corarr2 = N.corrcoef(self.tnodes[isecond], self.tnodes[isibl], bias=1)
      corval1 = abs(corarr1[0][1])
      corval2 = abs(corarr2[0][1])

      if 0 == iside:
        if corval1 > corval2:
          nleft, nright = [isecond, ifirst]
          self.tstruc[ind][0] = vsecond
          self.tstruc[ind][1] = vfirst

      if 1 == iside:
        if corval1 < corval2:
          nleft, nright = [isecond, ifirst]
          self.tstruc[ind][0] = vsecond
          self.tstruc[ind][1] = vfirst

      ileft = nleft - self.lcount
      iright = nright - self.lcount
      if 0 <= ileft:
        tside[ileft] = 0
      if 0 <= iright:
        tside[iright] = 1

  def createTree(self):
    #prepare correlation structures
    corrArray = N.empty((self.tcount,self.tcount))
    corrArray[:,:] = -2
    cnegArray = N.empty((self.tcount,self.tcount))
    cnegArray[:,:] = -2

    inicor = N.corrcoef(self.tnodes[:self.lcount], rowvar=1, bias=1)
    corrArray[:self.lcount,:self.lcount] = inicor
    cnegArray[:self.lcount,:self.lcount] = -1 * inicor

    for ind in range(self.tcount):
      corrArray[ind,ind] = -2
      cnegArray[ind,ind] = -2

    #assign the current active nodes
    cactive = N.empty((self.tcount), dtype=int)
    cactive[:self.lcount] = 1
    cactive[self.lcount:self.tcount] = 0

    #prepare max_value structures
    tornts = N.ones((self.tcount), dtype=int)
    imaxs = N.empty((self.tcount), dtype=int)
    vmaxs = N.empty((self.tcount))
    vmaxs[:] = -2

    #create the inner nodes
    for node in range(self.lcount, self.tcount):
      #set max_value data
      for ind in range(self.tcount):
        if (1 == cactive[ind]):
          indmaxpos = N.argmax(corrArray[ind])
          indmaxneg = N.argmax(cnegArray[ind])
          maxpos = corrArray[ind][indmaxpos]
          maxneg = cnegArray[ind][indmaxneg]

          maxval = maxpos
          indmax = indmaxpos
          curori = 1
          if maxneg > maxpos:
            maxval = maxneg
            indmax = indmaxneg
            curori = -1

          imaxs[ind] = indmax
          vmaxs[ind] = maxval
          tornts[ind] = curori
        else:
          vmaxs[ind] = -2

      #take max_similar pair
      ifirst = N.argmax(vmaxs[:ind])
      isecond = imaxs[ifirst]
      iorient = tornts[ifirst]

      nleft, nright = [ifirst, isecond]
      amn1 = self.amounts[ifirst]
      amn2 = self.amounts[isecond]
      if amn1 < amn2:
        nleft, nright = [isecond, ifirst]

      strind = node - self.lcount
      self.tstruc[strind, 0] = nleft + 1
      if 1 == iorient:
        self.tstruc[strind, 1] = nright + 1
      else:
        self.tstruc[strind, 1] = -1 * (nright + 1)

      #combine the pair into a new node
      newcent = N.zeros((1,self.ecount))
      newamount = self.amounts[ifirst] + self.amounts[isecond]

      newcent += self.tnodes[nleft] * self.amounts[nleft]
      if 1 == iorient:
        newcent += self.tnodes[nright] * self.amounts[nright]
      else:
        newcent -= self.tnodes[nright] * self.amounts[nright]
      newcent /= newamount

      self.tnodes[node] = newcent
      self.amounts[node] = newamount
      self.variats[node] = N.var(newcent)

      cactive[node] = 1
      cactive[ifirst] = 0
      cactive[isecond] = 0

      #make corrs to all existing non_lower nodes
      for ind in range(node):
        if (1 == cactive[ind]):
          corarr = N.corrcoef(self.tnodes[node], self.tnodes[ind], bias=1)
          curpos = corarr[0][1]
          curneg = -1 * curpos
          corrArray[node][ind] = curpos
          corrArray[ind][node] = curpos
          cnegArray[node][ind] = curneg
          cnegArray[ind][node] = curneg

          corrArray[ind][ifirst] = -2
          corrArray[ind][isecond] = -2
          cnegArray[ind][ifirst] = -2
          cnegArray[ind][isecond] = -2

    self.optTreeView()

    for node in range(self.lcount, self.tcount):
      strind = node - self.lcount
      ifirst0 = self.tstruc[strind, 0]
      isecond0 = self.tstruc[strind, 1]

      ifirst = abs(ifirst0) - 1
      isecond = abs(isecond0) - 1

      nleft, nright = [ifirst, isecond]

      self.lchild[node] = nleft
      self.rchild[node] = nright

      self.lsizes[node] = self.lsizes[nleft] + self.rsizes[nleft] + 1
      self.rsizes[node] = self.lsizes[nright] + self.rsizes[nright] + 1

      #set the lower nodes
      self.nlower[node] = [ifirst, isecond]
      for val in self.nlower[ifirst]:
        self.nlower[node].append(val)
      for val in self.nlower[isecond]:
        self.nlower[node].append(val)


    ilast = self.tcount - 1
    self.nupper[ilast] = []
    self.posits[ilast] = self.lsizes[ilast]

    for ind in range(ilast, self.lcount - 1, -1):
       indl = self.lchild[ind]
       indr = self.rchild[ind]
       curpos = self.posits[ind]

       self.nupper[indl] = [ind,]
       self.nupper[indr] = [ind,]
       for iupp in self.nupper[ind]:
         self.nupper[indl].append(iupp)
         self.nupper[indr].append(iupp)

       self.posits[indl] = curpos - (self.rsizes[indl] + 1)
       self.posits[indr] = curpos + (self.lsizes[indr] + 1)

    self.nodeBases()

  def renewTree(self, treestruc):
    #assign the current active nodes
    self.tstruc = treestruc

    cactive = N.empty((self.tcount), dtype=int)
    cactive[:self.lcount] = 1
    cactive[self.lcount:self.tcount] = 0

    #create the inner nodes
    for node in range(self.lcount, self.tcount):
      #set max_value data
      strind = node - self.lcount
      ifirst0 = self.tstruc[strind, 0]
      isecond0 = self.tstruc[strind, 1]

      ifirst = abs(ifirst0) - 1
      isecond = abs(isecond0) - 1

      if (0 > ifirst) or (self.tcount <= ifirst):
        return False
      if (0 > isecond) or (self.tcount <= isecond):
        return False
      if not 1 == cactive[ifirst]:
        return False
      if not 1 == cactive[isecond]:
        return False

      newamount = self.amounts[ifirst] + self.amounts[isecond]
      newcent = self.tnodes[node]

      self.amounts[node] = newamount
      self.variats[node] = N.var(newcent)

      cactive[node] = 1
      cactive[ifirst] = 0
      cactive[isecond] = 0

    if self.optview:
      self.optTreeView()

    for node in range(self.lcount, self.tcount):
      #set max_value data
      strind = node - self.lcount
      ifirst0 = self.tstruc[strind, 0]
      isecond0 = self.tstruc[strind, 1]

      ifirst = abs(ifirst0) - 1
      isecond = abs(isecond0) - 1

      nleft, nright = [ifirst, isecond]

      self.lchild[node] = nleft
      self.rchild[node] = nright

      self.lsizes[node] = self.lsizes[nleft] + self.rsizes[nleft] + 1
      self.rsizes[node] = self.lsizes[nright] + self.rsizes[nright] + 1

      #set the lower nodes
      self.nlower[node] = [ifirst, isecond]
      for val in self.nlower[ifirst]:
        self.nlower[node].append(val)
      for val in self.nlower[isecond]:
        self.nlower[node].append(val)

    ilast = self.tcount - 1
    self.nupper[ilast] = []
    self.posits[ilast] = self.lsizes[ilast]

    for ind in range(ilast, self.lcount - 1, -1):
       indl = self.lchild[ind]
       indr = self.rchild[ind]
       curpos = self.posits[ind]

       self.nupper[indl] = [ind,]
       self.nupper[indr] = [ind,]
       for iupp in self.nupper[ind]:
         self.nupper[indl].append(iupp)
         self.nupper[indr].append(iupp)

       self.posits[indl] = curpos - (self.rsizes[indl] + 1)
       self.posits[indr] = curpos + (self.lsizes[indr] + 1)

    self.nodeBases()
    return True

  def nodeBases(self):
    self.nodelists = self.tcount * [[],]
    for ind in range(self.lcount):
      self.nodelists[ind] = [ind + 1]
    for ind in range(self.lcount, self.tcount):
      self.nodelists[ind] = []
      strind = ind - self.lcount
      ifirst0 = self.tstruc[strind, 0]
      isecond0 = self.tstruc[strind, 1]
      ifirst = abs(ifirst0) - 1
      isecond = abs(isecond0) - 1
      if 0 < ifirst0:
        for ilist in self.nodelists[ifirst]:
          self.nodelists[ind].append(ilist)
      else:
        for ilist in self.nodelists[ifirst]:
          self.nodelists[ind].append(-1 * ilist)
      if 0 < isecond0:
        for ilist in self.nodelists[isecond]:
          self.nodelists[ind].append(ilist)
      else:
        for ilist in self.nodelists[isecond]:
          self.nodelists[ind].append(-1 * ilist)
      self.nodelists[ind].sort(key=lambda x: abs(x))

  def getNodeLists(self, ind):
    if len(self.nodelists) <= ind:
      return []
    else:
      return self.nodelists[ind]

class AxeTree(object):
  def __init__(self, master):
    self.canvcb = []
    self.canvsec = []
    self.tnodes = []
    self.tnodes2 = []
    self.tnodesout = []
    self.tree = []
    self.tntarg = -1

    self.printCanvas = []
    self.printCanvas2 = []
    self.vertical = 0

    self.minrotcnt = 3
    self.outergui = []
    self.outervar = []

    self.atf = T.Frame(master)
    self.atf.pack(side="bottom", fill="x", expand="1")

    self.bgcolor = "#ffffff"
    self.canv = T.Canvas(self.atf, width="15c", height="10c", background=self.bgcolor)
    self.canv2 = T.Canvas(self.atf, width="15c", height="10c", background=self.bgcolor)

    self.canv.config(scrollregion="0c 0c 15c 10c")
    self.pswidth = "15.0c"
    self.pswidth0 = "14.0c"
    self.mmwidth = 150

    self.csbx = T.Scrollbar(self.atf, command=self.canv.xview, orient="horizontal")
    self.canv.configure(xscrollcommand=self.csbx.set)

    self.canv.grid(row=0, column=0, sticky="we")
    self.csbx.grid(row=1, column=0, sticky="we")
    self.atf.columnconfigure(0, weight="1")

    self.csbx.bind("<Button-1>", self.tsbfocus)

  def delTree(self):
    self.canv.delete("all")
    self.canv2.delete("all")
    self.tree = []

    self.canv.config(scrollregion="0c 0c 15c 10c")
    self.pswidth = "15.0c"
    self.pswidth0 = "14.0c"
    self.mmwidth = 150

    self.canvcb = []
    self.canvsec = []
    self.tnodes = []
    self.tnodes2 = []
    self.tnodesout = []
    self.tntarg = -1

    self.outervar[0].set("")
    for wdg in self.outergui:
      wdg.configure(state="disabled")

    self.csbx.configure(command=self.canv.xview)

  def setTree(self, tree):

    if [] == self.tree:
      self.delTree()
    self.tree = tree
    self.updateGlobInfo()

    self.csbx.configure(command=self.canv.xview)

    if 100 == tree.lcount:
      self.canv.config(scrollregion="0.0c 0.0c 120.4c 10.0c")
      self.pswidth = "120.4c"
      self.mmwidth = 1204
    else:
      mcnt = 100 - tree.lcount
      mval = 12 * mcnt
      hend = 1204 - mval
      mmend = hend
      if 150 > mmend:
        mmend = 150
      psend = str(mmend)[0:-1] + "." + str(mmend)[-1:]
      self.canv.config(scrollregion="0c 0c " + psend + "c 10c")
      self.pswidth = str(hend)[0:-1] + "." + str(hend)[-1:] + "c"
      self.pswidth0 = str(hend-10)[0:-1] + "." + str(hend-10)[-1:] + "c"
      self.mmwidth = hend

    imgLr = self.canv.create_rectangle("0.0c", "0.0c", "1.0c", "10.0c", fill=self.bgcolor, outline=self.bgcolor, width=0)
    imgRr = self.canv.create_rectangle(self.pswidth0, "0.0c", self.pswidth, "10.0c", fill=self.bgcolor, outline=self.bgcolor, width=0)
    self.canv.tag_bind(imgLr, "<Double-Button-1>", self.printCanvasAll)
    self.canv.tag_bind(imgRr, "<Double-Button-1>", self.printCanvasAll)
    self.canv.tag_bind(imgLr, "<Control-Double-Button-1>", self.printCanvasSel)
    self.canv.tag_bind(imgRr, "<Control-Double-Button-1>", self.printCanvasSel)
    self.canv.tag_bind(imgLr, "<Double-Button-2>", self.printCanvasSel)
    self.canv.tag_bind(imgRr, "<Double-Button-2>", self.printCanvasSel)
    self.canv.tag_bind(imgLr, "<Double-Button-3>", self.printCanvasSel)
    self.canv.tag_bind(imgRr, "<Double-Button-3>", self.printCanvasSel)
    self.canv.tag_bind(imgLr, "<Shift-Double-Button-1>", self.printCanvasAllNN)
    self.canv.tag_bind(imgRr, "<Shift-Double-Button-1>", self.printCanvasAllNN)
    self.canv.tag_bind(imgLr, "<Shift-Control-Double-Button-1>", self.printCanvasSelNN)
    self.canv.tag_bind(imgRr, "<Shift-Control-Double-Button-1>", self.printCanvasSelNN)
    self.canv.tag_bind(imgLr, "<Shift-Double-Button-2>", self.printCanvasSelNN)
    self.canv.tag_bind(imgRr, "<Shift-Double-Button-2>", self.printCanvasSelNN)
    self.canv.tag_bind(imgLr, "<Shift-Double-Button-3>", self.printCanvasSelNN)
    self.canv.tag_bind(imgRr, "<Shift-Double-Button-3>", self.printCanvasSelNN)

    imgTr = self.canv.create_rectangle("1.0c", "0.0c", self.pswidth0, "1.0c", fill=self.bgcolor, outline=self.bgcolor, width=0)
    self.canv.tag_bind(imgTr, "<Double-Button-1>", self.printCanvasAll)
    self.canv.tag_bind(imgTr, "<Control-Double-Button-1>", self.printCanvasWin)
    self.canv.tag_bind(imgTr, "<Double-Button-2>", self.printCanvasWin)
    self.canv.tag_bind(imgTr, "<Double-Button-3>", self.printCanvasWin)
    self.canv.tag_bind(imgTr, "<Shift-Double-Button-1>", self.printCanvasAllNN)
    self.canv.tag_bind(imgTr, "<Shift-Control-Double-Button-1>", self.printCanvasWinNN)
    self.canv.tag_bind(imgTr, "<Shift-Double-Button-2>", self.printCanvasWinNN)
    self.canv.tag_bind(imgTr, "<Shift-Double-Button-3>", self.printCanvasWinNN)


    self.colstd = tree.tcount * ["",]
    self.colsel = tree.tcount * ["",]
    self.colinh = tree.tcount * ["",]

    self.xcens = tree.tcount * [0,]
    self.ycens = tree.tcount * [0,]

    hdcifs = "0123456789abcdef"
    for ind in range(tree.tcount):
      ratio = 0.0
      size = self.tree.amounts[ind]
      vmax = N.sqrt(size)
      vart = vmax * N.sqrt(self.tree.variats[ind])
      if 1 < size:
        ratio = (vart - 1) / (vmax - 1)
      elif 1 == size:
        ratio = vart
        if 0.000001 > ratio:
          ratio = 0.0
        if 0.999999 < ratio:
          ratio = 1.0
      if 0.0 > ratio:
        ratio = 0.0
      if 1.0 < ratio:
        ratio = 1.0

      fgreen = round(ratio * 160)
      fred = 160 - fgreen

      igreen0 = int(fgreen // 16)
      sgreen0 = hdcifs[igreen0]
      igreen1 = int(fgreen % 16)
      sgreen1 = hdcifs[igreen1]
      ired0 = int(fred // 16)
      sred0 = hdcifs[ired0]
      ired1 = int(fred % 16)
      sred1 = hdcifs[ired1]
      scolstd = "#" + sred0 + sred1 + sgreen0 + sgreen1 + "00"

      sgreen0 = hdcifs[igreen0 + 5]
      sred0 = hdcifs[ired0 + 5]
      scolsel = "#" + sred0 + sred1 + sgreen0 + sgreen1 + "50"

      fgreen = round(ratio * 80)
      fred = 80 - fgreen

      igreen0 = int(fgreen // 16)
      sgreen0 = hdcifs[igreen0 + 2]
      igreen1 = int(fgreen % 16)
      sgreen1 = hdcifs[igreen1]
      ired0 = int(fred // 16)
      sred0 = hdcifs[ired0 + 2]
      ired1 = int(fred % 16)
      sred1 = hdcifs[ired1]
      scolinh = "#" + sred0 + sred1 + sgreen0 + sgreen1 + "20"

      self.colstd[ind] = scolstd
      self.colsel[ind] = scolsel
      self.colinh[ind] = scolinh

    ilast = tree.tcount - 1
    totcnt = tree.amounts[ilast]
    totmin = N.min(tree.amounts)
    ymin0 = (1.0 * totmin) / totcnt
    if 1 == self.vertical:
      ymin0 = N.sqrt(ymin0)
    if 2 == self.vertical:
      ymin0 = N.sqrt(1.0 - ((1.0 - ymin0) ** 2))
    ymul0 = 1.0
    if 1.0 > ymin0:
      ymul0 = 1.0 / (1.0 - ymin0)
    for ind in range(tree.tcount):
      ncnt = tree.amounts[ind]
      npos = tree.posits[ind]

      xcen = 8 + (npos * 6)
      ycen0 = (1.0 * ncnt) / totcnt
      if 1 == self.vertical:
        ycen0 = N.sqrt(ycen0)
      if 2 == self.vertical:
        ycen0 = N.sqrt(1.0 - ((1.0 - ycen0) ** 2))
      ycen = 6.0 + 80.0 * ((1 - ycen0) * ymul0)
      self.xcens[ind] = xcen
      self.ycens[ind] = ycen

    self.tlabs = []
    self.tlabs2 = []
    self.thnds = []
    fontind = "Roman 9"
    for ind in range(tree.tcount):
      centh = self.xcens[ind]
      posh0 = centh - 2
      posh1 = posh0 + 4

      centv = self.ycens[ind]
      posv0 = centv - 2
      posv1 = posv0 + 4

      x0s = str(posh0 / 10.0) + "c"
      x1s = str(posh1 / 10.0) + "c"
      y0s = str(posv0 / 10.0) + "c"
      y1s = str(posv1 / 10.0) + "c"

      nodeout = self.canv.create_rectangle(x0s, y0s, x1s, y1s, fill=self.bgcolor, outline=self.bgcolor, width=3)

      self.tnodesout.append(nodeout)

      tx = str(centh / 10.0) + "c"
      ty = "9.3c"
      fc = "#909090"
      if ind < self.tree.lcount:
        ty = "9.6c"
      tlnode = self.canv.create_text(tx, ty, text=str(ind+1), fill=fc, font=fontind)
      tlnode2 = self.canv2.create_text(tx, ty, text=str(ind+1), fill=fc, font=fontind)
      self.canvsec.append(CanvSec(ind, self))
      if ind >= self.tree.lcount:
        self.canv.tag_bind(tlnode, "<Double-Button-1>", self.canvsec[ind].printOne)
        self.canv.tag_bind(tlnode, "<Control-Double-Button-1>", self.canvsec[ind].printSim)
        self.canv.tag_bind(tlnode, "<Double-Button-2>", self.canvsec[ind].printSim)
        self.canv.tag_bind(tlnode, "<Double-Button-3>", self.canvsec[ind].printSim)
        self.canv.tag_bind(tlnode, "<Shift-Double-Button-1>", self.canvsec[ind].printOneNN)
        self.canv.tag_bind(tlnode, "<Shift-Control-Double-Button-1>", self.canvsec[ind].printSimNN)
        self.canv.tag_bind(tlnode, "<Shift-Double-Button-2>", self.canvsec[ind].printSimNN)
        self.canv.tag_bind(tlnode, "<Shift-Double-Button-3>", self.canvsec[ind].printSimNN)
      self.tlabs.append(tlnode)
      self.tlabs2.append(tlnode2)
      tlhnd = CanvTHand(self.canv, tlnode)
      self.thnds.append(tlhnd)

    self.tlinesL = []
    self.tlinesR = []
    ilast = tree.tcount - 1
    for ind in range(ilast, tree.lcount - 1, -1):
      llow = tree.lchild[ind]
      rlow = tree.rchild[ind]

      upx = self.xcens[ind] / 10.0
      upy = self.ycens[ind] / 10.0

      llx = self.xcens[llow] / 10.0
      lly = self.ycens[llow] / 10.0

      rlx = self.xcens[rlow] / 10.0
      rly = self.ycens[rlow] / 10.0

      x0s = str(upx) + "c"
      y0s = str(upy) + "c"

      x1s = str(llx) + "c"
      y1s = str(lly) + "c"
      node = self.canv.create_line(x0s, y0s, x1s, y1s, fill="#804000")
      node2 = self.canv2.create_line(x0s, y0s, x1s, y1s, fill="#804000")
      self.tlinesL.append(node2)

      x1s = str(rlx) + "c"
      y1s = str(rly) + "c"
      node = self.canv.create_line(x0s, y0s, x1s, y1s, fill="#804000")
      node2 = self.canv2.create_line(x0s, y0s, x1s, y1s, fill="#804000")
      self.tlinesR.append(node2)

    for ind in range(tree.tcount):
      centh = self.xcens[ind]
      posh0 = centh - 2
      posh1 = posh0 + 4

      centv = self.ycens[ind]
      posv0 = centv - 2
      posv1 = posv0 + 4

      colfill = self.colsel[ind]
      coloutl = self.colstd[ind]

      x0s = str(posh0 / 10.0) + "c"
      x1s = str(posh1 / 10.0) + "c"
      y0s = str(posv0 / 10.0) + "c"
      y1s = str(posv1 / 10.0) + "c"
      node = self.canv.create_oval(x0s, y0s, x1s, y1s, fill=colfill, outline=coloutl, width=3)
      node2 = self.canv2.create_oval(x0s, y0s, x1s, y1s, fill=colfill, outline=coloutl, width=3)
      self.canvcb.append(CanvCB(ind, self))
      self.canv.tag_bind(node, "<Button-1>", self.canvcb[ind].treecb)
      self.canv.tag_bind(node, "<Double-Button-1>", self.canvcb[ind].treetg)
      self.canv.tag_bind(node, "<Shift-Button-1>", self.canvcb[ind].treetg)
      self.canv.tag_bind(node, "<Control-Button-1>", self.canvcb[ind].treelf)
      self.canv.tag_bind(node, "<Button-2>", self.canvcb[ind].treelf)
      self.canv.tag_bind(node, "<Button-3>", self.canvcb[ind].treelf)
      self.canv.tag_bind(node, "<Control-Double-Button-1>", self.canvcb[ind].treecl)
      self.canv.tag_bind(node, "<Control-Button-2>", self.canvcb[ind].treecl)
      self.canv.tag_bind(node, "<Control-Button-3>", self.canvcb[ind].treecl)
      self.canv.tag_bind(node, "<Double-Button-2>", self.canvcb[ind].treecl)
      self.canv.tag_bind(node, "<Double-Button-3>", self.canvcb[ind].treecl)

      self.tnodes.append(node)
      self.tnodes2.append(node2)
      self.canv.tag_bind(node, "<Enter>", self.thnds[ind].inn)
      self.canv.tag_bind(node, "<Leave>", self.thnds[ind].out)

    if 18 >= tree.lcount:
      self.csbx.configure(command="")


  def tclust(self, rank):
    if [] == self.tree:
      return

    if -1 == self.tntarg:
      nodeout = self.tnodesout[rank]
      self.canv.itemconfigure(nodeout, fill="#808080", outline="#808080")
      self.tntarg = rank
    else:
      nodeout = self.tnodesout[self.tntarg]
      self.canv.itemconfigure(nodeout, fill="#f8f8f8", outline="#f8f8f8")
      if rank == self.tntarg:
        self.tntarg = -1
      else:
        nodeout = self.tnodesout[rank]
        self.canv.itemconfigure(nodeout, fill="#808080", outline="#808080")
        self.tntarg = rank

    self.updateListInfo()

  def updateListInfo(self):
    tlist = ()

    if [] == self.tree:
      self.outervar[0].set(tlist)
      return

    if -1 == self.tntarg:
      self.updateGlobInfo()
    else:
      ratio = 0.0
      size = self.tree.amounts[self.tntarg]
      vmax = N.sqrt(size)
      vart = vmax * N.sqrt(self.tree.variats[self.tntarg])
      if 1 < size:
        ratio = (vart - 1) / (vmax - 1)
      elif 1 == size:
        ratio = vart
        if 0.000001 > ratio:
          ratio = 0.0
        if 0.999999 < ratio:
          ratio = 1.0
      if 0.0 > ratio:
        ratio = 0.0
      if 1.0 < ratio:
        ratio = 1.0

      wvart = "%(val)1.4e" % {'val': vart}
      wsize = str(size)
      wratio = "%(val).4f" % {'val': (100 * ratio)}
      tlist = ('Targeted', 'rank:   ' + str(self.tntarg + 1), 'size:    ' + wsize, 'scale:  ' + wvart, 'ratio:   ' + wratio + "%")

      self.outervar[0].set(tlist)

  # states: -n .. upper, 0 .. std, 1 selected, 2 lower
  def cclust(self, rank):
    if [] == self.tree:
      return

    curstate = self.tree.states[rank]

    if 0 > curstate:
      return
    if 2 == curstate:
      return

    if 0 == curstate:
      self.selectOne(rank)
      return

    if 1 == curstate:
      self.releaseOne(rank)
      return

  def lclust(self, rank):
    if [] == self.tree:
      return

    for ind in self.tree.nlower[rank]:
      if ind < self.tree.lcount:
        self.selectOne(ind)

  def eclust(self, rank):
    if [] == self.tree:
      return

    self.releaseOne(rank)
    for ind in self.tree.nlower[rank]:
      self.releaseOne(ind)

  def tsbfocus(self, e):
    self.csbx.focus()

  def selectOne(self, rank):
    if [] == self.tree:
      return

    if 0.0 == self.tree.variats[rank]:
      return

    if 0 == self.tree.states[rank]:
      newstate = 1

      cnode = self.tnodes[rank]
      self.canv.itemconfigure(cnode, outline="#0000ff")
      for ind in self.tree.nlower[rank]:
        self.tree.states[ind] = 2
        cnode = self.tnodes[ind]
        self.canv.itemconfigure(cnode, fill=self.colinh[ind])
      for ind in self.tree.nupper[rank]:
        self.tree.states[ind] -= 1
        cnode = self.tnodes[ind]
        self.canv.itemconfigure(cnode, fill=self.colinh[ind])

      self.tree.states[rank] = newstate
      self.updateSums()

  def releaseOne(self, rank):
    if [] == self.tree:
      return

    if 1 == self.tree.states[rank]:
      newstate = 0

      cnode = self.tnodes[rank]
      self.canv.itemconfigure(cnode, outline=self.colstd[rank])
      for ind in self.tree.nlower[rank]:
        self.tree.states[ind] = 0
        cnode = self.tnodes[ind]
        self.canv.itemconfigure(cnode, fill=self.colsel[ind])
      for ind in self.tree.nupper[rank]:
        self.tree.states[ind] += 1
        if 0 == self.tree.states[ind]:
          cnode = self.tnodes[ind]
          self.canv.itemconfigure(cnode, fill=self.colsel[ind])

      self.tree.states[rank] = newstate
      self.updateSums()

  def releaseAll(self):
    if [] == self.tree:
      return

    for rank in range(self.tree.tcount):
      if 1 == self.tree.states[rank]:
        self.releaseOne(rank)

  def selectBottom(self):
    if [] == self.tree:
      return

    for rank in range(self.tree.lcount, self.tree.tcount):
      if 1 == self.tree.states[rank]:
        self.releaseOne(rank)
    for rank in range(self.tree.lcount):
      if 0 == self.tree.states[rank]:
        self.selectOne(rank)

  def updateSums(self):
    if [] == self.tree:
      return

    totsel = 0
    wstate = "disabled"
    for rank in range(self.tree.tcount):
      if 1 == self.tree.states[rank]:
        totsel += 1
    if self.minrotcnt <= totsel:
      wstate = "normal"
    #update list var.
    if -1 == self.tntarg:
      self.updateGlobInfo()
    #update project state
    for wdg in self.outergui:
      wdg.configure(state=wstate)

  def updateGlobInfo(self):
    if [] == self.tree:
      return

    totsel = 0
    selsum = 0
    for rank in range(self.tree.tcount):
      if 1 == self.tree.states[rank]:
        totsel += 1
        selsum += self.tree.amounts[rank]

    strbot = "bottom:    " + str(self.tree.lcount)
    strall = "summed:  " + str(self.tree.totsize)
    strsel = "selected:  " + str(totsel)
    strsum = "summed:  " + str(selsum)
    tlist = ('Overall', strbot, strall, strsel, strsum)
    self.outervar[0].set(tlist)

  def getSelections(self):
    sels = []
    if [] == self.tree:
      return sels

    for rank in range(self.tree.tcount):
      if 1 == self.tree.states[rank]:
        sels.append(rank)

    return sels

  def printCanvasAll(self, event):
    rv, fname = self.printCanvas(True, True)
    self.updateImageInfo("all", rv, fname)

  def printCanvasAllNN(self, event):
    rv, fname = self.printCanvas(True, False)
    self.updateImageInfo("all", rv, fname)

  def printCanvasWin(self, event):
    rv, fname = self.printCanvas(False, True)
    self.updateImageInfo("win", rv, fname)

  def printCanvasWinNN(self, event):
    rv, fname = self.printCanvas(False, False)
    self.updateImageInfo("win", rv, fname)

  def printCanvasOne(self, index, withnodes, withlabels):
    lincnt = (2 * (self.tree.lcount - 1)) - index
    tcount = (2 * self.tree.lcount) - 1

    centh = self.xcens[index]
    centv = self.ycens[index]

    ldisp = self.tree.lsizes[index]
    rdisp = self.tree.rsizes[index]

    x0bor = int(centh - (8 + (6 * ldisp)))
    x1bor = int(centh + (8 + (6 * rdisp)))
    y0bor = int(centv - 6)
    y1bor = 100
    if not withlabels:
      y1bor = 90

    nodbeg = 0
    if withnodes:
      nodbeg = index + 1

    for lind in range(lincnt):
      rlin = self.tlinesL[lind]
      llin = self.tlinesR[lind]
      self.canv2.move(llin, "0.0c", "10.0c")
      self.canv2.move(rlin, "0.0c", "10.0c")

    for lind in range((index+1), tcount):
      tlab = self.tlabs2[lind]
      self.canv2.move(tlab, "0.0c", "10.0c")

    for lind in range(nodbeg, tcount):
      tnode = self.tnodes2[lind]
      self.canv2.move(tnode, "0.0c", "10.0c")

    rv, fname = self.printCanvas2(x0bor, y0bor, x1bor, y1bor)
    self.updateImageInfo("one", rv, fname)

    for lind in range(lincnt):
      rlin = self.tlinesL[lind]
      llin = self.tlinesR[lind]
      self.canv2.move(llin, "0.0c", "-10.0c")
      self.canv2.move(rlin, "0.0c", "-10.0c")

    for lind in range((index+1), tcount):
      tlab = self.tlabs2[lind]
      self.canv2.move(tlab, "0.0c", "-10.0c")

    for lind in range(nodbeg, tcount):
      tnode = self.tnodes2[lind]
      self.canv2.move(tnode, "0.0c", "-10.0c")

  def printCanvasSel(self, withlabels=True):
    lcount = self.tree.lcount
    tcount = (2 * lcount) - 1

    tmove = []

    for ind in range(tcount):
      state = self.tree.states[ind]
      if not 1 == state:
        tmove.append(ind)

    for lind in tmove:
      tnode = self.tnodes2[lind]
      self.canv2.move(tnode, "0.0c", "10.0c")

    x0bor = 0
    x1bor = self.mmwidth
    y0bor = 0
    y1bor = 100
    rv, fname = self.printCanvas2(x0bor, y0bor, x1bor, y1bor, True)
    self.updateImageInfo("sel", rv, fname)

    for lind in tmove:
      tnode = self.tnodes2[lind]
      self.canv2.move(tnode, "0.0c", "-10.0c")

  def printCanvasSelNN(self, withlabels=True):
    lcount = self.tree.lcount
    tcount = (2 * lcount) - 1

    tmove = []

    for ind in range(tcount):
      state = self.tree.states[ind]
      if not 1 == state:
        tmove.append(ind)

    for lind in tmove:
      tnode = self.tnodes2[lind]
      self.canv2.move(tnode, "0.0c", "10.0c")

    x0bor = 0
    x1bor = self.mmwidth
    y0bor = 0
    y1bor = 90
    rv, fname = self.printCanvas2(x0bor, y0bor, x1bor, y1bor, True)
    self.updateImageInfo("sel", rv, fname)

    for lind in tmove:
      tnode = self.tnodes2[lind]
      self.canv2.move(tnode, "0.0c", "-10.0c")

  def updateImageInfo(self, method, state, fname):
    if not state:
      if None == fname:
        return

    tlist = ("Cluster-Tree Image", "")
    if not state:
      tlist += ("shot failed",)
    else:
      tlist += ("shot written",)

      tinfo = ""
      if method in ["all", "sel"]:
        tinfo = "cluster tree"
      if "win" == method:
        tinfo = "tree window"
      if "one" == method:
        tinfo = "node subtree"
      tlist += (tinfo, fname)

    self.outervar[0].set(tlist)

class DataTable(object):
  def __init__(self, master, rnum, cnum, radd=2, cadd=0):
    self.data = []
    self.rcnt = 0
    self.ccnt = 0

    self.cells = []
    self.rlabs = []
    self.clabs = []
    self.tabed = 0
    self.rows = rnum + radd
    self.cols = cnum + cadd
    self.rres = rnum
    self.cres = cnum

    self.outergui = []

    self.ractv = []
    self.cactv = []
    self.rtvar = T.IntVar(0)
    self.ctvar = T.IntVar(0)

    self.rsel = False
    self.csel = False
    self.rfrom = -1
    self.rto = -1
    self.rhead = -1
    self.cfrom = -1
    self.cto = -1
    self.chead = -1

    self.rfgcols = self.rows * [0, ]
    self.rbgcols = self.rows * [0, ]
    self.cfgcols = self.cols * [0, ]
    self.cbgcols = self.cols * [0, ]

    self.sfgcols = [["#000000", "#00b000"], ["#b00000", "#202000"]]
    self.sbgcols = [["#f8f8f8", "#e8ffe8"], ["#ffe8e8", "#ffffe8"]]

    self.rlins = []
    self.clins = []
    self.rlsts = []
    self.clsts = []
    for ind in range(self.rows):
      self.ractv.append(0)
      self.rlins.append(0)
      self.rlsts.append(0)
    for ind in range(self.cols):
      self.cactv.append(0)
      self.clins.append(0)
      self.clsts.append(0)

    c0cf = T.Frame(master)
    c1cf = T.Frame(master)
    i0cf = T.Frame(master)
    r0cf = T.Frame(master)
    r1cf = T.Frame(master)
    d0tf = T.Frame(master)
    c0cf.grid(row=0, column=2)
    i0cf.grid(row=0, column=1)
    c1cf.grid(row=0, column=3)
    r0cf.grid(row=1, column=0)
    r1cf.grid(row=1, column=4, sticky="e", padx=8)
    d0tf.grid(row=1, column=2, columnspan=2, sticky="w")

    self.rds = T.Scale(r1cf, orient="vertical", from_=0, to=0, length=200, resolution=self.rres, command=self.setRange, variable=self.rtvar, bd=1)
    self.rds.configure(takefocus="on")
    self.rds.bind("<Button-1>", self.setRScaleFocus)

    self.cds = T.Scale(c1cf, orient="horizontal", from_=0, to=0, length=180, resolution=self.cres, command=self.setRange, variable=self.ctvar, bd=1)
    self.cds.bind("<Button-1>", self.setCScaleFocus)

    self.rselrb = T.IntVar()
    self.rselrb.set(1)
    self.rst1 = T.Radiobutton(r0cf, text="IDs", variable=self.rselrb, value=1, bd=1)
    self.rvt1 = T.Label(r0cf, text="---\n ")
    self.rst2 = T.Radiobutton(r0cf, text="start", variable=self.rselrb, value=2, bd=1)
    self.rvt2 = T.Label(r0cf, text="---\n ")
    self.rst3 = T.Radiobutton(r0cf, text="end", variable=self.rselrb, value=3, bd=1)
    self.rvt3 = T.Label(r0cf, text="---\n ")

    ivt0 = T.Label(i0cf, text="   ")
    ivt0.grid(row=0, column=0)

    self.cselrb = T.IntVar()
    self.cselrb.set(1)
    self.cvt0 = T.Label(c0cf, text="   ", width=5)
    self.cst1 = T.Radiobutton(c0cf, text="IDs:", variable=self.cselrb, value=1, bd=1)
    self.cvt1 = T.Label(c0cf, text="---", width=6, anchor="w")
    self.cst2 = T.Radiobutton(c0cf, text="start:", variable=self.cselrb, value=2, bd=1)
    self.cvt2 = T.Label(c0cf, text="---", width=6, anchor="w")
    self.cst3 = T.Radiobutton(c0cf, text="end:", variable=self.cselrb, value=3, bd=1)
    self.cvt3 = T.Label(c0cf, text="---", width=6, anchor="w")

    self.rds.grid(row=0, column=0)
    self.cds.grid(row=0, column=0)

    self.rst1.grid(row=0, column=0, sticky="w")
    self.rvt1.grid(row=1, column=0)
    self.rst2.grid(row=2, column=0, sticky="w")
    self.rvt2.grid(row=3, column=0)
    self.rst3.grid(row=4, column=0, sticky="w")
    self.rvt3.grid(row=5, column=0)

    self.cvt0.grid(row=0, column=0)
    self.cst1.grid(row=0, column=1)
    self.cvt1.grid(row=0, column=2, sticky="w")
    self.cst2.grid(row=0, column=3)
    self.cvt2.grid(row=0, column=4, sticky="w")
    self.cst3.grid(row=0, column=5)
    self.cvt3.grid(row=0, column=6, sticky="w")

    T.Label(d0tf, text="Data", width=7, height=1, background="#ffffff").grid(row=0, column=0, sticky="news")
    for icol in range(self.cols):
      clab = T.Label(d0tf, width=7, height=1)
      clab.configure(relief="raised", borderwidth=1)
      clab.configure(background="#efebe8")
      scol = icol + 1
      clab.configure(text=scol)
      self.clabs.append(clab)
      clab.grid(row=0, column=scol, sticky="nwes")
      clab.bind("<Enter>", self.clabenter)
      clab.bind("<Leave>", self.clableave)
      clab.bind("<Button-1>", self.clabpress)
      clab.bind("<ButtonRelease-1>", self.clabrelease)

    for irow in range(self.rows):
      rlab = T.Label(d0tf, width=7, height=1)
      rlab.configure(relief="raised", borderwidth=1)
      rlab.configure(background="#efebe8")
      srow = irow + 1
      rlab.configure(text=srow)
      self.rlabs.append(rlab)
      rlab.grid(row=srow, column=0, sticky="nwes")
      rlab.bind("<Enter>", self.rlabenter)
      rlab.bind("<Leave>", self.rlableave)
      rlab.bind("<Button-1>", self.rlabpress)
      rlab.bind("<ButtonRelease-1>", self.rlabrelease)

    for irow in range(self.rows):
      onerow = []
      for icol in range(self.cols):
        cell = T.Text(d0tf, width=6, height=1, takefocus=0)
        cell.configure(relief="sunken", borderwidth=1, padx=4, pady=2)
        cell.configure(background="#f8f8f8", foreground="#000000")
        cell.insert("1.0", "")
        cell.configure(state="disabled")
        cell.grid(row=irow+1, column=icol+1, sticky="news", ipadx=3)
        cell.bind("<Button-1>", self.focusthiscell)
        cell.bind("<FocusIn>", self.focusthiscell)

        onerow.append(cell)
      self.cells.append(onerow)
    self.cells[0][0].bind("<FocusOut>", self.focusoutcell)

    for icol in range(self.cols):
      d0tf.columnconfigure(icol, minsize=6, uniform=1)

  def clabenter(self, event):
    wdg = event.widget
    ind = -1
    for clab in self.clabs:
      ind += 1
      if wdg == clab:
        self.clins[ind] = 1

  def clableave(self, event):
    wdg = event.widget
    ind = -1
    for clab in self.clabs:
      ind += 1
      if wdg == clab:
        self.clins[ind] = 0

  def clabpress(self, event):
    wdg = event.widget
    ind = -1
    for clab in self.clabs:
      ind += 1
      if wdg == clab:
        if 0 == self.cactv[ind]:
          return
        wdg.configure(relief="sunken")

  def clabrelease(self, event):
    itype = self.cselrb.get()
    ioffs = self.ctvar.get()

    wdg = event.widget
    ind = -1
    for clab in self.clabs:
      ind += 1
      if wdg == clab:
        if 0 == self.cactv[ind]:
          return
        wdg.configure(relief="raised")
        if 1 == self.clins[ind]:
          irank = ind + ioffs
          oldst = self.clsts[ind]
          newst = 1 - oldst
          if 0 == newst:
            if 1 == itype:
              if irank == self.chead:
                self.chead = -1
                self.clsts[ind] = 0
                wdg.configure(background="#efebe8")
            if 2 == itype:
              if irank == self.cfrom:
                self.cfrom = -1
                self.clsts[ind] = 0
                wdg.configure(background="#efebe8")
                self.csel = False
            if 3 == itype:
              if irank == self.cto:
                self.cto = -1
                self.clsts[ind] = 0
                wdg.configure(background="#efebe8")
                self.csel = False
          else:
            if 1 == itype:
              ican = 1
              if (-1 < self.cfrom) and (-1 < self.cto):
                if (irank >= self.cfrom) and (irank <= self.cto):
                  ican = 0
              if 1 == ican:
                self.clsts[ind] = 1
                wdg.configure(background="#e0ffe0")
                iold = self.chead - ioffs
                if (0 <= iold) and (self.cols > iold):
                  self.clsts[iold] = 0
                  self.clabs[iold].configure(background="#efebe8")
                self.chead = irank
            if 2 == itype:
              ican = 1
              if irank == self.chead:
                ican = 0
              if -1 < self.cto:
                if irank >= self.cto:
                  ican = 0
                if (self.chead < self.cto) and (irank <= self.chead):
                  ican = 0
              if 1 == ican:
                self.clsts[ind] = 1
                wdg.configure(background="#e0ffe0")
                iold = self.cfrom - ioffs
                if (0 <= iold) and (self.cols > iold):
                  self.clsts[iold] = 0
                  self.clabs[iold].configure(background="#efebe8")
                self.cfrom = irank
            if 3 == itype:
              ican = 1
              if irank == self.chead:
                ican = 0
              if -1 < self.cfrom:
                if irank <= self.cfrom:
                  ican = 0
                if (self.chead > self.cfrom) and (irank >= self.chead):
                  ican = 0
              if 1 == ican:
                self.clsts[ind] = 1
                wdg.configure(background="#e0ffe0")
                iold = self.cto - ioffs
                if (0 <= iold) and (self.cols > iold):
                  self.clsts[iold] = 0
                  self.clabs[iold].configure(background="#efebe8")
                self.cto = irank
            if (-1 < self.cfrom) and (-1 < self.cto):
              self.csel = True

          self.updateLabels()
          self.updateColors()

  def rlabenter(self, event):
    wdg = event.widget
    ind = -1
    for rlab in self.rlabs:
      ind += 1
      if wdg == rlab:
        self.rlins[ind] = 1

  def rlableave(self, event):
    wdg = event.widget
    ind = -1
    for rlab in self.rlabs:
      ind += 1
      if wdg == rlab:
        self.rlins[ind] = 0

  def rlabpress(self, event):
    wdg = event.widget
    ind = -1
    for rlab in self.rlabs:
      ind += 1
      if wdg == rlab:
        if 0 == self.ractv[ind]:
          return
        wdg.configure(relief="sunken")

  def rlabrelease(self, event):
    itype = self.rselrb.get()
    ioffs = self.rtvar.get()

    wdg = event.widget
    ind = -1
    for rlab in self.rlabs:
      ind += 1
      if wdg == rlab:
        if 0 == self.ractv[ind]:
          return
        wdg.configure(relief="raised")
        if 1 == self.rlins[ind]:
          irank = ind + ioffs
          oldst = self.rlsts[ind]
          newst = 1 - oldst
          if 0 == newst:
            if 1 == itype:
              if irank == self.rhead:
                self.rhead = -1
                self.rlsts[ind] = 0
                wdg.configure(background="#efebe8")
            if 2 == itype:
              if irank == self.rfrom:
                self.rfrom = -1
                self.rlsts[ind] = 0
                wdg.configure(background="#efebe8")
                self.rsel = False
            if 3 == itype:
              if irank == self.rto:
                self.rto = -1
                self.rlsts[ind] = 0
                wdg.configure(background="#efebe8")
                self.rsel = False
          else:
            if 1 == itype:
              ican = 1
              if (-1 < self.rfrom) and (-1 < self.rto):
                if (irank >= self.rfrom) and (irank <= self.rto):
                  ican = 0
              if 1 == ican:
                self.rlsts[ind] = 1
                wdg.configure(background="#ffe0e0")
                iold = self.rhead - ioffs
                if (0 <= iold) and (self.rows > iold):
                  self.rlsts[iold] = 0
                  self.rlabs[iold].configure(background="#efebe8")
                self.rhead = irank
            if 2 == itype:
              ican = 1
              if irank == self.rhead:
                ican = 0
              if -1 < self.rto:
                if irank >= self.rto:
                  ican = 0
                if (self.rhead < self.rto) and (irank <= self.rhead):
                  ican = 0
              if 1 == ican:
                self.rlsts[ind] = 1
                wdg.configure(background="#ffe0e0")
                iold = self.rfrom - ioffs
                if (0 <= iold) and (self.rows > iold):
                  self.rlsts[iold] = 0
                  self.rlabs[iold].configure(background="#efebe8")
                self.rfrom = irank
            if 3 == itype:
              ican = 1
              if irank == self.rhead:
                ican = 0
              if -1 < self.rfrom:
                if irank <= self.rfrom:
                  ican = 0
                if (self.rhead > self.rfrom) and (irank >= self.rhead):
                  ican = 0
              if 1 == ican:
                self.rlsts[ind] = 1
                wdg.configure(background="#ffe0e0")
                iold = self.rto - ioffs
                if (0 <= iold) and (self.rows > iold):
                  self.rlsts[iold] = 0
                  self.rlabs[iold].configure(background="#efebe8")
                self.rto = irank
            if (-1 < self.rfrom) and (-1 < self.rto):
              self.rsel = True

          self.updateLabels()
          self.updateColors()

  def untabs(self):
    self.tabed = 0
    self.settabs(0)

  def settabs(self, ival):
    for crow in self.cells:
      for cell in crow:
        cell.configure(takefocus=ival)

  def focuscell(self, fcell):
    rind = -1
    for crow in self.cells:
      rind += 1
      cind = -1
      for cell in crow:
        cind += 1
        sfgc = self.sfgcols[self.rfgcols[rind]][self.cfgcols[cind]]
        sbgc = self.sbgcols[self.rbgcols[rind]][self.cbgcols[cind]]
        cell.configure(background=sbgc, foreground=sfgc)
    fcell.configure(background="#ffffff", foreground="#000000")
    fcell.focus()

  def focusoutcell(self, event):
    self.updateColors()

  def focusthiscell(self, event):
    if 0 == self.tabed:
      self.tabed = 1
      self.settabs(1)
    self.focuscell(event.widget)

  def setRange(self, value):

    irow = self.rtvar.get()
    icol = self.ctvar.get()

    for wdg in self.rlabs:
      wdg.configure(background="#efebe8")
    for wdg in self.clabs:
      wdg.configure(background="#efebe8")

    for ind in range(self.rows):
      self.rlsts[ind] = 0
    for ind in range(self.cols):
      self.clsts[ind] = 0

    for irank in [self.rhead, self.rfrom, self.rto]:
      if -1 < irank:
        ipos = irank - irow
        if (-1 < ipos) and (self.rows > ipos):
          self.rlabs[ipos].configure(background="#ffe0e0")
          self.rlsts[ipos] = 1

    for irank in [self.chead, self.cfrom, self.cto]:
      if -1 < irank:
        ipos = irank - icol
        if (-1 < ipos) and (self.cols > ipos):
          self.clabs[ipos].configure(background="#e0ffe0")
          self.clsts[ipos] = 1

    for rind in range(self.rows):
      cvsrind = rind + irow
      if cvsrind < self.rcnt:
        self.ractv[rind] = 1
        self.rlabs[rind].configure(text=str(cvsrind+1))
      else:
        self.ractv[rind] = 0
        self.rlabs[rind].configure(text="")

    for cind in range(self.cols):
      cvscind = cind + icol
      if cvscind < self.ccnt:
        self.cactv[cind] = 1
        self.clabs[cind].configure(text=str(cvscind+1))
      else:
        self.cactv[cind] = 0
        self.clabs[cind].configure(text="")

    for rind in range(self.rows):
      cvsrind = rind + irow
      if cvsrind < self.rcnt:
        cvsrow = self.data[cvsrind]
        for cind in range(self.cols):
          cvscind = cind + icol
          self.cells[rind][cind].configure(state="normal")
          self.cells[rind][cind].delete("1.0", "end")
          if cvscind < len(cvsrow):
            value = cvsrow[cvscind]
            self.cells[rind][cind].insert("1.0", str(value))
          self.cells[rind][cind].configure(state="disabled")
      else:
        for cind in range(self.cols):
          self.cells[rind][cind].configure(state="normal")
          self.cells[rind][cind].delete("1.0", "end")
          self.cells[rind][cind].configure(state="disabled")

    self.updateColors()

  def setData(self, cvsdata):
    self.data = cvsdata
    self.rcnt = len(cvsdata)
    self.ccnt = 0
    for rind in range(self.rcnt):
      curlen = len(self.data[rind])
      if self.ccnt < curlen:
        self.ccnt = curlen

    self.rtvar.set(0)
    self.ctvar.set(0)

    self.rsel = False
    self.csel = False
    self.rfrom = -1
    self.rto = -1
    self.rhead = -1
    self.cfrom = -1
    self.cto = -1
    self.chead = -1

    for ind in range(self.rows):
      self.ractv[ind] = 0
    for ind in range(self.cols):
      self.cactv[ind] = 0

    for rlab in self.rlabs:
      rlab.configure(background="#efebe8")

    for clab in self.clabs:
      clab.configure(background="#efebe8")

    rlast = self.rcnt % self.rres
    if 0 == rlast:
      rlast = self.rows
    rtoval = self.rcnt - rlast
    if 0 > rtoval:
      rtoval = 0
    self.rds.configure(to=rtoval)

    clast = self.ccnt % self.cres
    if 0 == clast:
      clast = self.cols
    ctoval = self.ccnt - clast
    if 0 > ctoval:
      ctoval = 0
    self.cds.configure(to=ctoval)

    self.setRange(0)
    self.updateLabels()

  def updateColors(self):
    irow = self.rtvar.get()
    icol = self.ctvar.get()

    irhead = self.rhead - irow
    irfrom = self.rfrom - irow
    irto = self.rto - irow
    ichead = self.chead - icol
    icfrom = self.cfrom - icol
    icto = self.cto - icol

    for ind in range(self.rows):
      self.rfgcols[ind] = 0
      self.rbgcols[ind] = 0
    for ind in range(self.cols):
      self.cfgcols[ind] = 0
      self.cbgcols[ind] = 0

    irmax = self.rows - 1
    icmax = self.cols - 1

    if ((0 <= irhead) and (irmax >= irhead)):
      self.rfgcols[irhead] = 1

    if ((0 <= ichead) and (icmax >= ichead)):
      self.cfgcols[ichead] = 1

    if self.rsel:
      if ((0 >= irfrom) and (irmax <= irto)):
        for ind in range(self.rows):
          self.rbgcols[ind] = 1
      else:
        if ((0 <= irfrom) and (irmax >= irto)):
          for ind in range(irfrom, (irto+1)):
            self.rbgcols[ind] = 1
        if ((0 > irfrom) and (irmax > irto)):
          for ind in range(irto+1):
            self.rbgcols[ind] = 1
        if ((0 < irfrom) and (irmax < irto)):
          for ind in range(irfrom, self.rows):
            self.rbgcols[ind] = 1

    if self.csel:
      if ((0 >= icfrom) and (icmax <= icto)):
        for ind in range(self.cols):
          self.cbgcols[ind] = 1
      else:
        if ((0 <= icfrom) and (icmax >= icto)):
          for ind in range(icfrom, (icto+1)):
            self.cbgcols[ind] = 1
        if ((0 > icfrom) and (icmax > icto)):
          for ind in range(icto+1):
            self.cbgcols[ind] = 1
        if ((0 < icfrom) and (icmax < icto)):
          for ind in range(icfrom, self.cols):
            self.cbgcols[ind] = 1

    for rind in range(self.rows):
      for cind in range(self.cols):
        sfgc = self.sfgcols[self.rfgcols[rind]][self.cfgcols[cind]]
        sbgc = self.sbgcols[self.rbgcols[rind]][self.cbgcols[cind]]
        self.cells[rind][cind].configure(background=sbgc, foreground=sfgc)

  def updateLabels(self):
    curval = "---"
    if -1 < self.rhead:
      curval = str(self.rhead + 1)
    curtext = curval + "\n "
    self.rvt1.configure(text=curtext)

    curval = "---"
    if -1 < self.rfrom:
      curval = str(self.rfrom + 1)
    curtext = curval + "\n "
    self.rvt2.configure(text=curtext)

    curval = "---"
    if -1 < self.rto:
      curval = str(self.rto + 1)
    curtext = curval + "\n "
    self.rvt3.configure(text=curtext)

    curval = "---"
    if -1 < self.chead:
      curval = str(self.chead + 1)
    curtext = curval + ""
    self.cvt1.configure(text=curtext)

    curval = "---"
    if -1 < self.cfrom:
      curval = str(self.cfrom + 1)
    curtext = curval + ""
    self.cvt2.configure(text=curtext)

    curval = "---"
    if -1 < self.cto:
      curval = str(self.cto + 1)
    curtext = curval + ""
    self.cvt3.configure(text=curtext)

    enablegui = False
    if self.rsel and self.csel:
      enablegui = True
      if 2 > (self.rto - self.rfrom):
        enablegui = False
      if 2 > (self.cto - self.cfrom):
        enablegui = False

    for wdg in self.outergui:
      if enablegui:
        wdg.configure(state="normal")
      else:
        wdg.configure(state="disabled")

  def setRScaleFocus(self, e):
    self.rds.focus()
    self.updateColors()

  def setCScaleFocus(self, e):
    self.cds.focus()
    self.updateColors()


class TreeSelect(object):
  def __init__(self, paramvals):
    self.view = []
    self.tree = []
    self.filled = False
    self.vertical = paramvals["vertical"]
    self.errmsg = ""

    self.datadir = ""
    self.datafile = ""

  def setclgui(self, lvar, gpro):
    self.view.outervar.append(lvar)
    self.view.outergui.append(gpro)

  def show(self, frame):
    self.view = AxeTree(frame)
    self.view.printCanvas = self.printCanvas
    self.view.printCanvas2 = self.printCanvas2
    self.view.vertical = self.vertical

  def empty(self):
    self.view.delTree()
    self.filled = False
    self.errmsg = ""
    self.tree = []

  def fill(self, clustered, groups):

    self.filled = False
    self.errmsg = ""
    tnodes = []
    lcount = 0
    ecount = 0
    gsizes = []
    if clustered:
      cenlist = groups.cenranks
      lcount = len(cenlist)
      if 3 > lcount:
        self.errmsg = "too few groups"
        if 0 == lcount:
          self.errmsg = " no group "
        return
      tcount = (2 * lcount) - 1
      tnodes = N.empty((tcount, groups.mcols))
      ecount = groups.mcols
      gsizes = N.empty(tcount, dtype=int)
      for ind in range(lcount):
        icen = cenlist[ind][0]
        tnodes[ind] = groups.centers[icen]
        gsizes[ind] = groups.censize[icen]
    else:
      lcount = groups.mrows
      ecount = groups.mcols
      tcount = (2 * lcount) - 1
      tnodes = N.empty((tcount,ecount))
      tnodes[:lcount,:] = groups.pmatrix
      for irow in range(lcount):
        for icol in range(ecount):
          cval = tnodes[irow][icol]
          isf = N.isfinite(cval)
          if isf is NotImplemented:
            tnodes[irow][icol] = 0.0
          elif not isf:
            tnodes[irow][icol] = 0.0

      gsizes = N.empty(tcount, dtype=int)
      gsizes[:] = 1

    self.tree = TreeData(tnodes, lcount, ecount, gsizes)
    self.tree.createTree()
    self.view.setTree(self.tree)

    self.filled = True

  def renewTree(self, tnodes, lcount, ecount, gsizes, treestruc, optview=True):
    self.filled = False

    self.tree = TreeData(tnodes, lcount, ecount, gsizes)
    self.tree.optview = optview

    rv = self.tree.renewTree(treestruc)
    if rv:
      self.view.setTree(self.tree)
      self.filled = True

  def emptyTreeSel(self):
    self.view.releaseAll()

  def bottomTreeSel(self):
    self.view.selectBottom()

  def selectOne(self, igrp):
    self.view.selectOne(igrp)

  def getSelections(self):
    sels = []
    if not [] == self.view:
      sels = self.view.getSelections()
    return sels

  def getImagePS(self, wholecanvas=True, withlabels=True):
    if [] == self.view:
      return ""
    if [] == self.view.canv:
      return ""

    vertcm = "10.0c"
    if not withlabels:
      vertcm = "9.0c"

    streps = ""
    if wholecanvas:
      streps = self.view.canv.postscript(width=self.view.pswidth, height=vertcm,
        x="0.0c", y="0.0c", pagex="0.0c", pagey=vertcm, pageanchor="nw", fontmap="fontmap")
    else:
      streps = self.view.canv.postscript(height=vertcm, y="0.0c", pageanchor="nw",
        pagey=vertcm, fontmap="fontmap")

    pos1 = streps.find("%%For:")
    pos1e = streps[pos1:].find("\n")
    if -1 < pos1:
      if -1 < pos1e:
        strtmp = streps[:pos1]
        strtmp += "%%For: CluDisc"
        if "\r" == streps[pos1+pos1e-1]:
          strtmp += "\r"
        strtmp += streps[(pos1+pos1e):]
        streps = strtmp

    pos2 = streps.find("%%Title:")
    pos2e = streps[pos2:].find("\n")
    if -1 < pos2:
      if -1 < pos2e:
        strtmp = streps[:pos2]
        if wholecanvas:
          strtmp += "%%Title: CluDisc Cluster-Tree"
        else:
          strtmp += "%%Title: CluDisc Cluster-Tree Window"
        if "\r" == streps[pos2+pos2e-1]:
          strtmp += "\r"
        strtmp += streps[(pos2+pos2e):]
        streps = strtmp

    return streps

  def getImagePS2(self, x0bor, y0bor, x1bor, y1bor, wholecanvas=False):
    if [] == self.view:
      return ""
    if [] == self.view.canv2:
      return ""

    x0bor = int(x0bor)
    x1bor = int(x1bor)
    y0bor = int(y0bor)
    y1bor = int(y1bor)

    widthmm = int(x1bor - x0bor)
    heightmm = int(y1bor - y0bor)
    widthstr = str(widthmm // 10) + "." + str(widthmm % 10) + "c"
    heightstr = str(heightmm // 10) + "." + str(heightmm % 10) + "c"

    anchx = str(x0bor // 10) + "." + str(x0bor % 10) + "c"
    anchy = str(y1bor // 10) + "." + str(y1bor % 10) + "c"

    posx = anchx
    posy = str(y0bor // 10) + "." + str(y0bor % 10) + "c"

    streps = self.view.canv2.postscript(width=widthstr, height=heightstr,
      x=posx, y=posy, pagex=anchx, pagey=anchy, pageanchor="nw", fontmap="fontmap")

    pos1 = streps.find("%%For:")
    pos1e = streps[pos1:].find("\n")
    if -1 < pos1:
      if -1 < pos1e:
        strtmp = streps[:pos1]
        strtmp += "%%For: CluDisc"
        if "\r" == streps[pos1+pos1e-1]:
          strtmp += "\r"
        strtmp += streps[(pos1+pos1e):]
        streps = strtmp

    pos2 = streps.find("%%Title:")
    pos2e = streps[pos2:].find("\n")
    if -1 < pos2:
      if -1 < pos2e:
        strtmp = streps[:pos2]
        if wholecanvas:
          strtmp += "%%Title: CluDisc Cluster-Tree"
        else:
          strtmp += "%%Title: CluDisc Cluster-SubTree"
        if "\r" == streps[pos2+pos2e-1]:
          strtmp += "\r"
        strtmp += streps[(pos2+pos2e):]
        streps = strtmp

    return streps

  def printCanvas(self, wholecanvas, withlabels):
    tstr = "cluster tree image window"
    if wholecanvas:
      tstr = "cluster tree image"
    fn = F.asksaveasfilename(title=tstr, filetypes=[("postscript files", "*.eps"), ("all files", "*")], initialdir=self.datadir, confirmoverwrite=False)
    if not fn:
      return [False, None]
    dn = os.path.dirname(fn)
    if not os.path.isdir(dn):
      return [False, ""]

    self.datafile = fn
    self.datadir = dn

    fbase = self.datafile
    for suff, slen in [[".ps", 3], [".eps", 4], [".pdf", 4]]:
      if fbase.endswith(suff):
        fbase = fbase[:(-1 * slen)]
        break

    fbase0 = os.path.basename(fbase)
    if 0 == len(fbase0):
      return  [False, ""]

    fnimg = fbase + ".eps"
    fwname = fbase0 + ".eps"

    isav = True
    if os.path.exists(fnimg):
      isav = False

    curind = 0
    while not isav:
      isav = True
      curind += 1
      fnimg = fbase + "-" + str(curind) + ".eps"
      fwname = fbase0 + "-" + str(curind) + ".eps"
      if os.path.exists(fnimg):
        isav = False

    try:
      pstr = self.getImagePS(wholecanvas, withlabels)
      hps = open(fnimg, "w")
      hps.write(pstr)
      hps.close()

    except:
      return  [False, ""]

    return [True, fwname]

  def printCanvas2(self, x0bor, y0bor, x1bor, y1bor, wholecanvas=False):
    tstr = "cluster sub-tree image"
    fn = F.asksaveasfilename(title=tstr, filetypes=[("postscript files", "*.eps"), ("all files", "*")], initialdir=self.datadir, confirmoverwrite=False)
    if not fn:
      return [False, None]
    dn = os.path.dirname(fn)
    if not os.path.isdir(dn):
      return [False, ""]

    self.datafile = fn
    self.datadir = dn

    fbase = self.datafile
    for suff, slen in [[".ps", 3], [".eps", 4], [".pdf", 4]]:
      if fbase.endswith(suff):
        fbase = fbase[:(-1 * slen)]
        break

    fbase0 = os.path.basename(fbase)
    if 0 == len(fbase0):
      return  [False, ""]

    fnimg = fbase + ".eps"
    fwname = fbase0 + ".eps"

    isav = True
    if os.path.exists(fnimg):
      isav = False

    curind = 0
    while not isav:
      isav = True
      curind += 1
      fnimg = fbase + "-" + str(curind) + ".eps"
      fwname = fbase0 + "-" + str(curind) + ".eps"
      if os.path.exists(fnimg):
        isav = False

    try:
      pstr = self.getImagePS2(x0bor, y0bor, x1bor, y1bor, wholecanvas)
      hps = open(fnimg, "w")
      hps.write(pstr)
      hps.close()

    except:
      return  [False, ""]

    return [True, fwname]

class StorageData(object):
  def __init__(self):
    self.csvdata = []
    self.delimiter = ""
    self.oname = ""
    self.fname = ""
    self.fsize = ""
    self.hexdig = ""

    self.csvcols = 0
    self.csvrows = 0

    self.dclog = 0
    self.dcwin = 0
    self.dcnor = 0
    self.props = 0
    self.basec = 0
    self.infrows = [-1, -1, -1]
    self.infcols = [-1, -1, -1]
    self.winstype = 0
    self.winsvalue = 0

    self.treeid = ""
    self.clust = 0

    self.dcnames = []
    self.prnames = []

    self.mydt = []

  def show(self, frame):
    self.mydt = DataTable(frame, 10, 10)
    self.mydt.setData(self.csvdata)

  def addclgui(self, wdg):
    self.mydt.outergui.append(wdg)

  def setData(self, csvdata):
    self.csvdata = csvdata
    if not [] == self.mydt:
      self.mydt.setData(self.csvdata)

  def clean(self):
    self.csvdata = []
    self.delimiter = ""
    self.oname = ""
    self.fname = ""
    self.fsize = ""
    self.hexdig = ""

    self.csvcols = 0
    self.csvrows = 0

    self.dclog = 0
    self.dcwin = 0
    self.dcnor = 0
    self.props = 0
    self.basec = 0
    self.infrows = [-1, -1, -1]
    self.infcols = [-1, -1, -1]
    self.winstype = 0
    self.winsvalue = 0

    self.treeid = ""
    self.clust = 0

    self.dcnames = []
    self.prnames = []

class ClusterCenters(object):
  def __init__(self, paramvals, isclibcd):
    self.cspeed = isclibcd
    self.error = 0

    self.cencount = 100
    self.maxsteps = paramvals["kmcycmax"]
    self.winstype = paramvals["winstype"]
    self.winsvalue = paramvals["winsvalue"]

    self.tolog = True
    self.towin = True
    self.dcnor = True
    self.dcrob = True

    self.mrows = 0
    self.mcols = 0

    self.pmatrix = []
    self.prepare()

    self.stat = []
    self.root = []

  def clean(self):
    self.cencount = 100

    self.tolog = True
    self.towin = True
    self.dcnor = True
    self.dcrob = True

    self.mrows = 0
    self.mcols = 0

    self.pmatrix = []
    self.prepare()

  def prepare(self):
    self.propmeans = []
    self.propvarts = []

    self.inicenters = []
    self.centers = []
    self.cenused = []

    self.orients = []
    self.cengrps = []
    self.grprops = []

    self.censize = []
    self.cenvoid = []
    self.cenlive = []

    self.cranks = []

    self.kmruncycles = 0
    self.kmconverged = 0

  def release(self):
    self.pmatrix = []
    self.centers = []
    self.cenused = []

    self.censize = []
    self.cenvoid = []
    self.cenlive = []

    self.cranks = []

  def normalize(self):
    self.root.update()
    self.prepare()
    self.error = 0

    pinf = N.inf
    minf = -1 * pinf
    for icol in range(self.mcols):
      curcol = self.pmatrix[:,icol]
      curfin = N.isfinite(curcol)
      curlen = len(curfin[True == curfin])
      curmax = 1
      curmin = -1
      if 0 < curlen:
        curmax = N.max(curcol[curfin])
        curmin = N.min(curcol[curfin])
      curcol[pinf == curcol] = curmax
      curcol[minf == curcol] = curmin

    self.root.update()
    #to winsorize data cases
    winmin = 0.0
    winmax = 0.0

    if self.towin:
      for icol in range(self.mcols):
        curcol = self.pmatrix[:,icol]
        curfin = N.isfinite(curcol)
        curlen = len(curfin[True == curfin])
        cntwin = 0
        cntwinmax = (curlen - 2) // 2
        if "q" == self.winstype:
          cntwin = int(N.sqrt(curlen / self.winsvalue))
        else:
          cntwin = int((curlen * self.winsvalue) // 200)
        if cntwinmax < cntwin:
          cntwin = cntwinmax
        if 0 < cntwin:
          curord = sorted(curcol[curfin])
          winmin = curord[cntwin]
          winmax = curord[curlen-(cntwin+1)]
          curcol[winmin > curcol] = winmin
          curcol[winmax < curcol] = winmax

    self.root.update()
    #into logarithmic values
    if self.tolog:
      if 0 < len(self.pmatrix[0 >= self.pmatrix]):
        self.error = 1
        return

      for irow in range(self.mrows):
        for icol in range(self.mcols):
          cval = self.pmatrix[irow][icol]
          if N.isfinite(cval):
            cval = N.log2(cval)
          self.pmatrix[irow][icol] = cval

    #data cases normalization
    renorstd = False
    renorrob = False
    if self.dcnor:
      if self.dcrob:
        renorrob = True
      else:
        renorstd = True

    self.root.update()
    if renorstd:
      for icol in range(self.mcols):
        curcol = self.pmatrix[:,icol]
        curfin = N.isfinite(curcol)
        curlen = len(curfin[True == curfin])
        if 1 < curlen:
          cmean = N.mean(curcol[curfin])
          curcol -= cmean
          cvart = N.var(curcol[curfin])
          if 0 < cvart:
            curcol /= N.sqrt(cvart)
          else:
            curcol[curfin] = N.nan
        else:
          curcol[curfin] = N.nan

    self.root.update()
    if renorrob:
      for icol in range(self.mcols):
        curcol = self.pmatrix[:,icol]
        curfin = N.isfinite(curcol)
        curlen = len(curfin[True == curfin])
        if 1 < curlen:
          cmedian = N.median(curcol[curfin])
          curcol -= cmedian
          cabsd = N.median(N.abs(curcol[curfin]))
          if 0 < cabsd:
            #1/qnorm(0.75) ~= 1.4826
            curcol /= (cabsd * 1.4826)
          else:
            curcol[curfin] = N.nan
        else:
          curcol[curfin] = N.nan

    self.root.update()
    #data properties normalization
    self.propmeans = N.empty(self.mrows)
    self.propvarts = N.empty(self.mrows)
    for irow in range(self.mrows):
      currow = self.pmatrix[irow]
      curfin = N.isfinite(currow)
      curlen = len(curfin[True == curfin])
      cmean = 0.0
      cvart = 0.0
      if 0 < curlen:
        cmean = N.mean(currow[curfin])
        currow -= cmean
        cvart = N.var(currow[curfin])
        if 0 < cvart:
          currow /= N.sqrt(cvart)
        else:
          currow[curfin] = N.nan
      else:
        currow[curfin] = N.nan
      self.propvarts[irow] = cvart
      self.propmeans[irow] = cmean

  def kmeans(self):
    self.selIniCens()
    self.root.update()

    self.censize = N.zeros(self.cencount, dtype=int)
    self.cenvoid = N.zeros(self.cencount, dtype=int)

    self.centers = N.empty((self.cencount, self.mcols))
    self.cenused = N.empty((self.cencount, self.mcols))
    self.cenlive = N.empty(self.cencount, dtype=C.c_long)

    for ind in range(self.cencount):
      irank = self.inicenters[ind]
      self.centers[ind] = self.pmatrix[irank]
      self.cenlive[ind] = 1

    self.cranks = N.empty(self.mrows, dtype=C.c_long)
    self.orients = N.zeros(self.mrows, dtype=C.c_long)

    if self.cspeed:
      self.pmatrix = N.ascontiguousarray(self.pmatrix)
      self.centers = N.ascontiguousarray(self.centers)
      self.cenlive = N.ascontiguousarray(self.cenlive)
      self.cranks = N.ascontiguousarray(self.cranks)
      self.orients = N.ascontiguousarray(self.orients)

    for irow in range(self.mrows):
      self.cranks[irow] = -1

    changed = True
    self.kmruncycles = 0
    self.kmconverged = 0
    for istep in range(self.maxsteps):
      cycrun = str(istep+1) + " / " + str(self.maxsteps)
      self.stat.configure(text=" \ncycle: " + cycrun + "\n ")
      self.root.update()
      changed = False
      auxres = N.empty(1)
      cdist = 0.0

      if self.cspeed:
        changed = cKMSearch(self.mrows, self.cencount, self.mcols, self.pmatrix,
          self.centers, self.cenlive, self.cranks, self.orients)
        self.root.update()

      else:
        for irow in range(self.mrows):
          self.root.update()

          ccen = self.cranks[irow]
          newldist = N.inf
          orient = 0
          if -1 != ccen:
            orient, newldist = self.calcKMDist(ccen, irow)
            self.orients[irow] = orient
            if 0 == orient:
              ccen = -1
              newldist = N.inf
              self.cranks[irow] = -1
              changed = True

          for icen in range(self.cencount):
            if not self.cenlive[icen]:
              continue
            if ccen == icen:
              continue

            orient, cdist = self.calcKMDist(icen, irow)

            if cdist < newldist:
              newldist = cdist
              self.cranks[irow] = icen
              self.orients[irow] = orient
              changed = True

      #empty group sizes
      self.censize[:] = 0
      #increase resp. grp sizes
      for irow in range(self.mrows):
        crank = self.cranks[irow]
        if -1 < crank:
          self.censize[crank] += 1

      #empty group centers coordinates
      for icen in range(self.cencount):
        if self.cenlive[icen]:
          if 0 != self.censize[icen]:
            self.centers[icen,:] = 0.0
            self.cenused[icen,:] = 0

      #adds coordinates into resp. centers
      for irow in range(self.mrows):
        crank = self.cranks[irow]
        cornt = self.orients[irow]
        if -1 < crank:
          curfin = N.isfinite(self.pmatrix[irow])
          if 1 == cornt:
            self.centers[crank][curfin] += self.pmatrix[irow][curfin]
            self.cenused[crank][curfin] += 1
          elif -1 == cornt:
            self.centers[crank][curfin] -= self.pmatrix[irow][curfin]
            self.cenused[crank][curfin] += 1

      #divide coord sums if not empty
      #assert not alive grp if empty for the second step
      for icen in range(self.cencount):
        if self.cenlive[icen]:
          csize = self.censize[0]
          if 0 < csize:
            self.cenvoid[icen] = 0
            for icol in range(self.mcols):
              cused = self.cenused[icen,icol]
              if 0 < cused:
                self.centers[icen,icol] /= cused
              else:
                self.centers[icen,icol] /= N.nan
          else:
            self.cenvoid[icen] += 1
            if 2 <= self.cenvoid[icen]:
              self.cenlive = False

      self.kmruncycles = istep + 1
      if not changed:
        self.kmconverged = 1
        break

    #to compute center variations
    self.cenranks = []
    for icen in range(self.cencount):
      cursize = self.censize[icen]
      if 0 == cursize:
        self.cenlive[icen] = 0

      if self.cenlive[icen]:
        for icol in range(self.mcols):
          cval = self.centers[icen][icol]
          isf = N.isfinite(cval)
          if isf is NotImplemented:
            self.centers[icen][icol] = 0.0
          elif not isf:
            self.centers[icen][icol] = 0.0

        curscale = N.var(self.centers[icen])
        auxmax = N.sqrt(cursize)
        curscale *= auxmax
        if 1 < cursize:
          curscale = (curscale - 1) / (auxmax - 1)
        self.cenranks.append([icen, curscale])

    self.cenranks.sort(key=lambda x: x[1], reverse=True)
    cenactl = len(self.cenranks)

    self.grprops = N.empty(self.mrows, dtype=int)
    self.grprops[:] = -1

    self.centrans = self.cencount * [-1]
    for inew in range(cenactl):
      iold = self.cenranks[inew]
      self.centrans[iold[0]] = inew

    self.prepCenGroups(cenactl)

    for irow in range(self.mrows):
      oldcen = self.cranks[irow]
      if -1 < oldcen:
        newcen = self.centrans[oldcen]
        self.cengrps[newcen].append(irow)
        self.grprops[irow] = newcen

  def selIniCens(self):
    self.inicenters = []

    if 0 < self.mrows:
      inicenters = list(range(self.mrows))
      random.shuffle(inicenters)
      self.inicenters = inicenters[:self.cencount]

  def prepCenGroups(self, count):
    if 0 >= count:
      self.cengrps = []
      return

    self.cengrps = count * [[],]
    for icen in range(count):
      self.cengrps[icen] = []

  def calcKMDist(self, icen, irow):
    common = 0
    cpdist = 0.0
    cndist = 0.0

    for icol in range(self.mcols):
      vrow = self.pmatrix[irow][icol]
      vcen = self.centers[icen][icol]
      if N.isfinite(vrow) and N.isfinite(vcen):
        common += 1
        cpdist += N.square(vrow - vcen)
        cndist += N.square(vrow + vcen)

    rv = [0, N.nan]
    if 0 < common:
      respos = N.sqrt(cpdist / common)
      resneg = N.sqrt(cndist / common)
      if resneg < respos:
        rv = [-1, resneg]
      else:
        rv = [1, respos]
    return rv

class CluDiscApp(object):
  """ The main and GUI part of CluDisc. """

  def __init__(self, p_sto, p_cls, p_cen, paramvals, isclibcd):
    """ to set the data, clustering and tree objects """
    self.sto = p_sto
    self.cls = p_cls
    self.cen = p_cen
    self.params = paramvals
    self.isclibcd = isclibcd

    self.rterror = 0

    self.atenstart = paramvals["atenstart"]
    self.atenstay = paramvals["atenstay"]
    self.atencount = paramvals["atencount"]
    self.resnorm = paramvals["resnorm"]
    self.basecount = paramvals["basecount"]
    self.project = paramvals["project"]
    self.vertical = paramvals["vertical"]
    self.extend = paramvals["extend"]

    self.alg = []
    self.selList = []
    self.works = False
    self.saved = False
    self.sname = ""
    self.disposed = True

    self.ctdftype = "1"
    self.ctdversion = "1.1"
    self.selftype = "2"
    self.selversion = "1.1"
    self.colftype = "3"
    self.colversion = "1.0"
    self.psdftype = "4"
    self.psdversion = "1.1"

    self.datadir = paramvals["inidatadir"]
    self.cls.datadir = self.datadir

    self.root = T.Tk()
    self.root.withdraw()
    try:
      self.app_icon = T.PhotoImage(data=ICON_APPLICATION)
    except:
      self.app_icon = None
    try:
      self.root.tk.call("wm", "iconphoto", self.root._w, "-default", self.app_icon)
    except:
      pass

    try:
      fontstd = "Roman 9"
      self.root.tk.call('set', 'fontmap(%s)' % fontstd, 'Arial 13')
    except:
      log_err("no suitable font available")

    self.withtile = True
    tclpath = str(paramvals["tclpath"])
    if 0 < len(tclpath):
      try:
        self.root.tk.eval('lappend auto_path "' + tclpath + '"')
      except:
        self.withtile = False

    try:
      self.root.tk.eval('package require tile')
    except:
      self.withtile = False

    self.fig = None
    self.collection_comment_placeholder = "put in collection comments"

  def prepFile(self):
    self.f2file = T.Frame(self.fbot)

    T.Label(self.f2file, text=" ", height=2).pack(side="top", expand=0, fill="none")
    fd = T.Frame(self.f2file)
    fd.pack(side="top", expand=0, fill="none")
    fi = T.Frame(self.f2file)
    fi.pack(side="top", expand=1, fill="y")
    fe = T.Frame(self.f2file)
    fe.pack(side="top", expand=1, fill="y")

    T.Label(fd, text="Load data table  ").grid(row=0, column=0, sticky="e")
    T.Label(fd, text="  csv separator:  ", height=1).grid(row=1, column=0, sticky="we", columnspan=2)

    self.vsep = T.IntVar()
    self.vsep.set(0)
    fdsep = T.Frame(fd)
    fdsep.grid(row=1, column=3, columnspan=3)
    T.Radiobutton(fdsep, text="guess  ", variable=self.vsep, value=0, bd=1).pack(side="left", anchor="w")
    T.Radiobutton(fdsep, text="semicolon  ", variable=self.vsep, value=1, bd=1).pack(side="left", anchor="w")
    T.Radiobutton(fdsep, text="comma  ", variable=self.vsep, value=2, bd=1).pack(side="left", anchor="w")
    T.Radiobutton(fdsep, text="tabulator  ", variable=self.vsep, value=3, bd=1).pack(side="left", anchor="w")

    lrgF = ("courier", 16, "normal")
    T.Label(fd, text=" ", font=lrgF).grid(row=2, column=1)
    T.Label(fd, text=" ", font=lrgF).grid(row=7, column=1)

    self.fbutt11 = T.Button(fd, text=" browse ", command=self.loadMatrix, bd=1)
    self.fbutt11.grid(row=0, column=1, sticky="sn")
    self.fbutt11.configure(state="normal")

    T.Label(fd, text="Load cluster-tree  ").grid(row=3, column=0, sticky="e")
    T.Label(fd, text="Save cluster-tree  ").grid(row=4, column=0, sticky="e")
    T.Label(fd, text="Load selections  ").grid(row=5, column=0, sticky="e")
    T.Label(fd, text="Save selections  ").grid(row=6, column=0, sticky="e")
    T.Label(fd, text="Load collections  ").grid(row=8, column=0, sticky="e")
    T.Label(fd, text="Save collections  ").grid(row=9, column=0, sticky="e")
    T.Label(fd, text="Save scaled data  ").grid(row=10, column=0, sticky="e")

    self.fbutt21 = T.Button(fd, text=" browse ", command=self.loadClusters, bd=1)
    self.fbutt22 = T.Button(fd, text=" browse ", command=self.saveClusters, bd=1)
    self.fbutt23 = T.Button(fd, text=" browse ", command=self.loadSelections, bd=1)
    self.fbutt24 = T.Button(fd, text=" browse ", command=self.saveSelections, bd=1)
    self.fbutt41 = T.Button(fd, text=" browse ", command=self.loadCollections, bd=1)
    self.fbutt42 = T.Button(fd, text=" browse ", command=self.saveCollections, bd=1)
    self.fbutt31 = T.Button(fd, text=" browse ", command=self.saveProjections, bd=1)

    self.fbutt21.grid(row=3, column=1, sticky="sn")
    self.fbutt22.grid(row=4, column=1, sticky="sn")
    self.fbutt23.grid(row=5, column=1, sticky="sn")
    self.fbutt24.grid(row=6, column=1, sticky="sn")
    self.fbutt41.grid(row=8, column=1, sticky="sn")
    self.fbutt42.grid(row=9, column=1, sticky="sn")
    self.fbutt31.grid(row=10, column=1, sticky="sn")

    self.fbutt21.configure(state="normal")
    self.fbutt22.configure(state="disabled")
    self.fbutt23.configure(state="disabled")
    self.fbutt24.configure(state="disabled")
    self.fbutt42.configure(state="disabled")
    self.fbutt31.configure(state="disabled")

    T.Label(fd, text=" ", width=10).grid(row=0, column=2)

    itlF = ("courier", 9, "normal italic")
    T.Label(fd, text="csv data matrix to be processed", font=itlF).grid(row=0, column=3, sticky="w")

    T.Label(fd, text=" ", height=2).grid(row=0, column=2, rowspan=1, pady=0)
    T.Label(fd, text=" ", height=2).grid(row=3, column=2, rowspan=1, pady=0)
    T.Label(fd, text=" ", height=2).grid(row=4, column=2, rowspan=1, pady=0)
    T.Label(fd, text=" ", height=2).grid(row=5, column=2, rowspan=1, pady=0)
    T.Label(fd, text=" ", height=2).grid(row=6, column=2, rowspan=1, pady=0)
    T.Label(fd, text=" ", height=2).grid(row=8, column=2, rowspan=1, pady=0)
    T.Label(fd, text=" ", height=2).grid(row=9, column=2, rowspan=1, pady=0)
    T.Label(fd, text=" ", height=2).grid(row=10, column=2, rowspan=1, pady=0)

    T.Label(fd, text="hierarchical picking of clusters", font=itlF, anchor="s").grid(row=3, column=3, sticky="ws", pady=0)
    T.Label(fd, text=" - to be used as axes for scales", font=itlF, anchor="n").grid(row=4, column=3, sticky="wn", pady=0)
    T.Label(fd, text="tree nodes being marked to use", font=itlF, anchor="s").grid(row=5, column=3, sticky="ws", pady=0)
    T.Label(fd, text=" - lightweight cluster-tree add-on", font=itlF, anchor="n").grid(row=6, column=3, sticky="wn", pady=0)

    T.Label(fd, text="case groups to 3D visualize", font=itlF, anchor="s").grid(row=8, column=3, sticky="ws", pady=0)
    T.Label(fd, text=" - definitions of display set", font=itlF, anchor="n").grid(row=9, column=3, sticky="wn", pady=0)
    T.Label(fd, text="final scaled data to visualize", font=itlF).grid(row=10, column=3, sticky="w", pady=0)
    T.Label(fd, text=" ", height=2).grid(row=11, column=0, sticky="nwe", columnspan=2)

    self.vspd = T.IntVar()
    self.vspd.set(0)
    self.vscc = T.IntVar()
    self.vscc.set(0)
    self.vscd = T.IntVar()
    self.vscd.set(0)
    self.vscd_saved = self.vscd.get()

    fdpro = T.Frame(fd)
    fdpro.grid(row=11, column=3, columnspan=3, sticky="n")
    T.Label(fdpro, text="3D: ").pack(side="left", anchor="w")
    self.cbpd = T.Checkbutton(fdpro, text=" linear ", width=8, anchor="w", variable=self.vspd, command=self.updateFileColls, bd=1)
    self.cbpd.pack(side="left", anchor="w")
    self.cbcc = T.Checkbutton(fdpro, text="along collections ", variable=self.vscc, command=self.emptyFileInfo, bd=1)
    self.cbcc.pack(side="left", anchor="w")
    self.cbcd = T.Checkbutton(fdpro, text="with axes ", variable=self.vscd, command=self.updateFileCollsAxes, bd=1)
    self.cbcd.pack(side="left", anchor="w")

    self.labelactfg = self.cbpd.cget("activeforeground")
    self.labeldisfg = self.cbpd.cget("disabledforeground")

    self.cbpd.configure(state="disabled")
    self.cbcc.configure(state="disabled")
    self.cbcd.configure(state="disabled")

    T.Label(fi, text=" ", height=1).grid(row=0, column=0)
    self.finfo = T.Label(fi, text=" ", height=1, width=40)
    self.finfo.grid(row=1, column=0)
    self.freas = T.Label(fi, text=" ", height=1, width=60, font=itlF)
    self.freas.grid(row=2, column=0)
    T.Label(fi, text=" ", height=2).grid(row=3, column=0)

    qb1 = T.Button(fe, text=" Quit ", command=self.viewQuit, bd=1)
    qb1.pack(side="top")

  def viewFile(self):
    if self.works:
      return

    if self.fcurr == self.f2file:
      return

    if 0 == self.rterror:
      self.updateFileCollsView()

    self.finfo.configure(text=" ")
    self.freas.configure(text=" ")

    self.fcurr.pack_forget()
    self.fcurr = self.f2file
    self.fcurr.pack(side="top", anchor="n", expand=1, fill="y")

    self.f1file.configure(relief="sunken")
    self.f1table.configure(relief="raised")
    self.f1clust.configure(relief="raised")
    self.f1state.configure(relief="raised")
    self.f1disp.configure(relief="raised")
    self.f1help.configure(relief="raised")

  def updateFileColls(self):
    self.emptyFileInfo()

    if not 0 == self.rterror:
      return
    if [] == self.alg:
      return
    if not 0 == self.alg.error:
      return

    if 1 == self.vspd.get():
      self.cbpd.configure(text="adjusted ")
      self.cbcd.configure(state="disabled")
      self.vscd_saved = self.vscd.get()
      self.vscd.set(0)
    else:
      self.cbpd.configure(text=" linear ")
      self.cbcd.configure(state="normal")
      self.vscd.set(self.vscd_saved)

  def updateFileCollsView(self):
    self.cbpd.configure(disabledforeground=self.labeldisfg)

    todisable = False
    if not 0 == self.rterror:
      todisable = True
    if [] == self.alg:
      todisable = True
    if not [] == self.alg:
      if not 0 == self.alg.error:
        todisable = True
    if todisable:
      self.cbpd.configure(state="disabled")
      self.cbcc.configure(state="disabled")
      self.cbcd.configure(state="disabled")
      self.vscd_saved = self.vscd.get()
      self.vscd.set(0)
      return

    self.cbpd.configure(disabledforeground=self.labelactfg)
    if [] == self.alg.getAdjusted():
      self.vspd.set(0)
      self.cbpd.configure(state="disabled")
    else:
      self.cbpd.configure(state="normal")

    if 0 == len(self.dispColls):
      self.cbcc.configure(state="disabled")
    else:
      self.cbcc.configure(state="normal")

    if not self.saved:
      self.vscd.set(0)
    if 0 == self.vspd.get():
      self.cbcd.configure(state="normal")
      self.vscd.set(self.vscd_saved)
    else:
      self.cbcd.configure(state="disabled")
      self.vscd_saved = self.vscd.get()
      self.vscd.set(0)

  def updateFileCollsAxes(self):
    self.emptyFileInfo()

    if not 0 == self.rterror:
      return

    if 0 == self.vspd.get():
      if 1 == self.vscd.get():
        if not self.saved:
          self.finfo.configure(text="save cluster-tree first")
          self.freas.configure(text="scale definition is cluster-tree add-on")

    if not self.saved:
      self.vscd.set(0)

  def emptyFileInfo(self):
    self.finfo.configure(text=" ")
    self.freas.configure(text=" ")

  def disableFileLoad(self):
    self.fbutt11.configure(state="disabled")
    self.fbutt21.configure(state="disabled")
    self.fbutt23.configure(state="disabled")
    self.fbutt24.configure(state="disabled")
    self.fbutt41.configure(state="disabled")
    self.fbutt42.configure(state="disabled")
    self.fbutt31.configure(state="disabled")

    self.cbpd.configure(state="disabled")
    self.cbcc.configure(state="disabled")
    self.cbcd.configure(state="disabled")
    self.vscd_saved = self.vscd.get()
    self.vscd.set(0)

    self.emptyFileInfo()

  def loadMatrix(self):
    self.freas.configure(text=" ")

    if not self.disposed:
      self.finfo.configure(text="not loaded")
      self.freas.configure(text="save or dispose the current cluster-tree first")
      return

    fn = False
    fnasked = False
    if not self.datadir:
      self.datadir = os.path.curdir
    if not fnasked:
      fn = F.askopenfilename(title="CSV data file", filetypes=[("csv files", "*.csv"), ("csv files", "*.txt"), ("all files", "*")], initialdir=self.datadir)
    if not fn:
      self.finfo.configure(text=" ")
      return
    if not os.path.isfile(fn):
      self.finfo.configure(text=" no such file ")
      return

    self.datafile = fn
    self.datadir = os.path.dirname(fn)
    self.finfo.configure(text=" ")

    self.saved = False
    self.sname = ""

    self.fbutt22.configure(state="disabled")
    self.fbutt23.configure(state="disabled")
    self.fbutt24.configure(state="disabled")
    self.fbutt31.configure(state="disabled")

    self.cleanTable()
    self.sto.oname = os.path.basename(self.datafile)
    self.sto.fname = ""

    self.cleanClust()
    self.stateClean()
    self.cen.clean()

    self.filestate = False
    try:
      self.loadMatrixFile()
    except:
      self.catchExcept()
      return
    if not self.filestate:
      self.sto.oname = ""
      self.setLabel(None)
      self.updateFileCollsView()
      self.finfo.configure(text=" load failed ")
      return

    self.setLabel(self.sto.oname)
    self.updateFileCollsView()
    self.finfo.configure(text=" data loaded ")

  def loadMatrixFile(self):
    csvsep = "\t"
    vsepval = self.vsep.get()
    csvdata = []
    msgdig = digmod()
    hexdig = ""
    fsize = os.stat(self.datafile)[stat.ST_SIZE]

    try:
      csvfile = open(self.datafile, 'rb')
      msgdig.update(csvfile.read())
      csvfile.close()
      hexdig = msgdig.hexdigest()

    except:
      return False

    if 0 == vsepval:
      csemic = 0
      ccomma = 0
      ctabul = 0
      try:
        csvfile = open(self.datafile, 'r')

        toskip = True
        while toskip:
          toskip = False
          line = csvfile.readline()
          if 0 == len(line):
            toskip = True
          else:
            pass

        for ind in range(3):
          line = csvfile.readline()
          csemic += len(line.split(";"))
          ccomma += len(line.split(","))
          ctabul += len(line.split("\t"))

        csvfile.close()

        csvcnt = ctabul
        if csvcnt <= ccomma:
          csvcnt = ccomma
          csvsep = ","
        if csvcnt <= csemic:
          csvcnt = csemic
          csvsep = ";"

      except:
        return False

    if 1 == vsepval:
      csvsep = ";"
    if 2 == vsepval:
      csvsep = ","
    if 3 == vsepval:
      csvsep = "\t"
    self.sto.delimiter = csvsep

    csvcols = 0
    csvrows = 0
    try:
      csvfile = open(self.datafile, "r")
      csvhand = csv.reader(csvfile, dialect='excel', delimiter=csvsep)
      for row in csvhand:
        csvdata.append(row)
        csvrows += 1
        if csvcols < len(row):
          csvcols = len(row)

      csvfile.close()

    except:
      return False

    if (0 == csvcols) or (0 == csvrows):
      self.freas.configure(text=" no data ")
      return False
    if (3 > csvcols) or (3 > csvrows):
      self.freas.configure(text=" too few data ")
      return False

    self.sto.csvcols = csvcols
    self.sto.csvrows = csvrows
    self.sto.setData(csvdata)
    self.sto.hexdig = hexdig
    self.sto.fsize = str(fsize)

    self.filestate = True
    return True

  def loadClusters(self):
    self.freas.configure(text=" ")

    if not self.disposed:
      self.finfo.configure(text="not loaded")
      self.freas.configure(text="save or dispose the current cluster-tree first")
      return

    fn = False
    fnasked = False
    if not self.datadir:
      self.datadir = os.path.curdir
    if not fnasked:
      fn = F.askopenfilename(title="cluster tree data", filetypes=[("hdf5 files", "*.h5"), ("all files", "*")], initialdir=self.datadir)
    if not fn:
      self.finfo.configure(text=" ")
      return
    if not os.path.isfile(fn):
      self.finfo.configure(text=" no such file ")
      return

    self.datafile = fn
    self.datadir = os.path.dirname(fn)
    self.finfo.configure(text=" ")

    self.saved = False
    self.sname = ""

    self.fbutt22.configure(state="disabled")
    self.fbutt23.configure(state="disabled")
    self.fbutt24.configure(state="disabled")
    self.fbutt31.configure(state="disabled")

    self.cleanClust()
    self.cen.clean()

    self.filestate = False
    try:
      self.loadClustersFile()
    except:
      self.catchExcept()
      return
    if not self.filestate:
      self.sto.dcnames = []
      self.sto.prnames = []
      self.stateGroups()
      self.stateAxes()
      if "" == self.sto.oname:
        self.setLabel(None)
      else:
        self.setLabel(self.sto.oname)
      self.finfo.configure(text=" load failed ")
      self.updateFileCollsView()
      return

    self.saved = True
    self.sname = os.path.basename(self.datafile)
    self.stateGroups()
    self.stateAxes()
    self.setDispCases()
    self.stateAxialSels(0)

    self.bs1.configure(state="normal")
    self.bs2.configure(state="normal")

    self.fbutt22.configure(state="normal")
    self.fbutt23.configure(state="normal")
    self.fbutt24.configure(state="normal")

    flabel = ""
    if "" == self.sto.oname:
      flabel = self.sto.fname + " / " + self.sname
    else:
      flabel = self.sto.oname + " / " + self.sname
    self.setLabel(flabel)
    self.updateFileCollsView()
    self.finfo.configure(text=" data loaded ")

  def loadClustersFile(self):
    """
    * root directory
    - [self.ctdftype, self.ctdversion, sto.treeid] [str] 3 {2x equal, nonempty}
    c [date, view] [str] 2 {date, "fixed"/"volatile"} optional for version == "1.1"
    - dclen=len(sto.dcnames), prlen=len(sto.prnames), cls.tree.lcount [int] 3 {finite >= 3}
    * conf directory (soft checks)
    - [sto.fname, sto.delimiter, sto.props] [str] 3 {2x non_empty, "rows"/"cols"}
    c [sto.fsize, sto.hexdig] [str] 2 {2x non_empty} for version == "1.1"
    - [sto.csvrows] + [sto.infrows] [int] 4 {finite >= 3, 3x (0..sto.csvrows)}
    - [sto.csvcols] + [sto.infcols] [int] 4 {finite >= 3, 3x (0..sto.csvcols)}
    - [sto.dclog, sto.dcwin, sto.dcnor, sto.basec] [int] 4 {0/1, 0/1, 0..2, 3..200}
    c [sto.winstype, sto.winsvalue] [int] 2 {0/1, 1..20} for version == "1.1"
    - [sto.clust, cen.kmruncycles, cen.kmconverged] [int] 3 {0/1, finite >= 0, 0/1}
    - [cen.cencount] + [cen.inicenters] [int] 1 + 0..200 {3..200, 1..prlen}
    * names directory
    - [sto.dcnames] [str] dclen {non_empty}
    - [sto.prnames] [str] prlen {non_empty}
    * props directory
    - [cen.propmeans] [N.float] prlen {finite}
    - [cen.propvarts] [N.float] prlen {finite}
    - [propgroups = cen.grprops * cen.orients] [N.int] prlen {-lcount .. lcount}
    * tree directory
    - [vgroup:axes_initial, vexact:axes_middle] [int] 2 {2x 0/1}
    c [vadjust:axes_final, adjloc:locality] [int] 2 {0/1, 1..50} for version == "1.1"
    - [len] + [cls.selectOne(each-1)] [int] 1 + 0..lcount {1 .. ((2 * lcount) - 1)}
    - [cls.tree.tnodes] [[N.float]] (((2 * lcount) - 1) x dclen) {finite}
    - [cls.tree.tstruc] [[N.int]] ((lcount - 1) x 2) {lesser, abs() > 0}
    """

    optview = True
    curversion = ""

    info = []
    dims = []

    treeid = ""
    dclen = -1
    prlen = -1

    lcount = -1
    icount = -1
    tcount = -1

    fileinfo = []
    rowsinfo = []
    colsinfo = []
    runsinfo = []

    dcnames = []
    prnames = []

    propmeans = []
    propvarts = []
    propgroups = []
    grprops = []
    orients = []

    seltype = []
    selgrps = []
    tnodes = []
    tstruc = []

    h5f = []

    if not os.access(self.datafile, os.R_OK):
      self.freas.configure(text="file not readable")
      return False

    try:
      h5f = H.openFile(self.datafile)
    except:
      self.freas.configure(text="wrong file fromat")
      return False

    try:
      info = h5f.root.info.read()
      if not type([]) == type(info):
        h5f.close()
        self.freas.configure(text="unknown format spec")
        return False
      for item_idx in range(len(info)):
        item = info[item_idx]
        if not type("") == type(item):
          if bytes == type(item):
            info[item_idx] = item.decode("utf-8")
          else:
            h5f.close()
            self.freas.configure(text="unknown format type")
            return False
      if 3 > len(info):
        h5f.close()
        return False
      if not self.ctdftype == str(info[0]):
        h5f.close()
        self.freas.configure(text="wrong format type")
        return False
      if not str(info[1]) in ["1.0", "1.1"]:
        h5f.close()
        self.freas.configure(text="wrong format version")
        return False
      curversion = str(info[1])
      treeid = str(info[2])
      if 0 >= len(treeid):
        h5f.close()
        self.freas.configure(text="empty tree id")
        return False
      if 5 <= len(info):
        viewtype = info[4]
        if "fixed" == viewtype:
          optview = False
      dims = h5f.root.dims.read()
      if not type([]) == type(dims):
        h5f.close()
        self.freas.configure(text="wrong data")
        return False
      if 3 > len(dims):
        h5f.close()
        self.freas.configure(text="wrong data")
        return False
      for ind in range(3):
        try:
          dims[ind] = int(dims[ind])
        except:
          h5f.close()
          self.freas.configure(text="wrong data")
          return False
        if 3 >= dims[ind]:
          h5f.close()
          self.freas.configure(text="wrong data")
          return False
      dclen = dims[0]
      prlen = dims[1]
      lcount = dims[2]

      icount = lcount - 1
      tcount = (2 * lcount) - 1

      fileinfo = h5f.root.conf.fileinfo.read()
      rowsinfo = h5f.root.conf.rowsinfo.read()
      colsinfo = h5f.root.conf.colsinfo.read()
      confinfo = h5f.root.conf.confinfo.read()
      runsinfo = h5f.root.conf.runsinfo.read()
      initinfo = h5f.root.conf.initinfo.read()

      dcnames = h5f.root.names.cases.read()
      prnames = h5f.root.names.props.read()

      propmeans = h5f.root.props.means.read()
      propvarts = h5f.root.props.varts.read()
      propgroups = h5f.root.props.groups.read()

      seltype = h5f.root.tree.seltype.read()
      selgrps = h5f.root.tree.selnodes.read()
      tnodes = h5f.root.tree.tnodes.read()
      tstruc = h5f.root.tree.tstruc.read()

      h5f.close()

    except:
      h5f.close()
      self.freas.configure(text="wrong file data")
      return False

    floatLong = ""
    isflong = True
    try:
      floatLong = N.float128
    except:
      isflong = False
    if not isflong:
      isflong = True
      try:
        floatLong = N.float96
      except:
        isflong = False
    if not isflong:
      floatLong = float

    try:
      if not type([]) == type(fileinfo):
        return False
      if 3 > len(fileinfo):
        return False
      for item_idx in range(len(fileinfo)):
        item = fileinfo[item_idx]
        if not type("") == type(item):
          if bytes == type(item):
            fileinfo[item_idx] = item.decode("utf-8")
          else:
            self.freas.configure(text="unknown fileinfo type")
            return False
      for fileinfoindex in range(3):
        item = fileinfo[fileinfoindex]
        if 0 == len(item):
          return False

      for ind in range(5):
        curinf = [rowsinfo, colsinfo, confinfo, runsinfo, initinfo][ind]
        curlen = [4, 4, 4, 3, 1][ind]
        if not type([]) == type(curinf):
          return False
        if curlen > len(curinf):
          return False
        for infind in range(len(curinf)):
          try:
            curinf[infind] = int(curinf[infind])
          except:
            return False

      if not type([]) == type(dcnames):
        return False
      if not dclen == len(dcnames):
        return False
      for item_idx in range(len(dcnames)):
        item = dcnames[item_idx]
        if not type("") == type(item):
          if bytes == type(item):
            dcnames[item_idx] = item.decode("utf-8")
          else:
            self.freas.configure(text="unknown names type")
            return False
        if 0 == len(dcnames[item_idx]):
          self.freas.configure(text="empty names data")
          return False

      if not type([]) == type(prnames):
        return False
      if not prlen == len(prnames):
        return False
      for item_idx in range(len(prnames)):
        item = prnames[item_idx]
        if not type("") == type(item):
          if bytes == type(item):
            prnames[item_idx] = item.decode("utf-8")
          else:
            self.freas.configure(text="unknown names type")
            return False
        if 0 == len(prnames[item_idx]):
          self.freas.configure(text="empty names data")
          return False

      for propinfo in (propmeans, propvarts):
        if not type(N.empty(0)) == type(propinfo):
          return False
        if not propinfo.dtype in [float, N.float32, N.float64, floatLong]:
          return False
        if not propinfo.shape == (prlen,):
          return False

      if not type(N.empty(0)) == type(propgroups):
        return False
      if not propgroups.dtype in [N.int32, N.int64]:
        return False
      if not propgroups.shape == (prlen,):
        return False
      for item in propgroups:
        if lcount < abs(item):
          return False

      if not type([]) == type(seltype):
        return False
      if not 2 <= len(seltype):
        return False
      for itemind in range(len(seltype)):
        try:
          seltype[itemind] = int(seltype[itemind])
        except:
          return False
      for seltypeindex in range(len(seltype)):
        if 3 > seltypeindex:
          if not seltype[seltypeindex] in [0, 1]:
            return False
        if 3 == seltypeindex:
          if not seltype[seltypeindex] in range(1, 51):
            return False

      if not type([]) == type(selgrps):
        return False
      if 0 == len(selgrps):
        return False
      for itemind in range(len(selgrps)):
        try:
          selgrps[itemind] = int(selgrps[itemind])
        except:
          return False
      sellen = selgrps[0]
      if not (sellen + 1) == len(selgrps):
        return False
      selgrps = selgrps[1:]
      for item in selgrps:
        if (1 > item) or (tcount < item):
          return False

      if not type(N.empty(0)) == type(tnodes):
        return False
      if not tnodes.dtype in [float, N.float32, N.float64, floatLong]:
        return False
      if not tnodes.shape == (tcount, dclen):
        return False
      if not (tcount * dclen) == N.sum(N.isfinite(tnodes)):
        return False

      if not type(N.empty(0)) == type(tstruc):
        return False
      if not tstruc.dtype in [N.int32, N.int64]:
        return False
      if not tstruc.shape == (icount, 2):
        return False
      for irow in range(icount):
        for icol in range(2):
          cval = abs(tstruc[irow, icol])
          if (0 >= cval) or ((lcount+irow) < cval):
            return False

    except:
      return False

    gsizes = N.zeros(tcount, dtype=int)
    grprops = N.zeros(prlen, dtype=int)
    orients = N.zeros(prlen, dtype=int)

    for ind in range(prlen):
      ival = propgroups[ind]
      grind = abs(ival) - 1
      if -1 < grind:
        grprops[ind] = grind
        gsizes[grind] += 1
        if 0 < ival:
          orients[ind] = 1
        elif 0 > ival:
          orients[ind] = -1

    for ind in range(lcount):
      if 0 == gsizes[ind]:
        self.freas.configure(text="empty basic group(s)")
        return False
    self.cls.renewTree(tnodes, lcount, dclen, gsizes, tstruc, optview)

    if not self.cls.filled:
      self.cls.empty()
      self.freas.configure(text="wrong tree structure")
      return False

    toclean = False
    if "1.0" == curversion:
      toclean = True
    if "1.1" == curversion:
      if 5 > len(fileinfo):
        toclean = True
    if not toclean:
      if not self.sto.delimiter == fileinfo[1]:
        toclean = True
      if not self.sto.fsize == fileinfo[3]:
        toclean = True
      if not self.sto.hexdig == fileinfo[4]:
        toclean = True
      if not self.sto.csvrows == rowsinfo[0]:
        toclean = True
      if not self.sto.csvcols == colsinfo[0]:
        toclean = True

    if toclean:
      self.cleanTable()
      self.sto.oname = ""

    self.sto.fsize = ""
    self.sto.hexdig == ""
    if "1.1" == curversion:
      if 5 <= len(fileinfo):
        self.sto.fsize = fileinfo[3]
        self.sto.hexdig == fileinfo[4]

    self.sto.winstype = 0
    self.sto.winsvalue = 0
    if "1.1" == curversion:
      if 6 <= len(confinfo):
        self.sto.winstype = confinfo[4]
        self.sto.winsvalue = confinfo[5]

    self.sto.treeid = treeid

    pscaled = seltype[0]
    padjust = 0
    plocal = 10

    self.vgroup.set(seltype[0])
    if 0 == seltype[0]:
      self.vgr1.configure(text="uninorm ")
    else:
      self.vgr1.configure(text="scaled ")
    if "1.1" == curversion:
      if 4 <= len(seltype):
        adjtype = seltype[2]
        self.vadjust.set(adjtype)
        padjust = seltype[2]
        plocal = seltype[3]
        if 1 == adjtype:
          self.adjloc.set(seltype[3])
          self.vgr3.configure(text="adjusted ")
        else:
          self.vgr3.configure(text="linear ")
      else:
        self.vadjust.set(0)
        self.vgr3.configure(text="linear ")
    else:
      self.vadjust.set(0)
      self.vgr3.configure(text="linear ")

    psels = []
    for onesel in selgrps:
      onesel -= 1
      self.cls.selectOne(onesel)
      psels.append(onesel)

    self.sto.dcnames = dcnames
    self.sto.prnames = prnames

    self.cen.propmeans = propmeans
    self.cen.propvarts = propvarts
    self.cen.grprops = grprops
    self.cen.orients = orients
    self.cen.cengrps = lcount * [[], ]
    for ind in range(lcount):
      self.cen.cengrps[ind] = []
    for ind in range(prlen):
      grind = grprops[ind]
      if -1 < grind:
        self.cen.cengrps[grind].append(ind)

    self.sto.fname = os.path.basename(fileinfo[0])
    self.sto.delimiter = fileinfo[1]
    self.sto.props = fileinfo[2]

    self.sto.csvrows = rowsinfo[0]
    self.sto.infrows = rowsinfo[1:]
    self.sto.csvcols = colsinfo[0]
    self.sto.infcols = colsinfo[1:]
    for ind in range(len(self.sto.infrows)):
      self.sto.infrows[ind] -= 1
    for ind in range(len(self.sto.infcols)):
      self.sto.infcols[ind] -= 1

    self.sto.dclog = confinfo[0]
    self.sto.dcwin = confinfo[1]
    self.sto.dcnor = confinfo[2]
    self.sto.basec = confinfo[3]

    self.sto.clust = runsinfo[0]
    self.cen.kmruncycles = runsinfo[1]
    self.cen.kmconverged = runsinfo[2]

    self.cen.cencount = initinfo[0]
    self.cen.inicenters = initinfo[1:]
    for ind in range(len(self.cen.inicenters)):
      self.cen.inicenters[ind] -= 1

    curparams = {"scaled": pscaled, "adjust": padjust, "local": plocal, "sels": psels}
    curparams["type"] = "loaded"

    self.selList.insert(0, curparams)
    self.updateSelsList()
    self.lbevH.select_clear(0, "end")
    self.lbevH.see(0)
    self.lbevH.activate(0)

    self.filestate = True
    return True

  def saveClusters(self):
    self.freas.configure(text=" ")

    fn = False
    fnasked = False
    if not self.datadir:
      self.datadir = os.path.curdir
    if not fnasked:
      fn = F.asksaveasfilename(title="cluster tree data", filetypes=[("hdf5 files", "*.h5"), ("all files", "*")], initialdir=self.datadir, confirmoverwrite=False)
    if not fn:
      self.finfo.configure(text=" ")
      return
    dn = os.path.dirname(fn)
    if not os.path.isdir(dn):
      self.finfo.configure(text=" no such directory ")
      return

    self.datafile = fn
    self.datadir = dn
    self.finfo.configure(text=" ")

    self.filestate = False
    self.saveClustersFile()
    if not self.filestate:
      self.finfo.configure(text=" save failed ")
      self.updateFileCollsView()
      return

    self.saved = True
    self.disposed = True
    self.bgclust.configure(text=" Cluster ")
    self.updateFileCollsView()
    self.finfo.configure(text=" data saved ")

  def saveClustersFile(self):
    fbase = self.datafile
    for suff, slen in [[".h5", 3], [".hdf", 4], [".hdf5", 5], [".csv", 4], [".txt", 4]]:
      if fbase.endswith(suff):
        fbase = fbase[:(-1 * slen)]
        break

    fbase0 = os.path.basename(fbase)
    if 0 == len(fbase0):
      self.freas.configure(text="no file-name")
      return False

    fnhdf = fbase + ".h5"
    fncsv = fbase + ".csv"
    fnimg = fbase + ".eps"
    fwname = fbase0 + ".h5"

    isav = True
    if os.path.exists(fnhdf):
      isav = False
    if os.path.exists(fncsv):
      isav = False
    if os.path.exists(fnimg):
      isav = False

    lcount = self.cls.tree.lcount
    tcount = (2 * lcount) - 1
    icount = lcount - 1

    curind = 0
    while not isav:
      isav = True
      curind += 1
      fnhdf = fbase + "-" + str(curind) + ".h5"
      fncsv = fbase + "-" + str(curind) + ".csv"
      fnimg = fbase + "-" + str(curind) + ".eps"
      fwname = fbase0 + "-" + str(curind) + ".h5"
      if os.path.exists(fnhdf):
        isav = False
      if os.path.exists(fncsv):
        isav = False
      if os.path.exists(fnimg):
        isav = False

    try:
      pstr = self.cls.getImagePS()
      hps = open(fnimg, "w")
      hps.write(pstr)
      hps.close()
    except:
      self.freas.configure(text="can not create files")
      return False

    try:
      h5f = H.openFile(fnhdf, "w")
    except:
      self.freas.configure(text="can not create files")
      return False

    try:
      cdate = time.strftime("%Y-%m-%d %H:%M:%S", time.gmtime())
      onelist = [self.ctdftype, self.ctdversion, self.sto.treeid, cdate, "volatile"]
      onelist += ["CluDisc cluster-tree data"]
      h5arr = h5f.createArray(h5f.root, "info", [item.encode("utf-8") for item in onelist], "data info")
      onelist = [len(self.sto.dcnames), len(self.sto.prnames), self.cls.tree.lcount]
      h5arr = h5f.createArray(h5f.root, "dims", onelist, "data dimensions")

      h5conf = h5f.createGroup(h5f.root, "conf", "configuration options")
      onelist = [self.sto.fname, self.sto.delimiter, self.sto.props]
      onelist += [self.sto.fsize, self.sto.hexdig]
      h5arr = h5f.createArray(h5conf, "fileinfo", [item.encode("utf-8") for item in onelist], "file info")
      onelist = [self.sto.csvrows]
      for iinf in self.sto.infrows:
        onelist.append(iinf + 1)
      h5arr = h5f.createArray(h5conf, "rowsinfo", onelist, "rows info")
      onelist = [self.sto.csvcols]
      for iinf in self.sto.infcols:
        onelist.append(iinf + 1)
      h5arr = h5f.createArray(h5conf, "colsinfo", onelist, "cols info")
      onelist = [self.sto.dclog, self.sto.dcwin, self.sto.dcnor, self.sto.basec]
      onelist += [self.sto.winstype, self.sto.winsvalue]
      h5arr = h5f.createArray(h5conf, "confinfo", onelist, "conf info")
      onelist = [self.sto.clust, self.cen.kmruncycles, self.cen.kmconverged]
      h5arr = h5f.createArray(h5conf, "runsinfo", onelist, "runs info")
      onelist = [self.cen.cencount]
      for icen in self.cen.inicenters:
        onelist.append(icen + 1)
      h5arr = h5f.createArray(h5conf, "initinfo", onelist, "init info")

      h5names = h5f.createGroup(h5f.root, "names", "data names")
      h5arr = h5f.createArray(h5names, "cases", [item.encode("utf-8") for item in self.sto.dcnames], "data case names")
      h5arr = h5f.createArray(h5names, "props", [item.encode("utf-8") for item in self.sto.prnames], "properties names")

      h5props = h5f.createGroup(h5f.root, "props", "properties info")
      h5arr = h5f.createArray(h5props, "means", self.cen.propmeans, "props means")
      h5arr = h5f.createArray(h5props, "varts", self.cen.propvarts, "props variations")
      prlen = len(self.sto.prnames)
      propgroups = N.empty(prlen, dtype=int)
      for ind in range(prlen):
        propgroups[ind] = (self.cen.grprops[ind] + 1) * self.cen.orients[ind]
      h5arr = h5f.createArray(h5props, "groups", propgroups, "props groups")

      h5tree = h5f.createGroup(h5f.root, "tree", "cluster tree")
      onelist = [self.vgroup.get(), self.vexact.get()]
      onelist += [self.vadjust.get(), self.adjloc.get()]
      h5arr = h5f.createArray(h5tree, "seltype", onelist, "selection type")
      sels = self.cls.getSelections()
      onelist = [len(sels)]
      for isel in sels:
        onelist.append(isel + 1)
      h5arr = h5f.createArray(h5tree, "selnodes", onelist, "selection nodes")
      h5arr = h5f.createArray(h5tree, "tnodes", self.cls.tree.tnodes, "tree nodes")
      h5arr = h5f.createArray(h5tree, "tstruc", self.cls.tree.tstruc, "tree structure")

      h5f.flush()
      h5f.close()

    except:
      h5f.close()
      self.freas.configure(text="can not write into files")
      return False

    fhcsv = []
    try:
      fhcsv = open(fncsv, "w")
    except:
      self.freas.configure(text="can not create files")
      return False

    try:
      cwcsv = csv.writer(fhcsv, delimiter=";", dialect='excel')
      cwcsv.writerow(["CluDisc tree groups"])
      cwcsv.writerow(["rank", "size", "variation", "ratio", "leaves"])

      for ind in range(tcount):
        crati = 0.0
        crank = str(ind + 1)
        csize = self.cls.tree.amounts[ind]
        cvmax = N.sqrt(csize)
        cvart = cvmax * self.cls.tree.variats[ind]
        if 1 < csize:
          crati = (cvart - 1) / (cvmax - 1)
        elif 1 == csize:
          crati = cvart
          if 0.000001 > crati:
            crati = 0.0
          if 0.999999 < crati:
            crati = 1.0
        if 0.0 > crati:
          crati = 0.0
        if 1.0 < crati:
          crati = 1.0
        wrank = str(crank)
        wsize = str(csize)
        wvart = "%(val)1.4e" % {'val': cvart}
        wrati = "%(val).4f" % {'val': (100 * crati)}
        clist = [wrank, wsize, wvart, wrati]
        clist += self.cls.tree.getNodeLists(ind)
        cwcsv.writerow(clist)

      cwcsv.writerow([])
      cwcsv.writerow(["CluDisc base clusters"])
      cwcsv.writerow(["rank", "content"])

      for ind in range(lcount):
        curgrp = self.cen.cengrps[ind]
        curlen = len(curgrp)
        clist = (curlen + 1) * [""]
        clist[0] = str(ind + 1)
        for ind in range(curlen):
          grind = curgrp[ind]
          orirank = self.cen.orients[grind] + 1
          orisign = ["-", "", "+"][orirank]
          clist[ind+1] = orisign + self.sto.prnames[grind]
        cwcsv.writerow(clist)

      fhcsv.close()

    except:
      fhcsv.close()
      self.freas.configure(text="can not write into files")
      return False

    self.sname = os.path.basename(fnhdf)
    self.freas.configure(text=fwname[:50])
    self.filestate = True
    return True

  def loadSelections(self):
    self.freas.configure(text=" ")

    fn = False
    fnasked = False
    if not self.datadir:
      self.datadir = os.path.curdir
    if not fnasked:
      fn = F.askopenfilename(title="tree node selections", filetypes=[("hdf5 files", "*.h5"), ("all files", "*")], initialdir=self.datadir)
    if not fn:
      self.finfo.configure(text=" ")
      return
    if not os.path.isfile(fn):
      self.finfo.configure(text=" no such file ")
      return

    self.datafile = fn
    self.datadir = os.path.dirname(fn)
    self.finfo.configure(text=" ")

    self.cls.emptyTreeSel()

    self.filestate = False
    try:
      self.loadSelectionsFile()
    except:
      self.catchExcept()
      return
    if not self.filestate:
      self.finfo.configure(text=" load failed ")
      self.stateAxialSels(0)
      return

    self.stateAxialSels(0)
    self.finfo.configure(text=" data loaded ")

  def loadSelectionsFile(self):
    curversion = ""

    lcount = self.cls.tree.lcount
    tcount = (2 * lcount) - 1

    seltype = []
    selgrps = []

    h5f = []

    if not os.access(self.datafile, os.R_OK):
      self.freas.configure(text="file not readable")
      return False

    try:
      h5f = H.openFile(self.datafile)
    except:
      self.freas.configure(text="wrong file fromat")
      return False

    try:
      info = h5f.root.info.read()
      if not type([]) == type(info):
        h5f.close()
        self.freas.configure(text="unknown format spec")
        return False
      if 3 > len(info):
        h5f.close()
        return False
      for item_idx in range(len(info)):
        item = info[item_idx]
        if not type("") == type(item):
          if bytes == type(item):
            info[item_idx] = item.decode("utf-8")
          else:
            self.freas.configure(text="unknown format type")
            return False
      if not self.selftype == str(info[0]):
        h5f.close()
        self.freas.configure(text="wrong format type")
        return False
      if not str(info[1]) in ["1.0", "1.1"]:
        h5f.close()
        self.freas.configure(text="wrong format version")
        return False
      if not self.sto.treeid == str(info[2]):
        h5f.close()
        self.freas.configure(text="different tree id")
        return False

      seltype = h5f.root.tree.seltype.read()
      selgrps = h5f.root.tree.selnodes.read()

      h5f.close()

    except:
      h5f.close()
      self.freas.configure(text="wrong file data")
      return False

    curversion = str(info[1])

    try:
      if not type([]) == type(seltype):
        return False
      if not 2 <= len(seltype):
        return False
      for itemind in range(len(seltype)):
        try:
          seltype[itemind] = int(seltype[itemind])
        except:
          return False
      for seltypeindex in range(len(seltype)):
        if 3 > seltypeindex:
          if not seltype[seltypeindex] in [0, 1]:
            return False
        if 3 == seltypeindex:
          if not seltype[seltypeindex] in range(1, 51):
            return False

      if not type([]) == type(selgrps):
        return False
      if 0 == len(selgrps):
        return False
      for itemind in range(len(selgrps)):
        try:
          selgrps[itemind] = int(selgrps[itemind])
        except:
          return False
      sellen = selgrps[0]
      if not (sellen + 1) == len(selgrps):
        return False
      selgrps = selgrps[1:]
      for item in selgrps:
        if (1 > item) or (tcount < item):
          return False

    except:
      return False

    pscaled = seltype[0]
    padjust = 0
    plocal = 10

    self.vgroup.set(seltype[0])
    if 0 == seltype[0]:
      self.vgr1.configure(text="uninorm ")
    else:
      self.vgr1.configure(text="scaled ")
    if "1.1" == curversion:
      if 4 <= len(seltype):
        adjtype = seltype[2]
        self.vadjust.set(adjtype)
        padjust = seltype[2]
        plocal = seltype[3]
        if 1 == adjtype:
          self.adjloc.set(seltype[3])
          self.vgr3.configure(text="adjusted ")
        else:
          self.vgr3.configure(text="linear ")
      else:
        self.vadjust.set(0)
        self.vgr3.configure(text="linear ")
    else:
      self.vadjust.set(0)
      self.vgr3.configure(text="linear ")

    psels = []
    for isel in selgrps:
      isel -= 1
      self.cls.selectOne(isel)
      psels.append(isel)

    curparams = {"scaled": pscaled, "adjust": padjust, "local": plocal, "sels": psels}
    curparams["type"] = "loaded"

    self.selList.insert(0, curparams)
    self.updateSelsList()
    self.lbevH.select_clear(0, "end")
    self.lbevH.see(0)
    self.lbevH.activate(0)

    self.filestate = True
    return True

  def saveSelections(self):
    self.freas.configure(text=" ")

    if not self.saved:
      self.finfo.configure(text="save cluster-tree first")
      self.freas.configure(text="selectins are cluster-tree add-on")
      return

    inifn = "sels"
    namebase = ""
    dotind = self.sname.rfind(".")
    if not -1 == dotind:
      namebase = self.sname[:dotind]
    else:
      namebase = self.sname
    if not "" == namebase:
      inifn = namebase + "-sels"

    fn = False
    fnasked = False
    if not self.datadir:
      self.datadir = os.path.curdir
    if not fnasked:
      fn = F.asksaveasfilename(title="tree node selections", filetypes=[("hdf5 files", "*.h5"), ("all files", "*")], initialdir=self.datadir, initialfile=inifn, confirmoverwrite=False)
    if not fn:
      self.finfo.configure(text=" ")
      return
    dn = os.path.dirname(fn)
    if not os.path.isdir(dn):
      self.finfo.configure(text=" no such directory ")
      return

    self.datafile = fn
    self.datadir = dn
    self.finfo.configure(text=" ")

    self.filestate = False
    self.saveSelectionsFile()
    if not self.filestate:
      self.finfo.configure(text=" save failed ")
      return

    self.fbutt23.configure(state="normal")
    self.finfo.configure(text=" data saved ")

  def saveSelectionsFile(self):
    fbase = self.datafile
    for suff, slen in [[".h5", 3], [".hdf", 4], [".hdf5", 5], [".csv", 4], [".txt", 4]]:
      if fbase.endswith(suff):
        fbase = fbase[:(-1 * slen)]
        break

    fbase0 = os.path.basename(fbase)
    if 0 == len(fbase0):
      self.freas.configure(text="no file-name")
      return False

    fnhdf = fbase + ".h5"
    fncsv = fbase + ".csv"
    fwname = fbase0 + ".h5"

    isav = True
    if os.path.exists(fnhdf):
      isav = False
    if os.path.exists(fncsv):
      isav = False

    curind = 0
    while not isav:
      isav = True
      curind += 1
      fnhdf = fbase + "-" + str(curind) + ".h5"
      fncsv = fbase + "-" + str(curind) + ".csv"
      fwname = fbase0 + "-" + str(curind) + ".h5"
      if os.path.exists(fnhdf):
        isav = False
      if os.path.exists(fncsv):
        isav = False

    try:
      h5f = H.openFile(fnhdf, "w")

    except:
      self.freas.configure(text="can not create files")
      return False

    try:
      cdate = time.strftime("%Y-%m-%d %H:%M:%S", time.gmtime())
      onelist = [self.selftype, self.selversion, self.sto.treeid, cdate, "CluDisc selection data"]
      h5arr = h5f.createArray(h5f.root, "info", [item.encode("utf-8") for item in onelist], "data info")
      h5tree = h5f.createGroup(h5f.root, "tree", "cluster tree")
      onelist = [self.vgroup.get(), self.vexact.get()]
      onelist += [self.vadjust.get(), self.adjloc.get()]
      h5arr = h5f.createArray(h5tree, "seltype", onelist, "selection type")
      sels = self.cls.getSelections()
      onelist = [len(sels)]
      for isel in sels:
        onelist.append(isel + 1)
      h5arr = h5f.createArray(h5tree, "selnodes", onelist, "selection nodes")

      h5f.flush()
      h5f.close()

    except:
      h5f.close()
      self.freas.configure(text="can not write into files")
      return False

    fhcsv = []
    try:
      fhcsv = open(fncsv, "w")

    except:
      self.freas.configure(text="can not create files")
      return False

    try:
      cwcsv = csv.writer(fhcsv, delimiter=";", dialect='excel')
      cwcsv.writerow(["CluDisc tree selections"])

      item = ["uninorm", "scaled"][self.vgroup.get()]
      cwcsv.writerow([item])
      item = ["fail-safe", "exact"][self.vexact.get()]
      cwcsv.writerow([item])
      item = ["linear", "adjusted"][self.vadjust.get()]
      oneline = [item]
      if 1 == self.vadjust.get():
        oneline += [str(self.adjloc.get())]
      cwcsv.writerow(oneline)

      for isel in self.cls.getSelections():
        item = str(isel + 1)
        cwcsv.writerow([item])

      fhcsv.close()

    except:
      fhcsv.close()
      self.freas.configure(text="can not write into files")
      return False

    self.freas.configure(text=fwname[:50])
    self.filestate = True
    return True

  def loadCollections(self):
    self.freas.configure(text=" ")

    fn = False
    fnasked = False
    if not self.datadir:
      self.datadir = os.path.curdir
    if not fnasked:
      fn = F.askopenfilename(title="collection defs", filetypes=[("hdf5 files", "*.h5"), ("all files", "*")], initialdir=self.datadir)
    if not fn:
      self.finfo.configure(text=" ")
      return
    if not os.path.isfile(fn):
      self.finfo.configure(text=" no such file ")
      return

    self.datafile = fn
    self.datadir = os.path.dirname(fn)
    self.finfo.configure(text=" ")

    self.filestate = False
    try:
      self.loadCollectionsFile()
    except:
      self.catchExcept()
      return
    if not self.filestate:
      self.finfo.configure(text=" load failed ")
      self.updateFileCollsView()
      return

    self.updateFileCollsView()
    self.finfo.configure(text=" data loaded ")

  def loadCollectionsFile(self):
    dclen = 0
    collen = 0

    dcnames = []
    dccolls = []

    collranks = []
    collsizes = []
    collcolors = []
    collcomms = []

    h5f = []

    if not os.access(self.datafile, os.R_OK):
      self.freas.configure(text="file not readable")
      return False

    try:
      h5f = H.openFile(self.datafile)
    except:
      self.freas.configure(text="wrong file fromat")
      return False

    try:
      info = h5f.root.info.read()
      if not type([]) == type(info):
        h5f.close()
        return False
      if 2 > len(info):
        h5f.close()
        return False
      for item_idx in range(len(info)):
        item = info[item_idx]
        if not type("") == type(item):
          if bytes == type(item):
            info[item_idx] = item.decode("utf-8")
      if not self.colftype == str(info[0]):
        h5f.close()
        self.freas.configure(text="wrong format type")
        return False
      if not str(info[1]) in ["1.0"]:
        h5f.close()
        self.freas.configure(text="wrong format version")
        return False

      dims = h5f.root.dims.read()
      if not type([]) == type(dims):
        h5f.close()
        self.freas.configure(text="wrong data")
        return False
      if 2 > len(dims):
        h5f.close()
        self.freas.configure(text="wrong data")
        return False
      for ind in range(2):
        try:
          dims[ind] = int(dims[ind])
        except:
          h5f.close()
          self.freas.configure(text="wrong data")
          return False
      for ind, minlen in [[0, 0], [1, 1]]:
        if minlen > dims[ind]:
          h5f.close()
          self.freas.configure(text="wrong data")
          return False
      dclen = dims[0]
      collen = dims[1]

      dcnames = h5f.root.cases.names.read()
      if type([]) == type(dcnames):
        for item_idx in range(len(dcnames)):
          item = dcnames[item_idx]
          if not type("") == type(item):
            if bytes == type(item):
              dcnames[item_idx] = item.decode("utf-8")
      dccolls = h5f.root.cases.colls.read()

      collranks = h5f.root.colls.ranks.read()
      collsizes = h5f.root.colls.sizes.read()
      collcolors = h5f.root.colls.colors.read()
      if type([]) == type(collcolors):
        for item_idx in range(len(collcolors)):
          item = collcolors[item_idx]
          if not type("") == type(item):
            if bytes == type(item):
              collcolors[item_idx] = item.decode("utf-8")
      collcomms = h5f.root.colls.comms.read()
      if type([]) == type(collcomms):
        for item_idx in range(len(collcomms)):
          item = collcomms[item_idx]
          if not type("") == type(item):
            if bytes == type(item):
              collcomms[item_idx] = item.decode("utf-8")

      h5f.close()

    except:
      h5f.close()
      self.freas.configure(text="wrong file data")
      return False

    try:
      namlists = [dcnames, dccolls]
      collists = [collranks, collsizes, collcolors, collcomms]

      for onelist in (namlists + collists):
        if not type([]) == type(onelist):
          return False

      for onelist in namlists:
        if not len(onelist) >= dclen:
          return False

      for onelist in collists:
        if not len(onelist) == collen:
          return False

      for onelist in [dcnames, collcolors, collcomms]:
        for item in onelist:
          if not type("") == type(item):
            return False

      for onelist in [dccolls, collranks, collsizes]:
        for itemind in range(len(onelist)):
          try:
            onelist[itemind] = int(onelist[itemind])
          except:
            return False

      for item in dccolls:
        if (0 > item) or (collen < item):
          return False

      curmax = 0
      for item in collranks:
        if not curmax < item:
          return False
        curmax = item

      for item in collsizes:
        if not item in range(1, 21):
          return False

      for item in collcolors:
        if not item in self.rosacolors:
          return False

    except:
      return False

    dcnames = dcnames[:dclen]
    dccolls = dccolls[:dclen]

    if not dcnames == self.dispCases:
      self.dispHigh = -1
      self.dispHighName.set("")

    if [] == self.sto.dcnames:
      self.dispCases = dcnames
    self.dispCaseRanks = list(range(len(self.dispCases)))

    self.dispColls = []
    self.dispCollCases = []
    self.dispCollDiameters = []
    self.dispCollColors = []
    self.dispCollComments = []

    for ind in range(collen):
      rank = collranks[ind]
      self.dispColls.append("group " + str(rank))

      self.dispCollCases.append([])
      self.dispCollDiameters.append(collsizes[ind])
      clrstr = collcolors[ind]
      clrind = list(self.rosacolors).index(clrstr)
      self.dispCollColors.append(clrind)
      self.dispCollComments.append(collcomms[ind])

    todiff = True
    curlen = 0
    if [] == self.sto.dcnames:
      todiff = False
    else:
      if dcnames == self.sto.dcnames:
        todiff = False
      else:
        curlen = len(self.sto.dcnames)

    if todiff:
      probnames = dcnames
      probcolls = dccolls
      self.distCaseColls(probnames, probcolls)

    else:
      for ind in range((dclen-1), -1, -1):
        colind = dccolls[ind] - 1
        if -1 < colind:
          self.dispCaseRanks.pop(ind)
          self.dispCollCases[colind].append(ind)
      for ind in range(collen):
        self.dispCollCases[ind].sort()

    self.setDispCasesData()

    self.vgc1info.set(tuple(self.dispColls))
    listind = 0
    self.lbgc1.activate(listind)
    self.lbgc1.select_clear(0, "end")
    self.lbgc1.select_set(listind)
    self.lbgc1.see(listind)

    self.setDispCollsData(listind)
    self.fbutt42.configure(state="normal")
    self.lbgc2.focus()

    self.filestate = True
    return True

  def saveCollections(self):
    self.freas.configure(text=" ")

    fn = False
    fnasked = False
    if not self.datadir:
      self.datadir = os.path.curdir
    if not fnasked:
      fn = F.asksaveasfilename(title="collection defs", filetypes=[("hdf5 files", "*.h5"), ("all files", "*")], initialdir=self.datadir, confirmoverwrite=False)
    if not fn:
      self.finfo.configure(text=" ")
      return
    dn = os.path.dirname(fn)
    if not os.path.isdir(dn):
      self.finfo.configure(text=" no such directory ")
      return

    self.datafile = fn
    self.datadir = dn
    self.finfo.configure(text=" ")

    self.filestate = False
    self.saveCollectionsFile()
    if not self.filestate:
      self.finfo.configure(text=" save failed ")
      return

    self.finfo.configure(text=" data saved ")


  def saveCollectionsFile(self):
    fbase = self.datafile
    for suff, slen in [[".h5", 3], [".hdf", 4], [".hdf5", 5], [".csv", 4], [".txt", 4]]:
      if fbase.endswith(suff):
        fbase = fbase[:(-1 * slen)]
        break

    fbase0 = os.path.basename(fbase)
    if 0 == len(fbase0):
      self.freas.configure(text="no file-name")
      return False

    fnhdf = fbase + ".h5"
    fncsv = fbase + ".csv"
    fwname = fbase0 + ".h5"

    isav = True
    if os.path.exists(fnhdf):
      isav = False
    if os.path.exists(fncsv):
      isav = False

    curind = 0
    while not isav:
      isav = True
      curind += 1
      fnhdf = fbase + "-" + str(curind) + ".h5"
      fncsv = fbase + "-" + str(curind) + ".csv"
      fwname = fbase0 + "-" + str(curind) + ".h5"
      if os.path.exists(fnhdf):
        isav = False
      if os.path.exists(fncsv):
        isav = False

    collen = len(self.dispCollCases)
    if 0 == collen:
      return

    savenames = []
    savecolls = []
    dclen = 0

    tmpcolls = len(self.dispCases) * [0]
    withnames = False
    for collind in range(collen):
      for caseind in self.dispCollCases[collind]:
        tmpcolls[caseind] = collind + 1
        withnames = True

    if withnames:
      for ind in range(len(tmpcolls)):
        currank = tmpcolls[ind]
        if 0 < currank:
          savenames.append(self.dispCases[ind])
          savecolls.append(currank)
      dclen = len(savecolls)

    else:
      savenames.append("")
      savecolls.append(0)

    try:
      h5f = H.openFile(fnhdf, "w")

    except:
      self.freas.configure(text="can not create files")
      return False

    try:
      cdate = time.strftime("%Y-%m-%d %H:%M:%S", time.gmtime())
      onelist = [self.colftype, self.colversion, cdate, "CluDisc collection defs"]
      h5arr = h5f.createArray(h5f.root, "info", [item.encode("utf-8") for item in onelist], "data info")
      onelist = [dclen, collen]
      h5arr = h5f.createArray(h5f.root, "dims", onelist, "data dimensions")

      h5cases = h5f.createGroup(h5f.root, "cases", "data cases info")
      h5colls = h5f.createGroup(h5f.root, "colls", "data collections info")

      h5arr = h5f.createArray(h5cases, "names", [item.encode("utf-8") for item in savenames], "data names")
      h5arr = h5f.createArray(h5cases, "colls", savecolls, "collection groups")

      collranks = []
      for onecoll in self.dispColls:
        currank = onecoll.split()[-1]
        collranks.append(int(currank))
      h5arr = h5f.createArray(h5colls, "ranks", collranks, "collection ranks")

      collsizes = self.dispCollDiameters
      h5arr = h5f.createArray(h5colls, "sizes", collsizes, "collection sizes")

      collcolors = []
      for clrind in self.dispCollColors:
        curclr = self.rosacolors[clrind]
        collcolors.append(curclr)
      h5arr = h5f.createArray(h5colls, "colors", [item.encode("utf-8") for item in collcolors], "collection colors")

      collcomms = [(item if (item != self.collection_comment_placeholder) else "") for item in self.dispCollComments]
      h5arr = h5f.createArray(h5colls, "comms", [item.encode("utf-8") for item in collcomms], "collection comments")

      h5f.flush()
      h5f.close()

    except:
      h5f.close()
      self.freas.configure(text="can not write into files")
      return False

    fhcsv = []
    try:
      fhcsv = open(fncsv, "w")

    except:
      self.freas.configure(text="can not create files")
      return False

    try:
      cwcsv = csv.writer(fhcsv, delimiter=";", dialect='excel')
      cwcsv.writerow(["CluDisc collection defs"])

      for onecol in range(collen):
        cwcsv.writerow(self.dispColls[onecol].split())
        cursizeint = self.dispCollDiameters[onecol]
        cursize = str(cursizeint // 10) + "." + str(cursizeint % 10)
        curcolorind = self.dispCollColors[onecol]
        curcolor = self.rosacolors[curcolorind]
        cwcsv.writerow([cursize, curcolor])
        curcomm = self.dispCollComments[onecol]
        cwcsv.writerow([curcomm])
        for nmind in sorted(self.dispCollCases[onecol]):
          cwcsv.writerow([self.dispCases[nmind]])
        cwcsv.writerow([])

      fhcsv.close()

    except:
      fhcsv.close()
      self.freas.configure(text="can not write into files")
      return False

    self.freas.configure(text=fwname[:50])
    self.filestate = True
    return True


  def saveProjections(self):
    self.freas.configure(text=" ")

    if 0 == self.vspd.get():
      if 1 == self.vscd.get():
        if not self.saved:
          self.finfo.configure(text="save cluster-tree first")
          self.freas.configure(text="scale definition is cluster-tree add-on")
          return

    if [] == self.alg:
      self.finfo.configure(text="no data")
      return
    if [] == self.alg.projections:
      self.finfo.configure(text="no data")
      return
    if 1 == self.vspd.get():
      if [] == self.alg.getAdjusted():
        self.finfo.configure(text="no data")
        return

    inifn = "3d"
    namebase = ""
    dotind = self.sname.rfind(".")
    if not -1 == dotind:
      namebase = self.sname[:dotind]
    else:
      namebase = self.sname
    if not "" == namebase:
      inifn = namebase + "-3d"

    fn = False
    fnasked = False
    if not self.datadir:
      self.datadir = os.path.curdir
    if not fnasked:
      fn = F.asksaveasfilename(title="scaled data", filetypes=[("csv files", "*.csv"), ("all files", "*")], initialdir=self.datadir, initialfile=inifn, confirmoverwrite=False)
    if not fn:
      self.finfo.configure(text=" ")
      return
    dn = os.path.dirname(fn)
    if not os.path.isdir(dn):
      self.finfo.configure(text=" no such directory ")
      return

    self.datafile = fn
    self.datadir = dn
    self.finfo.configure(text=" ")

    self.filestate = False
    self.saveProjectionsFile()
    if not self.filestate:
      self.finfo.configure(text=" save failed ")
      return

    self.finfo.configure(text=" data saved ")

  def saveProjectionsFile(self):
    fbase = self.datafile
    for suff, slen in [[".h5", 3], [".hdf", 4], [".hdf5", 5], [".csv", 4], [".txt", 4]]:
      if fbase.endswith(suff):
        fbase = fbase[:(-1 * slen)]
        break

    fbase0 = os.path.basename(fbase)
    if 0 == len(fbase0):
      self.freas.configure(text="no file-name")
      return False

    saveh5 = False
    savecs = False
    dtype = self.vspd.get()
    if 0 == dtype:
      if 1 == self.vscd.get():
        saveh5 = True
    if 0 < len(self.dispColls):
      if 1 == self.vscc.get():
        savecs = True

    fnhdf = fbase + ".h5"
    fncsv = fbase + ".csv"
    fncol = fbase + "-colls.csv"
    fwname = fbase0 + ".csv"

    isav = True
    if os.path.exists(fnhdf):
      isav = False
    if os.path.exists(fncsv):
      isav = False
    if os.path.exists(fncol):
      isav = False

    curind = 0
    while not isav:
      isav = True
      curind += 1
      fnhdf = fbase + "-" + str(curind) + ".h5"
      fncsv = fbase + "-" + str(curind) + ".csv"
      fncol = fbase + "-" + str(curind) + "-colls.csv"
      fwname = fbase0 + "-" + str(curind) + ".csv"
      if os.path.exists(fnhdf):
        isav = False
      if os.path.exists(fncsv):
        isav = False
      if os.path.exists(fncol):
        isav = False

    sels = self.alg.getSelections()
    pars = self.alg.getParameters()
    vecs = self.alg.getEigenVectors()

    if saveh5:
      dims = [0, 0]
      if not [] == vecs:
        dims = [vecs.shape[0], vecs.shape[1]]

      try:
        h5f = H.openFile(fnhdf, "w")

      except:
        self.freas.configure(text="can not create files")
        return False

      try:
        cdate = time.strftime("%Y-%m-%d %H:%M:%S", time.gmtime())
        onelist = [self.psdftype, self.psdversion, self.sto.treeid, cdate, "CluDisc projection data"]
        h5arr = h5f.createArray(h5f.root, "info", [item.encode("utf-8") for item in onelist], "data info")

        h5tree = h5f.createGroup(h5f.root, "tree", "cluster tree selection info")
        onelist = [pars[0], self.project, self.resnorm]
        h5arr = h5f.createArray(h5tree, "seltype", onelist, "selection type")
        onelist = [len(sels)]
        for isel in sels:
          onelist.append(isel + 1)
        h5arr = h5f.createArray(h5tree, "selnodes", onelist, "selection nodes")

        h5base = h5f.createGroup(h5f.root, "base", "projection info")
        h5arr = h5f.createArray(h5base, "dimensions", dims, "basis dimensions")
        h5arr = h5f.createArray(h5base, "eigenvectors", vecs, "projection basis")

        h5scale = h5f.createGroup(h5f.root, "scale", "projection info")
        onelist = self.alg.getAverages()
        h5arr = h5f.createArray(h5scale, "averages", onelist, "scale averages")
        onelist = self.alg.getScales()
        h5arr = h5f.createArray(h5scale, "ratios", onelist, "scale ratios")

        h5f.flush()
        h5f.close()

      except:
        h5f.close()
        self.freas.configure(text="can not write into files")
        return False

    vecs = []
    psels = []
    for isel in sels:
      psels.append(str(isel + 1))
    ppars = ["", "", "", ""]
    ppars[0] = ["uninorm", "scaled"][pars[0]]
    ppars[1] = ["fail-safe:" + str(self.project), "exact"][pars[1]]
    ppars[2] = str(self.resnorm)
    ppars[3] = ["linear", "adjusted"][dtype]
    if 1 == dtype:
      ppars.append(str(pars[3]))
      item = "square"
      if 1 == self.extend:
        item = "linear"
      ppars.append(item)
      item = str(self.atenstart) + ":" + str(self.atenstay) + ":" + str(self.atencount)
      ppars.append(item)

    if 0 == dtype:
      vecs = self.alg.getProjections()
      if [] == vecs:
        return False

    else:
      vecs = N.asmatrix(self.alg.getAdjusted())
      if [] == vecs:
        return False

    dims = vecs.shape
    if not len(self.sto.dcnames) == dims[0]:
      return False
    if not 3 == dims[1]:
      return False

    origname = self.sto.fname

    fhcsv = []
    try:
      fhcsv = open(fncsv, "w")
    except:
      self.freas.configure(text="can not create files")
      return False

    try:
      cwcsv = csv.writer(fhcsv, delimiter=";", dialect='excel')
      cwcsv.writerow(["#CluDisc scaled data", origname, self.sto.treeid])
      cwcsv.writerow(["#parameters"] + ppars)
      cwcsv.writerow(["#nodes"] + psels)
      cwcsv.writerow("")

      for ind in range(dims[0]):
        cname = self.sto.dcnames[ind]
        onerow = vecs[ind]
        sval1 = "%(val)1.4e" % {'val': onerow[0,0]}
        sval2 = "%(val)1.4e" % {'val': onerow[0,1]}
        sval3 = "%(val)1.4e" % {'val': onerow[0,2]}
        onecsv = [cname, sval1, sval2, sval3]
        cwcsv.writerow(onecsv)

      fhcsv.close()

    except:
      fhcsv.close()
      self.freas.configure(text="can not write into files")
      return False

    collen = len(self.dispCollCases)
    if 0 == collen:
      savecs = False

    if savecs:

      fhcol = []
      try:
        fhcol = open(fncol, "w")
      except:
        self.freas.configure(text="can not create files")
        return False

      try:
        cwcol = csv.writer(fhcol, delimiter=";", dialect='excel')
        cwcol.writerow(["#CluDisc scaled data", origname, self.sto.treeid])
        cwcol.writerow(["#parameters"] + ppars)
        cwcol.writerow(["#nodes"] + psels)
        cwcol.writerow("")

        for onecol in range(collen):
          curcomm = "#" + self.dispCollComments[onecol]
          cwcol.writerow([curcomm])
          onerow = ["#collection", self.dispColls[onecol].split()[-1]]
          cursizeint = self.dispCollDiameters[onecol]
          cursize = str(cursizeint // 10) + "." + str(cursizeint % 10)
          curcolorind = self.dispCollColors[onecol]
          curcolor = self.rosacolors[curcolorind]
          onerow += ["size", cursize, "color", curcolor]
          cwcol.writerow(onerow)
          for nmind in sorted(self.dispCollCases[onecol]):
            cname = self.sto.dcnames[nmind]
            onerow = vecs[nmind]
            sval1 = "%(val)1.4e" % {'val': onerow[0,0]}
            sval2 = "%(val)1.4e" % {'val': onerow[0,1]}
            sval3 = "%(val)1.4e" % {'val': onerow[0,2]}
            onecsv = [cname, sval1, sval2, sval3]
            cwcol.writerow(onecsv)
          cwcol.writerow([])

        fhcol.close()

      except:
        fhcol.close()
        self.freas.configure(text="can not write into files")
        return False

    self.freas.configure(text=fwname[:50])
    self.filestate = True
    return True

  def prepTable(self):
    self.f2table = T.Frame(self.fbot)

    fc = T.Frame(self.f2table)
    fc.pack(side="top", expand=1, fill="y")
    T.Label(fc, text=" ", width=1, height=3).pack(side="top")

    fd = T.Frame(self.f2table)
    fd.pack(side="top", expand=0, fill="none")

    fe = T.Frame(self.f2table)
    fe.pack(side="top", expand=1, fill="y")
    T.Label(fe, text=" ", width=1, height=4).pack(side="top")

    ff = T.Frame(self.f2table)
    ff.pack(side="top", expand=0, fill="none")

    fg = T.Frame(self.f2table)
    fg.pack(side="top", expand=1, fill="y")
    T.Label(fg, text=" ", width=1, height=1).pack(side="top")

    fd1 = T.Frame(fd)
    fd2 = T.Frame(fd)
    fd3 = T.Frame(fd)
    fd1.pack(side="left")
    fd2.pack(side="left")
    fd3.pack(side="left")

    itlF = ("courier", 9, "normal italic")

    T.Label(fd1, text="Data cases", width=20, pady=4).grid(row=0, column=0)
    self.tdcselview = T.Label(fd1, text=" - columns\n", width=11, anchor="w", font=itlF)
    self.tdcselview.grid(row=1, column=0)

    foren = T.Frame(fd1)
    foren.grid(row=2, column=0, rowspan=1)

    self.vobjlog = T.IntVar()
    self.vobjlog.set(0)
    self.vobjwin = T.IntVar()
    self.vobjwin.set(0)
    self.vobjren = T.IntVar()
    self.vobjren.set(1)
    self.vobjrob = T.IntVar()
    self.vobjrob.set(0)

    self.cblog = T.Checkbutton(foren, text="logarithmize ", variable=self.vobjlog, command=self.updateDataType, pady=0, bd=1)
    self.cbwin = T.Checkbutton(foren, text="winsorize ", variable=self.vobjwin, command=self.updateDataType, pady=0, bd=1)
    self.cbren = T.Checkbutton(foren, text="normalize ", variable=self.vobjren, command=self.updateNormType, pady=0, disabledforeground="#000000", bd=1)
    T.Label(foren, text=" ", width=1).grid(row=1, column=1, sticky="w")
    self.cbrob = T.Checkbutton(foren, text="robust ", variable=self.vobjrob, pady=0, bd=1)
    self.cblog.grid(row=0, column=0, sticky="w")
    self.cbwin.grid(row=1, column=0, sticky="w")
    self.cbren.grid(row=0, column=1, sticky="w", columnspan=2)
    self.cbrob.grid(row=1, column=2, sticky="w")

    T.Label(fd1, text=" ", width=4).grid(row=0, column=1)

    T.Label(fd1, text="Properties", pady=4).grid(row=0, column=2)
    T.Label(fd1, text=" - to cluster\n", font=itlF).grid(row=1, column=2)

    fodir = T.Frame(fd1)
    fodir.grid(row=2, column=2, rowspan=1)

    self.vobjdir = T.Variable()
    self.vobjdir.set(2)
    rodir1 = T.Radiobutton(fodir, text="columns ", variable=self.vobjdir, value=1, command=self.updatePropSelection, pady=0, bd=1)
    rodir2 = T.Radiobutton(fodir, text="rows ", variable=self.vobjdir, value=2, command=self.updatePropSelection, pady=0, bd=1)
    rodir1.grid(row=1, column=0, sticky="w")
    rodir2.grid(row=2, column=0, sticky="w")

    T.Label(fd1, text=" ", width=6).grid(row=0, column=3)

    T.Label(fd1, text="Base clusters", pady=4).grid(row=0, column=4)
    T.Label(fd1, text=" - center count\n", font=itlF).grid(row=1, column=4)

    self.vreqbas = T.IntVar()
    inicount = self.basecount // 2
    if 3 > inicount:
      inicount = 3
    self.vreqbas.set(inicount)

    self.sbclcnt = T.Scale(fd1, from_=3, to=self.basecount, orient="horizontal", length=120, variable=self.vreqbas, bd=1)
    self.sbclcnt.grid(row=2, column=4)
    self.sbclcnt.bind("<1>", self.focusBaseScale)

    T.Label(fd2, text=" ", width=9).grid(row=0, column=0)

    self.bgclust = T.Button(fd3, text=" Cluster ", width=7, command=self.procClusteringExt)
    self.bgclust.grid(row=0, column=1)
    self.bgclust.configure(state="disabled")

    itlF = ("courier", 9, "normal italic")

    self.lgclust = T.Label(fd3, text=" \n none \n ", width=20, font=itlF)
    self.lgclust.grid(row=1, column=1)

    self.sto.show(ff)
    self.sto.addclgui(self.bgclust)

    self.cen.root = self.root
    self.cen.stat = self.lgclust

  def viewTable(self):
    if self.works:
      return
    if not 0 == self.rterror:
      return

    if self.fcurr == self.f2table:
      return

    self.fcurr.pack_forget()
    self.fcurr = self.f2table
    self.fcurr.pack(side="top", anchor="n", expand=1, fill="both")

    self.f1file.configure(relief="raised")
    self.f1table.configure(relief="sunken")
    self.f1clust.configure(relief="raised")
    self.f1state.configure(relief="raised")
    self.f1disp.configure(relief="raised")
    self.f1help.configure(relief="raised")

  def cleanTable(self):
    self.sto.clean()
    self.csvdata = []
    self.sto.setData(self.csvdata)
    self.bgclust.configure(state="disabled")
    self.lgclust.configure(text=" \n none \n ")

  def updateDataType(self):
    tolog = self.vobjlog.get()
    towin = self.vobjwin.get()
    if (1 == tolog) or (1 == towin):
      self.vobjren.set(1)
      self.cbren.configure(state="disabled")
      self.cbrob.configure(state="normal")
    else:
      self.cbren.configure(state="normal")

  def updateNormType(self):
    if 1 == self.vobjren.get():
      self.cbrob.configure(state="normal")
    else:
      self.vobjrob.set(0)
      self.cbrob.configure(state="disabled")

  def updatePropSelection(self):
    if 1 == self.vobjdir.get():
      self.tdcselview.configure(text=" - rows\n")
    else:
      self.tdcselview.configure(text=" - columns\n")

  def focusBaseScale(self, event):
    self.sbclcnt.focus()

  def procClusteringExt(self):
    try:
      self.procClustering()
    except:
      self.catchExcept()
      return

  def procClustering(self):
    if self.works:
      return

    if not self.disposed:
      self.bgclust.configure(text=" Cluster ")
      self.disposed = True
      return

    if not self.sto.mydt.rsel:
      return
    if not self.sto.mydt.csel:
      return

    self.lgclust.configure(text=" \n works \n ")

    rfrom = self.sto.mydt.rfrom
    rto = self.sto.mydt.rto
    rhead = self.sto.mydt.rhead

    cfrom = self.sto.mydt.cfrom
    cto = self.sto.mydt.cto
    chead = self.sto.mydt.chead

    rcnt = (rto - rfrom) + 1
    ccnt = (cto - cfrom) + 1
    if (3 > rcnt):
      self.lgclust.configure(text=" \n none \n ")
      return
    if (3 > ccnt):
      self.lgclust.configure(text=" \n none \n ")
      return

    self.works = True
    self.saved = False
    self.sname = ""
    self.alg = []
    self.selList = []
    self.bgclust.configure(fg="#8f8f8f")

    self.cen.clean()
    self.setLabel(self.sto.oname)

    maxhclust = self.vreqbas.get()
    self.cen.cencount = maxhclust

    mrows = 0
    mcols = 0
    trans = False
    if 1 == self.vobjdir.get():
      mrows = ccnt
      mcols = rcnt
      trans = True
    else:
      mrows = rcnt
      mcols = ccnt
    self.bykmeans = True
    if maxhclust >= mrows:
      self.bykmeans = False
    pmatrix = N.empty((mrows, mcols))

    coff = cfrom
    roff = rfrom
    vdef = N.nan
    vmis = 0

    self.sto.dcnames = []
    self.sto.prnames = []

    if trans:
      #data case names
      if -1 < chead:
        for irow in range(mrows):
          name = str(irow + 1) + "."
          try:
            readname = self.sto.csvdata[roff+irow][chead].strip()
            if 0 < len(readname):
              name = readname
          except:
            pass
          self.sto.dcnames.append(name)
      else:
        for irow in range(mrows):
          name = str(irow + 1) + "."
          self.sto.dcnames.append(name)

      #property names
      if -1 < rhead:
        for icol in range(mcols):
          name = str(icol + 1) + "."
          try:
            readname = self.sto.csvdata[rhead][coff+icol].strip()
            if 0 < len(readname):
              name = readname
          except:
            pass
          self.sto.prnames.append(name)
      else:
        for icol in range(mcols):
          name = str(icol + 1) + "."
          self.sto.prnames.append(name)

      #data values
      for irow in range(mrows):
        for icol in range(mcols):
          value = vdef
          try:
            value = float(self.sto.csvdata[coff+icol][roff+irow])
          except:
            vmis += 1
          pmatrix[irow][icol] = value

    else:
      #property names
      if -1 < chead:
        for irow in range(mrows):
          name = str(irow + 1) + "."
          try:
            readname = self.sto.csvdata[roff+irow][chead].strip()
            if 0 < len(readname):
              name = readname
          except:
            pass
          self.sto.prnames.append(name)
      else:
        for irow in range(mrows):
          name = str(irow + 1) + "."
          self.sto.prnames.append(name)

      #data case names
      if -1 < rhead:
        for icol in range(mcols):
          name = str(icol + 1) + "."
          try:
            readname = self.sto.csvdata[rhead][coff+icol].strip()
            if 0 < len(readname):
              name = readname
          except:
            pass
          self.sto.dcnames.append(name)
      else:
        for icol in range(mcols):
          name = str(icol + 1) + "."
          self.sto.dcnames.append(name)

      #data values
      for irow in range(mrows):
        for icol in range(mcols):
          value = vdef
          try:
            value = float(self.sto.csvdata[roff+irow][coff+icol])
          except:
            vmis += 1
          pmatrix[irow][icol] = value

    self.cen.pmatrix = pmatrix
    self.cen.mrows = mrows
    self.cen.mcols = mcols

    self.sto.dclog = 0
    self.sto.dcwin = 0
    self.sto.dcnor = 0
    self.sto.props = "rows"
    if trans:
      self.sto.props = "cols"
    self.sto.basec = maxhclust
    self.sto.infrows = [rhead, rfrom, rto]
    self.sto.infcols = [chead, cfrom, cto]

    self.sto.treeid = ""
    rndstr = str(time.time())
    for ind in range(20):
      rndstr += str(random.random())
    self.sto.treeid = digmod(rndstr.encode("utf-8")).hexdigest()

    if 1 == self.vobjwin.get():
      self.cen.towin = True
      self.sto.dcwin = 1
    else:
      self.cen.towin = False

    self.cen.dcrob = False
    if 1 == self.vobjlog.get():
      self.cen.tolog = True
      self.sto.dclog = 1
    else:
      self.cen.tolog = False

    if 1 == self.vobjren.get():
      self.cen.dcnor = True
      self.sto.dcnor = 1
      if 1 == self.vobjrob.get():
        self.cen.dcrob = True
        self.sto.dcnor = 2
    else:
      self.cen.dcnor = False

    self.lscale.configure(text=" \n none \n ")
    self.cls.empty()
    self.vtrinfo.set('')

    self.stateClean()

    if self.bykmeans:
      self.sto.clust = 1
    else:
      self.sto.clust = 0
      self.cen.kmruncycles = 0
      self.cen.kmconverged = 1

    self.cen.normalize()
    if 0 < self.cen.error:
      self.lgclust.configure(text=" \n wrong data \n ")
      self.clustNone()
      self.bgclust.configure(fg="#000000")
      self.works = False
      return

    if self.bykmeans:
      self.cen.kmeans()
      self.lgclust.configure(text=" \n at tree \n ")
      self.root.update()
      self.cls.fill(True, self.cen)
    else:
      self.cen.orients = N.ones(mrows, dtype=int)
      self.cen.grprops = N.empty(mrows, dtype=int)
      for ind in range(mrows):
        self.cen.grprops[ind] = ind
      self.cen.prepCenGroups(mrows)
      for ind in range(mrows):
        self.cen.cengrps[ind] = [ind]
      self.lgclust.configure(text=" \n at tree \n ")
      self.root.update()
      self.cls.fill(False, self.cen)

    self.bscale.configure(state="disabled")
    self.fbutt23.configure(state="disabled")
    self.fbutt31.configure(state="disabled")

    self.cen.release()

    if self.cls.filled:
      self.sto.fname = self.sto.oname
      self.sto.winstype = 0
      if "q" == self.cen.winstype:
        self.sto.winstype = 1
      self.sto.winsvalue = self.cen.winsvalue
      self.bs1.configure(state="normal")
      self.bs2.configure(state="normal")
      self.fbutt22.configure(state="normal")
      self.fbutt24.configure(state="normal")
    else:
      self.clustNone()
      if 0 == len(self.cls.errmsg):
        self.lgclust.configure(text=" \n no group \n ")
      else:
        errmsg = " \n" + self.cls.errmsg + "\n "
        self.lgclust.configure(text=errmsg)
        self.cls.errmsg = ""
      self.works = False
      self.bgclust.configure(fg="#000000")
      return

    self.stateGroups()
    self.stateAxes()
    self.setDispCases()
    self.lgclust.configure(text=" \n done \n ")
    self.bgclust.configure(fg="#000000")
    self.bgclust.configure(text=" Dispose ")
    self.disposed = False
    self.works = False

  def clustNone(self):
    self.cen.clean()

    self.fbutt22.configure(state="disabled")
    self.fbutt23.configure(state="disabled")
    self.fbutt24.configure(state="disabled")

    self.fbutt31.configure(state="disabled")

    self.bs1.configure(state="disabled")
    self.bs2.configure(state="disabled")
    self.bscale.configure(state="disabled")

  def prepClust(self):
    self.f2clust = T.Frame(self.fbot)

    T.Label(self.f2clust, text=" ").pack(side="top", expand=1, fill="y")
    fd = T.Frame(self.f2clust)
    fd.pack(side="top", expand=1, fill="y")
    T.Label(self.f2clust, text=" ").pack(side="top", expand=1, fill="y")
    fe = T.Frame(self.f2clust)
    fe.pack(side="bottom", anchor="s", expand=1, fill="x")

    self.cls.show(fe)

    fd1 = T.Frame(fd)
    fd2 = T.Frame(fd)
    fd3 = T.Frame(fd)
    fd4 = T.Frame(fd)
    fd5 = T.Frame(fd)
    fd1.grid(row=0, column=0, stick="sn")
    fd2.grid(row=0, column=1, stick="sn")
    fd3.grid(row=0, column=2, stick="sn")
    fd4.grid(row=0, column=3, stick="sn")
    fd5.grid(row=0, column=4, stick="sn")

    fd1.rowconfigure(0, weight=1)
    fd1.rowconfigure(2, weight=2)

    lb1 = T.Label(fd1, text="Select\n ", anchor="n")
    self.bs1 = T.Button(fd1, text="empty", bd=1, width=7, command=self.cls.emptyTreeSel)
    self.bs2 = T.Button(fd1, text="bottom", bd=1, width=7, command=self.cls.bottomTreeSel)

    lb1.grid(row=0, column=0, sticky="n")
    self.bs1.grid(row=1, column=0, sticky="swe")
    self.bs2.grid(row=2, column=0, sticky="nwe")

    self.bs1.configure(state="disabled")
    self.bs2.configure(state="disabled")

    i1 = T.Label(fd1, text=" ", width=4)
    i1.grid(row=1, column=1)

    fF = T.Frame(fd2)
    fF.pack(side="top", expand="t", fill="x", padx=5, pady=4)

    self.vtrinfo = T.Variable()
    self.lbF1 = T.Listbox(fF, listvariable=self.vtrinfo, height=5, selectmode="extended", bd=1)

    self.vtrinfo.set('')

    self.lbF1.pack(side="left", fill="both", expand=1)

    self.lbF1.bind("<1>", self.focusTargList)

    i2 = T.Label(fd3, text=" ", width=4)
    i2.grid(row=0, column=0)

    self.vgroup = T.IntVar()
    self.vgroup.set(1)
    self.vexact = T.IntVar()
    self.vexact.set(1)
    if 0 < self.project:
      self.vexact.set(0)
    self.vadjust = T.IntVar()
    self.vadjust.set(0)

    smlF = ("courier", 6, "normal italic")
    lb2 = T.Label(fd3, text="Axes")
    lb2.grid(row=0, column=1, sticky="n")
    T.Label(fd3, text=" \n ", font=smlF).grid(row=1, column=1, sticky="s")

    fd3.rowconfigure(2, weight=2)
    fcltype = T.Frame(fd3)
    fcltype.grid(row=2, column=1, rowspan=2, sticky="n")

    T.Label(fcltype, text="initial: ", pady=0).grid(row=0, column=0, sticky="e")
    T.Label(fcltype, text="final: ", pady=0).grid(row=1, column=0, sticky="e")
    self.vgr1 = T.Checkbutton(fcltype, text="scaled ", width=8, anchor="w", variable=self.vgroup, command=self.updateAxesInitial, pady=0, bd=1)
    self.vgr3 = T.Checkbutton(fcltype, text="linear ", width=8, anchor="w", variable=self.vadjust, command=self.updateAxesFinal, pady=0, bd=1)

    self.vgr1.grid(row=0, column=1, sticky="w")
    self.vgr3.grid(row=1, column=1, sticky="w")

    self.adjloc = T.IntVar()
    self.adjloc.set(10)
    T.Label(fd4, text=" ", width=4).grid(row=0, column=0)
    T.Label(fd4, text="Locality\n").grid(row=0, column=1)
    self.scloc = T.Scale(fd4, from_=5, to=50, length=100, resolution=5, orient="horizontal", variable=self.adjloc, bd=1)
    self.scloc.grid(row=1, column=1)
    T.Label(fd4, text="% of distances", height=1).grid(row=2, column=1)

    self.scloc.bind("<1>", self.focusLocScale)

    T.Label(fd5, text=" ", width=6).grid(row=0, column=0)
    self.bscale = T.Button(fd5, text=" Project ", command=self.procProjectingExt)
    self.bscale.grid(row=1, column=1)
    self.bscale.configure(state="disabled")

    itlF = ("courier", 9, "normal italic")

    self.lscale = T.Label(fd5, text=" \n none \n ", width=10, font=itlF)
    self.lscale.grid(row=2, column=1)

    self.cls.setclgui(self.vtrinfo, self.bscale)

  def viewClust(self):
    if self.works:
      return
    if not 0 == self.rterror:
      return

    if self.fcurr == self.f2clust:
      return

    self.fcurr.pack_forget()
    self.fcurr = self.f2clust
    self.fcurr.pack(side="top", anchor="n", expand=1, fill="both")

    self.f1file.configure(relief="raised")
    self.f1table.configure(relief="raised")
    self.f1clust.configure(relief="sunken")
    self.f1state.configure(relief="raised")
    self.f1disp.configure(relief="raised")
    self.f1help.configure(relief="raised")

    if not [] == self.cls.view:
      self.cls.view.updateListInfo()
      self.cls.view.csbx.focus()

  def cleanClust(self):
    self.cls.empty()
    self.vtrinfo.set('')
    self.bs1.configure(state="disabled")
    self.bs2.configure(state="disabled")
    self.bscale.configure(state="disabled")
    self.lscale.configure(text=" \n none \n ")
    self.alg = []
    self.selList = []

  def focusTargList(self, event):
    self.lbF1.focus()

  def focusLocScale(self, event):
    self.scloc.focus()

  def updateAxesInitial(self):
    if 1 == self.vgroup.get():
      self.vgr1.configure(text="scaled ")
    else:
      self.vgr1.configure(text="uninorm ")

  def updateAxesMiddle(self):
    if 1 == self.vexact.get():
      self.vgr2.configure(text="exact ")
    else:
      self.vgr2.configure(text="fail-safe ")

  def updateAxesFinal(self):
    if 1 == self.vadjust.get():
      self.vgr3.configure(text="adjusted ")
    else:
      self.vgr3.configure(text="linear ")
    self.dispRosaAutoLinadj()

  def procProjectingExt(self):
    try:
      self.procProjecting()
    except:
      self.catchExcept()
      return

  def procProjecting(self):
    self.lscale.configure(text=" \n works \n ")

    sels = self.cls.getSelections()
    grcnt = len(sels)
    if 3 > grcnt:
      self.lscale.configure(text=" \n none \n ")
      return

    dccnt = len(self.sto.dcnames)
    dscaled = self.vgroup.get()
    dexact = self.vexact.get()
    dadjust = self.vadjust.get()
    dlocal = self.adjloc.get()

    nochange = False
    toproject = True
    contained = False
    if not [] == self.alg:
      nochange = True
      toproject = False
      oldsels = self.alg.getSelections()
      oldpars = self.alg.getParameters()
      oscaled = oldpars[0]
      oexact = oldpars[1]
      oadjust = oldpars[2]
      olocal = oldpars[3]
      if not sels == oldsels:
        nochange = False
        toproject = True
      if not dscaled == oscaled:
        nochange = False
        toproject = True
      if not dexact == oexact:
        nochange = False
        toproject = True
      if not dadjust == oadjust:
        nochange = False
      if not dlocal == olocal:
        nochange = False
      if not toproject:
        if not nochange:
          if 0 == dadjust:
            contained = True

    if nochange:
      self.lscale.configure(text=" \n same \n ")
      return

    if contained:
      self.lscale.configure(text=" \n part \n ")
      return

    self.works = True
    curparams = {"scaled": dscaled, "adjust": dadjust, "local": dlocal, "sels": sels}
    curparams["type"] = "projected"
    self.selList.insert(0, curparams)
    self.updateSelsList()
    self.lbevH.select_clear(0, "end")
    self.lbevH.see(0)
    self.lbevH.activate(0)

    if toproject:
      self.alg = []

      tnodes = self.cls.tree.tnodes
      tsizes = self.cls.tree.amounts
      tvarts = self.cls.tree.variats

      scaledata = N.empty((grcnt,dccnt))
      groupsizes = N.empty(grcnt)
      groupvarts = N.empty(grcnt)

      for ind in range(grcnt):
        grind = sels[ind]
        scaledata[ind] = tnodes[grind]
        groupsizes[ind] = tsizes[grind]
        groupvarts[ind] = tvarts[grind]

      if 1 == dscaled:
        #scaled, multiply with sqrt(size)
        for ind in range(grcnt):
          vscale = N.sqrt(groupsizes[ind])
          scaledata[ind] *= vscale
      else:
        #uninorm, divide by sqrt(variation)
        for ind in range(grcnt):
          vscale = N.sqrt(groupvarts[ind])
          if 0.0 < vscale:
            scaledata[ind] /= vscale

      algrtpars = [self.project, self.resnorm, self.extend]
      algrtpars += [self.atenstart, self.atenstay, self.atencount]
      self.alg = DataScale(scaledata, sels, [dscaled, dexact, dadjust, dlocal], algrtpars, isclibcd)
      self.alg.compEigenVecs()

      if 0 == self.alg.error:
        self.alg.compProjections()

    else:
      self.alg.updateLocals([dadjust, dlocal])

    if 0 == self.alg.error:
      if not [] == self.alg.getAdjusted():
        self.alg.releaseAdjusted()

    if 0 == self.alg.error:
      if 1 == dadjust:
        self.alg.localAdjust()

    if not 0 == self.alg.error:
      self.alg = []

    self.stateAxesVals()
    self.stateAxialSels(0)
    if 1 == dadjust:
      self.sttype.set(1)
      self.stateType()
    else:
      self.sttype.set(0)
      self.stateType()

    if not 0 == self.alg.error:
      self.lscale.configure(text=" \n failed \n ")
      self.fbutt31.configure(state="disabled")

    else:
      self.dispRosaAutoProj()
      self.lscale.configure(text=" \n done \n ")
      self.fbutt31.configure(state="normal")

    self.works = False

  def prepState(self):
    self.f2state = T.Frame(self.fbot)

    T.Label(self.f2state, text=" ").pack(side="top", expand=1, fill="y")
    fd = T.Frame(self.f2state)
    fd.pack(side="top", expand=1, fill="y")
    fe = T.Frame(self.f2state)
    fe.pack(side="top", expand=0, fill="none")
    T.Label(self.f2state, text=" ").pack(side="top", expand=1, fill="y")

    fgr0 = T.LabelFrame(fd, text="Tree groups content")
    fgr0.grid(row=0, column=0, sticky="sn")
    T.Label(fd, text=" ", width=15).grid(row=0, column=1)
    fev0 = T.LabelFrame(fd, text="Scaled axes values")
    fev0.grid(row=0, column=2, sticky="sn")

    fd.rowconfigure(0, uniform=1)

    fgr = T.Frame(fgr0)
    fgr.grid(row=1, column=1)
    T.Label(fgr0, text=" ", width=2, height=1).grid(row=0, column=0)
    T.Label(fgr0, text=" ", width=2, height=1).grid(row=2, column=0)
    T.Label(fgr0, text=" ", width=2, height=1).grid(row=0, column=2)
    T.Label(fgr0, text=" ", width=2, height=1).grid(row=2, column=2)

    fev = T.Frame(fev0)
    fev.grid(row=1, column=1, sticky="sn")
    T.Label(fev0, text=" ", width=2, height=1).grid(row=0, column=0)
    T.Label(fev0, text=" ", width=2, height=1).grid(row=2, column=0)
    T.Label(fev0, text=" ", width=2, height=1).grid(row=0, column=2)
    T.Label(fev0, text=" ", width=2, height=1).grid(row=2, column=2)
    fev0.rowconfigure(1, weight=2)

    T.Label(fgr, text=" ", width=1).grid(row=1, column=1)

    fgr00 = T.Frame(fgr)
    fgr00.grid(row=0, column=0, sticky="n")
    self.vgr1Iinfo = T.Variable()
    self.sbgr1I = T.Scrollbar(fgr00, orient="vertical", bd=1)
    self.lbgr1I = T.Listbox(fgr00, listvariable=self.vgr1Iinfo, height=12, width=8, selectmode="browse", bd=1)

    self.lbgr1I.configure(yscrollcommand=self.sbgr1I.set)
    self.sbgr1I.configure(command=self.lbgr1I.yview)

    self.sbgr1I.grid(row=0, column=0, sticky="sn")
    self.lbgr1I.grid(row=0, column=1, sticky="sn")

    self.sbgr1I.bind("<1>", self.focusGroupList)
    self.lbgr1I.bind("<1>", self.focusGroupList)

    self.lbgr1I.bind("<KeyRelease>", self.focusGroupPartInn)
    self.lbgr1I.bind("<ButtonRelease>", self.focusGroupPartInn)

    fgr10 = T.Frame(fgr)
    fgr10.grid(row=2, column=0, sticky="n")
    self.vgr1Binfo = T.Variable()
    self.sbgr1B = T.Scrollbar(fgr10, orient="vertical", bd=1)
    self.lbgr1B = T.Listbox(fgr10, listvariable=self.vgr1Binfo, height=12, width=8, selectmode="browse", bd=1)

    self.lbgr1B.configure(yscrollcommand=self.sbgr1B.set)
    self.sbgr1B.configure(command=self.lbgr1B.yview)

    self.sbgr1B.grid(row=0, column=0, sticky="sn")
    self.lbgr1B.grid(row=0, column=1, sticky="sn")

    self.sbgr1B.bind("<1>", self.focusGroupList)
    self.lbgr1B.bind("<1>", self.focusGroupList)

    self.lbgr1B.bind("<KeyRelease>", self.focusGroupPartBot)
    self.lbgr1B.bind("<ButtonRelease>", self.focusGroupPartBot)

    fgr01 = T.Frame(fgr)
    fgr01.grid(row=0, column=2, sticky="sn")
    self.vgr2Iinfo = T.Variable()
    self.vgr2Iinfo.set(('group: --', ''))
    self.lbgr2I = T.Listbox(fgr01, listvariable=self.vgr2Iinfo, height=12, width=15, selectmode="extended", bd=1)
    self.sbgr2Ii = T.Scrollbar(fgr01, orient="horizontal", bd=1)

    self.sbgr2I = T.Scrollbar(fgr01, orient="vertical", bd=1)

    self.lbgr2I.configure(yscrollcommand=self.sbgr2I.set)
    self.lbgr2I.configure(xscrollcommand=self.sbgr2Ii.set)
    self.sbgr2I.configure(command=self.lbgr2I.yview)

    self.lbgr2I.grid(row=0, column=0, sticky="sn")
    self.sbgr2I.grid(row=0, column=1, sticky="sn")

    self.sbgr2I.bind("<1>", self.focusGroupList)
    self.lbgr2I.bind("<1>", self.focusGroupList)

    fgr11 = T.Frame(fgr)
    fgr11.grid(row=2, column=2, sticky="sn")
    self.vgr2Binfo = T.Variable()
    self.vgr2Binfo.set(('group: --', ''))
    self.lbgr2B = T.Listbox(fgr11, listvariable=self.vgr2Binfo, height=12, width=15, selectmode="extended", bd=1)
    self.sbgr2Bi = T.Scrollbar(fgr11, orient="horizontal", bd=1)

    self.sbgr2B = T.Scrollbar(fgr11, orient="vertical", bd=1)

    self.lbgr2B.configure(yscrollcommand=self.sbgr2B.set)
    self.lbgr2B.configure(xscrollcommand=self.sbgr2Bi.set)
    self.sbgr2B.configure(command=self.lbgr2B.yview)
    self.sbgr2Bi.configure(command=self.lbgr2B.xview)

    self.lbgr2B.grid(row=0, column=0, sticky="sn")
    self.sbgr2Bi.grid(row=1, column=0, sticky="we")
    self.sbgr2B.grid(row=0, column=1, sticky="sn")

    self.sbgr2B.bind("<1>", self.focusGroupList)
    self.sbgr2Bi.bind("<1>", self.focusGroupList)
    self.lbgr2B.bind("<1>", self.focusGroupList)

    T.Label(fev, text="selections   ").grid(row=0, column=0)
    T.Label(fev, text=" ", width=2).grid(row=2, column=1)

    self.sttype = T.IntVar()
    self.sttype.set(1)
    self.cbtype = T.Checkbutton(fev, text="adjusted", width=8, anchor="w", variable=self.sttype, command=self.stateType, bd=1)
    self.cbtype.grid(row=0, column=2, sticky="n")
    self.cbtype.configure(disabledforeground="#000000")

    fhv = T.Frame(fev)
    fhv.grid(row=1, column=0)

    self.vevHist = T.Variable()
    self.vevHist.set(())
    self.lbevH = T.Listbox(fhv, listvariable=self.vevHist, height=5, width=12, selectmode="browse", bd=1)
    self.sbevH = T.Scrollbar(fhv, orient="vertical", bd=1)

    self.lbevH.configure(yscrollcommand=self.sbevH.set)
    self.sbevH.configure(command=self.lbevH.yview)

    self.lbevH.grid(row=0, column=0, sticky="sn")
    self.sbevH.grid(row=0, column=1, sticky="sn")

    self.sbevH.bind("<1>", self.focusEigenList)
    self.lbevH.bind("<1>", self.focusEigenList)

    self.lbevH.bind("<KeyRelease>", self.updateAxialSels)
    self.lbevH.bind("<ButtonRelease>", self.updateAxialSels)

    self.vev2info = T.Variable()
    self.vev2info.set(('ID: --', 's: --', '1: --', '2: --', '3: --'))
    self.lbev2 = T.Listbox(fev, listvariable=self.vev2info, height=5, width=15, selectmode="extended", bd=1)
    self.lbev2.bind("<1>", self.focusEigenList)

    self.lbev2.grid(row=1, column=2, sticky="we")

    fsel = T.Frame(fev)
    fsel.grid(row=3, column=0, sticky="snew")

    self.vev3info = T.Variable()
    self.vev3info.set(('current:', '  initial axes', '  locality: %', 'groups:'))
    self.lbev3 = T.Listbox(fsel, listvariable=self.vev3info, height=16, width=12, selectmode="extended", bd=1)

    self.sbev3 = T.Scrollbar(fsel, orient="vertical", bd=1)

    self.lbev3.configure(yscrollcommand=self.sbev3.set)
    self.sbev3.configure(command=self.lbev3.yview)

    self.lbev3.pack(side="left", fill="both", expand=1)
    self.sbev3.pack(side="left", fill="y", expand=1)

    self.lbev3.bind("<1>", self.focusEigenList)
    self.sbev3.bind("<1>", self.focusEigenList)

    self.vev1info = T.Variable()
    self.vev1info.set('')
    fevi = T.Frame(fev)
    fevi.grid(row=3, column=2, sticky="snew")

    self.lbev1 = T.Listbox(fevi, listvariable=self.vev1info, height=15, width=15, selectmode="browse", bd=1)
    self.sbev1i = T.Scrollbar(fevi, orient="horizontal", bd=1)
    self.sbev1 = T.Scrollbar(fevi, orient="vertical", bd=1)

    self.lbev1.configure(yscrollcommand=self.sbev1.set)
    self.lbev1.configure(xscrollcommand=self.sbev1i.set)
    self.sbev1.configure(command=self.lbev1.yview)
    self.sbev1i.configure(command=self.lbev1.xview)

    self.lbev1.grid(row=0, column=0, sticky="sn")
    self.sbev1i.grid(row=1, column=0, sticky="we")
    self.sbev1.grid(row=0, column=1, sticky="sn")
    fevi.rowconfigure(0, weight=2)

    self.sbev1.bind("<1>", self.focusEigenList)
    self.lbev1.bind("<1>", self.focusEigenList)
    self.sbev1i.bind("<1>", self.focusEigenList)

    self.lbev1.bind("<KeyRelease>", self.focusEigenPart)
    self.lbev1.bind("<ButtonRelease>", self.focusEigenPart)

    fev.columnconfigure(2, uniform=1)
    fev.rowconfigure(3, weight=2)

  def viewState(self):
    if self.works:
      return
    if not 0 == self.rterror:
      return

    if self.fcurr == self.f2state:
      return

    self.fcurr.pack_forget()
    self.fcurr = self.f2state
    self.fcurr.pack(side="top", anchor="n", expand=1, fill="both")

    self.f1file.configure(relief="raised")
    self.f1table.configure(relief="raised")
    self.f1clust.configure(relief="raised")
    self.f1state.configure(relief="sunken")
    self.f1disp.configure(relief="raised")
    self.f1help.configure(relief="raised")

  def focusGroupList(self, event):
    event.widget.focus()

  def focusGroupPartInn(self, event):
    if [] == self.cls.tree:
      self.vgr2Iinfo.set(('group: --', ''))
      return

    lcount = self.cls.tree.lcount
    tcount = (2 * lcount) - 1

    cursel = self.lbgr1I.curselection()
    if 0 == len(cursel):
      self.vgr2Iinfo.set(('group: --', ''))
      return
    curind = int(cursel[0])
    curind += lcount

    if (curind <= lcount) or (curind > tcount):
      self.vgr2Iinfo.set(('group: --', ''))
      return

    curtup = ("group: " + str(curind), "")
    leaves = self.cls.tree.getNodeLists(curind - 1)
    curtup += tuple(leaves)
    self.vgr2Iinfo.set(curtup)

  def focusGroupPartBot(self, event):
    if [] == self.cls.tree:
      self.vgr2Binfo.set(('group: --', ''))
      return

    lcount = self.cls.tree.lcount
    tcount = (2 * lcount) - 1

    cursel = self.lbgr1B.curselection()
    if 0 == len(cursel):
      self.vgr2Binfo.set(('group: --', ''))
      return
    curind = int(cursel[0])

    if (0 == curind) or (curind > lcount):
      self.vgr2Binfo.set(('group: --', ''))
      return

    if (curind <= lcount):
      curgrp = self.cen.cengrps[curind - 1]
      curlen = len(curgrp)
      gnames = []
      if 0 < curlen:
        gnames = curlen * [""]
        for ind in range(curlen):
          grind = curgrp[ind]
          orirank = self.cen.orients[grind] + 1
          orisign = ["-", "", "+"][orirank]
          gnames[ind] = orisign + self.sto.prnames[grind]

      curtup = ("group: " + str(curind), "")
      curtup += tuple(gnames)
      self.vgr2Binfo.set(curtup)

  def updateSelsList(self):
    if [] == self.selList:
      self.vevHist.set(())
      return

    ilast = 0
    curHist = []
    for cursel in self.selList:
      if "projected" == cursel["type"]:
        if 0 == ilast:
          curHist.append("current")
        else:
          curHist.append(str(ilast) + ". last")
        ilast += 1
      else:
        curHist.append("loaded")

    self.vevHist.set(tuple(curHist))

  def focusEigenList(self, event):
    event.widget.focus()

  def focusEigenPart(self, event):
    vproj = ["ID: --", "s: --", "1: --", "2: --", "3: --"]

    if [] == self.alg:
      self.vev2info.set(tuple(vproj))
      return
    if 0 < self.alg.error:
      self.vev2info.set(tuple(vproj))
      return
    if [] == self.alg.projections:
      self.vev2info.set(tuple(vproj))
      return

    curtype = self.sttype.get()

    cursel = self.lbev1.curselection()
    if 0 == len(cursel):
      return
    curind = int(cursel[0])
    if curind >= self.alg.projections.shape[0]:
      self.vev2info.set(tuple(vproj))
      return
    if curind >= len(self.sto.dcnames):
      self.vev2info.set(tuple(vproj))
      return

    cname = self.sto.dcnames[curind]
    vproj[0] = "ID: " + str(cname)

    if 0 == curtype:
      projections = self.alg.getProjections()
      for ind in range(3):
        fval = projections[curind, ind]
        if 0.0 > fval:
          sval = "%(val)1.4e" % {'val': fval}
        else:
          sval = " %(val)1.4e" % {'val': fval}
        vproj[ind+2] = str(ind+1) + ": " + sval
      vsize = self.alg.scales[curind]
      if N.isfinite(vsize):
        vproj[1] = "s:  %(val)1.4e" % {'val': vsize}
      else:
        vproj[1] = "s:  " + str(vsize)

    else:
      adjusted = self.alg.getAdjusted()
      if not [] == adjusted:
        for ind in range(3):
          fval = adjusted[curind, ind]
          if 0.0 > fval:
            sval = "%(val)1.4e" % {'val': fval}
          else:
            sval = " %(val)1.4e" % {'val': fval}
          vproj[ind+2] = str(ind+1) + ": " + sval
        vsize = self.alg.adjmoved[curind]
        if N.isfinite(vsize):
          vproj[1] = "s:  %(val)1.4e" % {'val': vsize}
        else:
          vproj[1] = "s:  " + str(vsize)

    self.vev2info.set(tuple(vproj))

  def stateClean(self):
    self.vgr1Iinfo.set('')
    self.vgr2Iinfo.set(('group: --', ''))
    self.vgr1Binfo.set('')
    self.vgr2Binfo.set(('group: --', ''))
    self.stateCleanAxes()

  def stateCleanAxes(self):
    self.vev1info.set('')
    self.stateAxesVals()

  def stateGroups(self):
    self.vgr1Iinfo.set(())
    self.vgr1Binfo.set(())
    self.vgr2Iinfo.set(('group: --', ''))
    self.vgr2Binfo.set(('group: --', ''))

    curtuple = ()
    if not [] == self.cls.tree:
      lcount = self.cls.tree.lcount
      tcount = (2 * lcount) - 1
      curtupleI = ('inner',)
      curtupleI += tuple(range(lcount+1, tcount+1))
      curtupleB = ('bottom',)
      curtupleB += tuple(range(1, lcount+1))
      self.vgr1Iinfo.set(curtupleI)
      self.vgr1Binfo.set(curtupleB)

  def stateAxes(self):
    self.stateAxesVals()
    self.vev1info.set(tuple(self.sto.dcnames))

  def stateAxesVals(self):
    self.vev2info.set(('ID: --', 's: --', '1: --', '2: --', '3: --'))
    self.vev3info.set(('current:', '  initial axes', '  locality: %', 'groups:'))
    self.updateSelsList()

  def updateAxialSels(self, event):
    rank = 0
    cursels = self.lbevH.curselection()
    if 0 < len(cursels):
      rank = int(cursels[0])

    self.stateAxialSels(rank)

  def stateAxialSels(self, rank):
    if 0 == len(self.selList):
      return

    if len(self.selList) <= rank:
      self.vev3info.set(())
      return

    curparams = self.selList[rank]
    dscaled = curparams["scaled"]
    dadjust = curparams["adjust"]
    dlocal = curparams["local"]
    sels = curparams["sels"]

    if not dscaled in [0, 1]:
      return

    if 1 == dscaled:
      self.vgroup.set(1)
      self.vgr1.configure(text="scaled ")
    else:
      self.vgroup.set(0)
      self.vgr1.configure(text="uninorm ")

    if 1 == dadjust:
      self.vadjust.set(1)
      self.vgr3.configure(text="adjusted ")
      self.adjloc.set(dlocal)
    else:
      self.vadjust.set(0)
      self.vgr3.configure(text="linear ")

    self.cls.emptyTreeSel()
    for node in sels:
      self.cls.selectOne(node)

    if not type([]) == type(sels):
      return
    valini = ['  uninorm', '  scaled'][dscaled]
    strloc = "  linear"
    if 1 == dadjust:
      strloc = "  " + str(dlocal) + "% local"

    msgrank = str(self.lbevH.get(rank)) + ":"
    curtup = (msgrank, valini, strloc, "groups:")

    curlen = len(sels)
    if 0 < curlen:
      selnew = curlen * [""]
      for ind in range(curlen):
        curind = 1 + sels[ind]
        selnew[ind] = str(curind)
      curtup += tuple(selnew)

    self.vev3info.set(curtup)

  def stateType(self):
    curtype = self.sttype.get()
    if 1 == curtype:
      self.cbtype.configure(text="adjusted")
    else:
      self.cbtype.configure(text=" linear")

    rank = int(self.lbev1.index("active"))
    self.lbev1.select_clear(0, "end")
    if 0 <= rank:
      self.lbev1.select_set(rank)

    self.focusEigenPart(None)

  def prepDisp(self):
    self.f2disp = T.Frame(self.fbot)

    self.dispCases = []
    self.dispCaseRanks = []
    self.dispColls = []
    self.dispCollCases = []
    self.dispCollDiameters = []
    self.dispCollColors = []
    self.dispCollComments = []

    self.dispHigh = -1
    self.dispHighType = 0
    self.dispHighName = T.Variable()
    self.dispHighName.set("")

    fitems = T.Frame(self.f2disp)
    fitems.pack(side="top", expand=1, fill="y")

    fd = T.Frame(fitems)
    fd.pack(side="left")
    T.Label(fitems, text=" ", width=5).pack(side="left")
    fe = T.Frame(fitems)
    fe.pack(side="left")

    fdc0 = T.Frame(fd)
    fdc0.pack(side="left")
    T.Label(fd, text=" ", width=15).pack(side="left")

    fdc = T.Frame(fdc0)
    fdc.grid(row=1, column=1, rowspan=5, sticky="sn")
    T.Label(fdc0, text=" ", width=2, height=1).grid(row=0, column=0)
    T.Label(fdc0, text="   Free cases", height=1).grid(row=0, column=1)
    T.Label(fdc0, text=" ", width=2, height=1).grid(row=6, column=0)
    T.Label(fdc0, text=" ", width=2, height=1).grid(row=0, column=2)
    T.Label(fdc0, text=" ", width=2, height=1).grid(row=6, column=2)
    T.Label(fdc0, text="Collections   ", height=1).grid(row=0, column=3)
    T.Label(fdc0, text=" ", width=2, height=1).grid(row=6, column=4)

    self.vdc1info = T.Variable()
    self.sbdc1 = T.Scrollbar(fdc, orient="vertical", bd=1)
    self.lbdc1 = T.Listbox(fdc, listvariable=self.vdc1info, height=25, width=15, selectmode="extended", bd=1)

    self.lbdc1.configure(yscrollcommand=self.sbdc1.set)
    self.sbdc1.configure(command=self.lbdc1.yview)

    self.sbdc1.pack(side="left", fill="both", expand=1)
    self.lbdc1.pack(side="left", fill="both", expand=1)

    self.sbdc1.bind("<1>", self.focusDispList)
    self.lbdc1.bind("<1>", self.focusDispList)

    self.lbdc1.bind("<Double-Button-1>", self.focusCasesPart1)

    fgcS = T.Frame(fdc0)
    fgcS.grid(row=1, column=3, rowspan=1, sticky="n")

    fgcE = T.Frame(fdc0)
    fgcE.grid(row=2, column=3, rowspan=1, sticky="sn")

    fgcD = T.Frame(fdc0)
    fgcD.grid(row=3, column=3, rowspan=1, sticky="s")

    fgcC = T.Frame(fdc0)
    fgcC.grid(row=4, column=3, rowspan=1, sticky="n")

    self.vgc1info = T.Variable()
    self.lbgc1 = T.Listbox(fgcS, listvariable=self.vgc1info, height=5, width=15, selectmode="browse", bd=1)
    self.sbgc1 = T.Scrollbar(fgcS, orient="vertical", bd=1)

    self.lbgc1.configure(yscrollcommand=self.sbgc1.set)
    self.sbgc1.configure(command=self.lbgc1.yview)

    self.lbgc1.pack(side="left", fill="both", expand=1)
    self.sbgc1.pack(side="left", fill="both", expand=1)

    self.sbgc1.bind("<1>", self.focusDispList)
    self.lbgc1.bind("<1>", self.focusDispList)

    self.lbgc1.bind("<KeyRelease>", self.focusCollsPart)
    self.lbgc1.bind("<ButtonRelease>", self.focusCollsPart)

    self.vgc2info = T.Variable()
    self.lbgc2 = T.Listbox(fgcE, listvariable=self.vgc2info, height=10, width=15, selectmode="extended", bd=1)
    self.sbgc2 = T.Scrollbar(fgcE, orient="vertical", bd=1)

    self.lbgc2.configure(yscrollcommand=self.sbgc2.set)
    self.sbgc2.configure(command=self.lbgc2.yview)

    self.lbgc2.pack(side="left", fill="both", expand=1)
    self.sbgc2.pack(side="left", fill="both", expand=1)

    self.sbgc2.bind("<1>", self.focusDispList)
    self.lbgc2.bind("<1>", self.focusDispList)

    self.lbgc2.bind("<Double-Button-1>", self.focusCasesPart2)

    self.vcoldia = T.IntVar()
    self.fcoldia = T.Variable()
    self.vcoldia.set(5)
    self.fcoldia.set("0.5")
    self.scoldia = T.Scale(fgcD, from_=0.1, to=2.0, length=100, orient="horizontal", resolution=0.1, bd=1)
    self.scoldia.configure(variable=self.fcoldia, command=self.dispSetVColDia)
    self.scoldia.grid(row=0, column=0)

    self.scoldia.bind("<1>", self.focusDispList)

    self.vgc3info = T.Variable()
    self.rosacolors = ("red", "green", "blue", "yellow")
    self.rosacolors += ("magenta", "cyan", "orange", "steel")
    self.vgc3info.set(self.rosacolors)
    self.lbgc3 = T.Listbox(fgcC, listvariable=self.vgc3info, height=5, width=15, selectmode="browse", bd=1)
    self.sbgc3 = T.Scrollbar(fgcC, orient="vertical", bd=1)
    self.lbgc3.activate(0)

    self.lbgc3.configure(yscrollcommand=self.sbgc3.set)
    self.sbgc3.configure(command=self.lbgc3.yview)

    self.lbgc3.pack(side="left", fill="both", expand=1)
    self.sbgc3.pack(side="left", fill="both", expand=1)

    self.sbgc3.bind("<1>", self.focusDispList)
    self.lbgc3.bind("<1>", self.focusDispList)

    self.lbgc3.bind("<KeyRelease>", self.focusColorsPart)
    self.lbgc3.bind("<Shift-ButtonRelease>", self.focusColorsPart)
    self.lbgc3.bind("<Control-ButtonRelease>", self.focusColorsPart)

    self.curcom = T.Variable()
    self.curcom.set(self.collection_comment_placeholder)
    self.colcom = T.Entry(fdc0, textvariable=self.curcom, bd=1)
    self.colcom.grid(row=5, column=3, rowspan=1, columnspan=3, sticky="we")

    self.colcom.bind("<1>", self.emptyDispComm)

    itlF = ("courier", 9, "normal italic")
    couF = ("courier", 9, "normal")

    fdb1 = T.Frame(fdc0)
    fdb1.grid(row=1, column=5, sticky="nwe")
    fdb2 = T.Frame(fdc0)
    fdb2.grid(row=2, column=5, sticky="nwe")
    fdb3 = T.Frame(fdc0)
    fdb3.grid(row=3, column=5, sticky="swe")

    fdc0.columnconfigure(5, uniform=1)

    self.clnew = T.Button(fdb1, text="New", bd=1, command=self.dispCollsNew)
    self.cldel = T.Button(fdb1, text="Delete", bd=1, command=self.dispCollsDelete)
    self.clins = T.Button(fdb2, text="Insert", bd=1, command=self.dispCollsInsert)
    self.clrem = T.Button(fdb2, text="Remove", bd=1, command=self.dispCollsRemove)
    self.hllab = T.Label(fdb2, text="\nhighlight", font=couF, width=8)
    self.hlval = T.Entry(fdb2, width=8, textvariable=self.dispHighName, bd=0)
    self.hlval.configure(background="#d9d9d9", disabledbackground="#d9d9d9", readonlybackground="#d9d9d9")
    self.hlval.configure(state="readonly")
    self.clset = T.Button(fdb3, text="Set", bd=1, command=self.dispCollsSet)

    self.clnew.pack(side="top", fill="both", expand="yes")
    self.cldel.pack(side="top", fill="both", expand="yes")
    self.clins.pack(side="top", fill="both")
    self.clrem.pack(side="top", fill="both")
    self.hllab.pack(side="top", fill="both")
    self.hlval.pack(side="top", fill="both")
    T.Label(fdb2, text=" ", width=8).pack(side="top", fill="both")
    self.clset.pack(side="top", fill="both")
    T.Label(fdc0, text="\ndiameter\ncolor\ncomment", font=itlF).grid(row=4, column=5, sticky="n")

    self.rosaaut_proj = T.IntVar()
    self.rosaaut_proj.set(1)
    self.rosaaut_linadj = T.IntVar()
    self.rosaaut_linadj.set(1)
    self.rosaaut_hl = T.IntVar()
    self.rosaaut_hl.set(1)

    self.rbshow = T.Button(fe, text="Show", command=self.dispRosaShow)
    self.rbshow.grid(row=0, column=1, sticky="we")
    T.Label(fe, text=" ").grid(row=1, column=0)
    T.Label(fe, text=" ", width=5).grid(row=0, column=0)
    T.Label(fe, text=" ", width=5).grid(row=0, column=2)
    self.lbaut = T.Label(fe, text="automatically plot\nafter:", bd=1, justify="left")
    self.fraut = T.Frame(fe, bd=1)
    self.cbaut_proj = T.Checkbutton(self.fraut, text="new projection", variable=self.rosaaut_proj, bd=1)
    self.cbaut_linadj = T.Checkbutton(self.fraut, text="linear/adjusted swap", variable=self.rosaaut_linadj, bd=1)
    self.cbaut_hl = T.Checkbutton(self.fraut, text="highlight (un)set", variable=self.rosaaut_hl, bd=1)
    self.lbaut.grid(row=3, column=0, sticky="we", columnspan=3)
    self.fraut.grid(row=4, column=0, sticky="we", columnspan=3)
    self.cbaut_proj.grid(row=4, column=0, sticky="w", columnspan=3)
    self.cbaut_linadj.grid(row=5, column=0, sticky="w", columnspan=3)
    self.cbaut_hl.grid(row=6, column=0, sticky="w", columnspan=3)
    T.Label(fe, text=" ", height=2).grid(row=7, column=1)
    T.Label(fe, text=" ", height=2, width=5).grid(row=12, column=1)
    self.statrosa = T.Label(fe, text="none set\n ", height=2, width=20, font=itlF)
    self.statrosa.grid(row=13, column=0, columnspan=3, sticky="we")
    T.Label(fe, text=" ", height=2, width=5).grid(row=14, column=1)

    fe.columnconfigure(1, uniform=1)

    self.rbshow.configure(state="disabled")

  def dispSetVColDia(self, value):
    intvalue = (10 * int(value[:-2])) + int(value[-1:])
    self.vcoldia.set(intvalue)

  def viewDisp(self):
    if self.works:
      return
    if not 0 == self.rterror:
      return

    if self.fcurr == self.f2disp:
      return

    self.fcurr.pack_forget()
    self.fcurr = self.f2disp
    self.fcurr.pack(side="top", anchor="n", expand=1, fill="both")

    self.f1file.configure(relief="raised")
    self.f1table.configure(relief="raised")
    self.f1clust.configure(relief="raised")
    self.f1state.configure(relief="raised")
    self.f1disp.configure(relief="sunken")
    self.f1help.configure(relief="raised")

    todisable = True
    if not [] == self.alg:
      if 0 == self.alg.error:
        todisable = False

    if todisable:
      self.rbshow.configure(state="disabled")
    else:
      self.rbshow.configure(state="normal")

    self.statrosa.configure(text=" \n ")

  def focusDispList(self, event):
    event.widget.focus()

  def focusCasesPart(self, index):
    tohigh = True
    hname = ""
    htype = ""
    if -1 == index:
      self.dispHigh = -1
      tohigh = False
    if self.dispHigh == index:
      self.dispHigh = -1
      tohigh = False
    if tohigh:
      self.dispHigh = index
      hname = self.dispCases[index]

    self.hlval.configure(state="normal")
    self.dispHighName.set(hname)
    self.hlval.configure(state="readonly")
    self.dispRosaAutoHl()

  def focusCasesPart1(self, event):
    if 0 == len(self.dispCaseRanks):
      return

    index = int(self.lbdc1.index("active"))
    rank = self.dispCaseRanks[index]
    self.focusCasesPart(rank)

  def focusCasesPart2(self, event):
    if 0 == len(self.dispColls):
      return

    index = int(self.lbgc2.index("active")) - 2
    if 0 > index:
      return

    curcoll = int(self.lbgc1.index("active"))
    collname = self.lbgc2.get(0)
    if not collname == self.dispColls[curcoll]:
      return

    curlist = self.dispCollCases[curcoll]
    if 0 > len(curlist):
      return

    rank = curlist[index]
    self.focusCasesPart(rank)

  def emptyDispComm(self, event):
    if self.curcom.get() == self.collection_comment_placeholder:
      self.curcom.set("")

  def distCaseColls(self, probnames, probcolls):
    realnames = self.sto.dcnames

    reallen = len(realnames)
    problen = len(probnames)

    if 0 == reallen:
      return
    if 0 == problen:
      return
    if not len(self.dispCaseRanks) == reallen:
      return
    if not len(probcolls) == problen:
      return

    dcmax = -1
    realcolls = reallen * [0]
    for probind in range(problen):
      dcprob = probnames[probind]
      curstart = dcmax + 1
      if curstart == reallen:
        break
      for realind in range(curstart, reallen):
        dcreal = realnames[realind]
        if dcreal == dcprob:
          dcmax = realind
          insind = probcolls[probind]
          realcolls[realind] = insind

    for ind in range((reallen-1), -1, -1):
      colind = realcolls[ind] - 1
      if -1 < colind:
        self.dispCaseRanks.pop(ind)
        self.dispCollCases[colind].append(ind)

    collen = len(self.dispCollCases)
    for ind in range(collen):
      self.dispCollCases[ind].sort()

    return

  def setDispCases(self):
    if self.sto.dcnames == self.dispCases:
      return

    self.dispHigh = -1
    self.dispHighName.set("")

    probnames = self.dispCases
    probcolls = len(probnames) * [0]
    todiff = False
    for collind in range(len(self.dispCollCases)):
      for caseind in self.dispCollCases[collind]:
        probcolls[caseind] = collind + 1
        todiff = True
      self.dispCollCases[collind] = []

    self.dispCases = []
    for item in self.sto.dcnames:
      self.dispCases.append(item)
    self.dispCaseRanks = list(range(len(self.dispCases)))

    if todiff:
      self.distCaseColls(probnames, probcolls)

    self.setDispCasesData()

    collen = len(self.dispColls)
    if 0 < collen:
      curind = self.lbgc1.index("active")
      self.setDispCollsData(curind)
    else:
      self.dispDefaultSet()


  def setDispCasesData(self):
    curlist = []
    for ind in self.dispCaseRanks:
      item = self.dispCases[ind]
      curlist.append(item)
    self.vdc1info.set(tuple(curlist))


  def setDispCollsData(self, index):
    curgrp = self.dispColls[index]
    curdia = self.dispCollDiameters[index]
    diavalue = str(curdia // 10) + "." + str(curdia % 10)
    colorind = self.dispCollColors[index]
    curcolor = self.rosacolors[colorind]
    curlist = [curgrp, diavalue + " / " + curcolor]
    for ind in self.dispCollCases[index]:
      item = self.dispCases[ind]
      curlist.append(item)
    self.vgc2info.set(tuple(curlist))

    self.vcoldia.set(curdia)
    self.fcoldia.set(diavalue)
    self.lbgc3.activate(colorind)
    self.lbgc3.select_clear(0, "end")
    self.lbgc3.see(colorind)

    curcomment = self.dispCollComments[index]
    self.curcom.set(curcomment)

  def focusCollsPart(self, event):
    if 0 < len(self.dispColls):
      sellist = self.lbgc1.curselection()
      if 0 < len(sellist):
        curind = int(sellist[0])
        self.setDispCollsData(curind)

  def focusColorsPart(self, event):
    curind = self.lbgc3.index("active")
    self.lbgc3.select_clear(0, "end")
    self.lbgc3.select_set(curind)

  def dispCollsNew(self):
    rank = 1
    if 0 < len(self.dispColls):
      laststr = self.dispColls[-1]
      lastind = int(laststr.split()[-1])
      rank = lastind + 1

    self.dispColls.append("group " + str(rank))
    self.vgc1info.set(tuple(self.dispColls))
    listind = len(self.dispColls) - 1
    self.lbgc1.activate(listind)
    self.lbgc1.select_clear(0, "end")
    self.lbgc1.select_set(listind)
    self.lbgc1.see(listind)

    colorind = self.lbgc3.index("active")

    self.dispCollCases.append([])
    self.dispCollDiameters.append(self.vcoldia.get())
    self.dispCollColors.append(colorind)
    self.dispCollComments.append(self.curcom.get())

    self.setDispCollsData(listind)
    self.fbutt42.configure(state="normal")
    self.lbgc2.focus()

  def dispDefaultSet(self):
    self.vgc2info.set(())
    self.vcoldia.set(5)
    self.fcoldia.set("0.5")
    self.lbgc3.activate(0)
    self.lbgc3.select_clear(0, "end")
    self.lbgc3.see(0)
    self.curcom.set(self.collection_comment_placeholder)

  def dispCollsDelete(self):
    index = -1
    collen = len(self.dispColls)
    sellist = self.lbgc1.curselection()
    if 0 < len(sellist):
      index = int(sellist[0])
      if len(self.dispColls) > index:
        self.dispColls.pop(index)
        remlist = self.dispCollCases.pop(index)
        self.dispCollDiameters.pop(index)
        self.dispCollColors.pop(index)
        self.dispCollComments.pop(index)
        self.vgc1info.set(tuple(self.dispColls))
        for rank in remlist:
          self.dispCaseRanks.append(rank)
        self.dispCaseRanks.sort()
        self.setDispCasesData()
      collen = len(self.dispColls)
      if 0 < collen:
        if not index < collen:
          index -= 1
    if 0 < collen:
      if -1 < index:
        self.lbgc1.activate(index)
        self.lbgc1.select_clear(0, "end")
        self.lbgc1.select_set(index)
        self.setDispCollsData(index)
    if 0 == collen:
      self.dispDefaultSet()
      self.fbutt42.configure(state="disabled")

  def dispCollsInsert(self):
    collen = len(self.dispColls)
    if 0 < collen:
      curcoll = self.lbgc1.index("active")
      collname = self.lbgc2.get(0)
      if not collname == self.dispColls[curcoll]:
        return

      selarea = self.lbdc1.curselection()
      sellist = []
      for ind in selarea:
        sellist.append(int(ind))
      sellist.sort(reverse=True)
      if 0 < len(sellist):
        curlist = self.dispCollCases[curcoll]
        for ind in sellist:
          rank = self.dispCaseRanks.pop(int(ind))
          curlist.append(rank)
        curlist.sort()
        self.dispCollCases[curcoll] = curlist
        self.setDispCollsData(curcoll)
        self.lbdc1.select_clear(0, "end")
        self.setDispCasesData()

  def dispCollsRemove(self):
    collen = len(self.dispColls)
    if 0 < collen:
      curcoll = self.lbgc1.index("active")
      collname = self.lbgc2.get(0)
      if not collname == self.dispColls[curcoll]:
        return

      selarea = self.lbgc2.curselection()
      sellist = []
      for ind in selarea:
        if 1 < int(ind):
          sellist.append(int(ind) - 2)
      sellist.sort(reverse=True)
      if 0 < len(sellist):
        curlist = self.dispCollCases[curcoll]
        for ind in sellist:
          rank = curlist.pop(ind)
          self.dispCaseRanks.append(rank)
        self.dispCaseRanks.sort()
        self.dispCollCases[curcoll] = curlist
        self.lbgc2.select_clear(0, "end")
        self.setDispCollsData(curcoll)
        self.setDispCasesData()

  def dispCollsSet(self):
    collen = len(self.dispColls)
    if 0 < collen:
      index = self.lbgc1.index("active")
      collname = self.lbgc2.get(0)
      if not collname == self.dispColls[index]:
        return

      self.dispCollDiameters[index] = self.vcoldia.get()
      colorind = self.lbgc3.index("active")
      self.dispCollColors[index] = colorind
      self.dispCollComments[index] = self.curcom.get()

      self.setDispCollsData(index)

  def dispRosaShow(self):
    rinfo = "no action\n"
    if [] == self.alg:
      self.statrosa.configure(text=rinfo)
      return
    collen = len(self.dispColls)
    errcnt = 0
    sntcntlist = []

    if 0 == collen:
      self.statrosa.configure(text=rinfo)
      return

    datasource = []
    datatype = ""
    if 1 == self.vadjust.get():
      datasource = self.alg.getAdjusted()
      datatype = "adj"
    if datasource == []:
      datasource = self.alg.getProjections()
      datatype = "lin"
    if datasource == []:
      self.statrosa.configure(text=rinfo)
      return

    fig_nums = plt.get_fignums()
    if (not self.fig) or (not fig_nums):
      self.fig = plt.figure()
    else:
      self.fig.clf()

    plt.ion()
    ax = self.fig.add_subplot(111, projection="3d")
    ax.set_xlabel("1. pc")
    ax.set_ylabel("2. pc")
    ax.set_zlabel("3. pc")
    scatter_list_obj = []
    scatter_list_desc = []

    dispHighIndex = self.dispHigh
    sntcnt = 0
    for index in range(collen):
      scatter_list_x = []
      scatter_list_y = []
      scatter_list_z = []

      hlindex = -1
      indexlist = self.dispCollCases[index]
      if dispHighIndex in indexlist:
        hlindex = indexlist.index(self.dispHigh)

      collsize = len(indexlist)
      if 0 < collsize:
        sntcnt += 1
        try:
          curcoll = self.dispColls[index]
          rank = curcoll.split()[-1]
          diam = str(self.dispCollDiameters[index])
          colorind = self.dispCollColors[index]
          color = self.rosacolors[colorind]
          if color == "steel":
              color = "steelblue"
          comment = self.dispCollComments[index].strip()
          xhval = 0
          yhval = 0
          zhval = 0
          for caseind in range(collsize):
            caserank = indexlist[caseind]
            """
            # not displaying case names now
            caselabel = self.dispCases[caserank]
            caselablist = caselabel.split()
            caselabel = caselablist[0]
            for auxind in range(1, len(caselablist)):
              caselabel += "_" + str(caselablist[auxind])
            """
            xfval = datasource[caserank, 0]
            scatter_list_x.append(xfval)
            yfval = datasource[caserank, 1]
            scatter_list_y.append(yfval)
            zfval = datasource[caserank, 2]
            scatter_list_z.append(zfval)
            if caseind == hlindex:
              xhval = xfval
              yhval = yfval
              zhval = zfval
          scatter_list_obj.append(ax.scatter(scatter_list_x, scatter_list_y, scatter_list_z, s=float(diam)*10, c=color, marker="o"))
          curr_desc = str(rank) + " (" + str(datatype) + ")"
          if (comment != "") and (comment != self.collection_comment_placeholder):
              curr_desc += " " + comment
          scatter_list_desc.append(curr_desc)

          if hlindex > -1:
            ax.scatter([xhval], [yhval], [zhval], s=max(100,(float(diam)*14+60)), c="black", marker="x")
        except:
          errcnt += 1

    sntcntlist.append(sntcnt)
    if len(scatter_list_obj):
      plt.legend(scatter_list_obj, scatter_list_desc, loc="upper left", scatterpoints=1, ncol=2, fontsize=12, bbox_to_anchor=(-0.1, 1.1))

    plt.show()

    sntmsg = "none"
    if 0 < len(sntcntlist):
      sntmsg = str(sntcntlist[0])
      for icnt in sntcntlist[1:]:
        sntmsg += "+" + str(icnt)
    rinfo = sntmsg + " collections\n"
    if 0 == errcnt:
      rinfo += ""
    elif 1 == errcnt:
      rinfo += "1 error"
    else:
      rinfo += str(errcnt) + " errors"

    self.statrosa.configure(text=rinfo)

  def dispRosaAutoProj(self):
    tores = self.rosaaut_proj.get()
    if 0 == tores:
      return

    collen = len(self.dispColls)
    if 0 == collen:
      return

    self.dispRosaShow()

  def dispRosaAutoLinadj(self):
    tores = self.rosaaut_linadj.get()
    if 0 == tores:
      return

    if [] == self.alg:
      return
    if 1 == self.vadjust.get():
      if [] == self.alg.getAdjusted():
        return

    collen = len(self.dispColls)
    if 0 == collen:
      return

    self.dispRosaShow()

  def dispRosaAutoHl(self):
    tores = self.rosaaut_hl.get()
    if 0 == tores:
      return

    collen = len(self.dispColls)
    if 0 == collen:
      return

    self.dispRosaShow()

  def prepHelp(self):
    self.f2help = T.Frame(self.fbot)

    fc = T.Frame(self.f2help)
    fc.pack(side="top", expand=0, fill="none")
    T.Label(fc, text=" ", width=1, height=1).pack(side="top")

    fd = T.Frame(self.f2help)
    fd.pack(side="top", expand=1, fill="both")
    T.Label(self.f2help, text=" ", height=2).pack(side="top", expand=0, fill="none")

    finner = T.Frame(self.f2help)
    finner.pack(side="top", expand=0, fill="none")

    fe = T.Frame(finner)
    fe.grid(row=0, column=0)
    T.Label(finner, text=" ", height=3, width=10).grid(row=0, column=1)
    fdesc = T.Frame(finner)
    fdesc.grid(row=0, column=2)
    T.Label(self.f2help, text=" ", height=3).pack(side="top", expand=0, fill="none")

    fpars = T.Frame(self.f2help)
    fpars.pack(side="top", expand=0, fill="none")
    fbottom = T.Frame(self.f2help)
    fbottom.pack(side="top", expand=1, fill="both")
    T.Label(fbottom, text=" ").grid(row=0, column=0)

    itlF = ("courier", 9, "normal italic")

    labF = ("helvetica", 12, "bold")
    l1 = T.Label(fd, text="\nCluDisc", foreground="red4", font=labF)
    l1b = T.Text(fd, height=1, width=18, bd=0, font=("courier", 9))
    l1b.insert("end", "www.bioplexity.org\n")
    l1b.configure(state="disabled")
    linedesc = "two step MDS: cluster-center subscales, then discoidal projection"
    l2 = T.Label(fd, text="\nClustered Scaling - " + linedesc + "\n ")
    cdver = "CluDisc version: " + str(CluDiscVersion)
    if isclibcd:
      cdver += " (with speed-up)"
    else:
      cdver += " (without speed-up)"
    l3a = T.Label(fd, text=cdver, font=itlF)
    l3b = T.Label(fd, text="open source software", font=itlF)
    for i in (l1, l1b, l2, l3a, l3b):
      i.pack(side="top", expand=0, fill="none")

    T.Label(fdesc, text="1) load csv matrix or tree data", font=("ariel", 9)).grid(row=0, column=0, sticky="w")
    T.Label(fdesc, text="2) if loaded csv data:", font=("ariel", 9)).grid(row=1, column=0, sticky="w")
    T.Label(fdesc, text="     a) select data region and orientation", font=("ariel", 9)).grid(row=2, column=0, sticky="w")
    T.Label(fdesc, text="     b) choose data normalization options ", font=("ariel", 9)).grid(row=3, column=0, sticky="w")
    T.Label(fdesc, text="     c) set the cluster count and cluster it", font=("ariel", 9)).grid(row=4, column=0, sticky="w")
    T.Label(fdesc, text="3) prepare for projecting:", font=("ariel", 9)).grid(row=5, column=0, sticky="w")
    T.Label(fdesc, text="     a) select particular groups in the tree", font=("ariel", 9)).grid(row=6, column=0, sticky="w")
    T.Label(fdesc, text="     b) choose scaled vs. uninorm centers ", font=("ariel", 9)).grid(row=7, column=0, sticky="w")
    T.Label(fdesc, text="     c) scale, possibly in a non-linear mode", font=("ariel", 9)).grid(row=8, column=0, sticky="w")
    T.Label(fdesc, text="4) display the 3D projections", font=("ariel", 9)).grid(row=9, column=0, sticky="w")
    T.Label(fdesc, text="     *) save the tree, projection data", font=("ariel", 9)).grid(row=10, column=0, sticky="w")

    ld11 = T.Label(fe, text="File: ")
    ld11t = T.Label(fe, text="data files load, save")
    ld12 = T.Label(fe, text="Table: ")
    ld12t = T.Label(fe, text="input data selection")
    ld13 = T.Label(fe, text="Clusters: ")
    ld13t = T.Label(fe, text="axial centers choice ")
    ld14 = T.Label(fe, text="State: ")
    ld14t = T.Label(fe, text="result content view ")
    ld15 = T.Label(fe, text="Display: ")
    ld15t = T.Label(fe, text="result projections ")
    lin = T.Label(fe, text=" ")
    lint = T.Label(fe, text=" ")
    helpLabs = (ld11, ld12, ld13, ld14, ld15, lin)
    helpTexts = (ld11t, ld12t, ld13t, ld14t, ld15t, lint)
    for i in range(len(helpLabs)):
      helpLabs[i].grid(row=i, column=0, sticky="e")
      helpTexts[i].grid(row=i, column=1, sticky="w")

    rtlist = []

    rtwins = ""
    if "l" == self.params["winstype"]:
      rtwins = "linear / "
    else:
      rtwins = "quadratic / "
    rtwins += str(self.params["winsvalue"])
    rtlist.append(" winsorize:   " + rtwins)

    rtcycs = str(self.params["kmcycmax"])
    rtlist.append("clustering:   max " + rtcycs + " cycles")

    rtvert = "proportional"
    if 1 == self.params["vertical"]:
      rtvert = "square root"
    if 2 == self.params["vertical"]:
      rtvert = "circular"
    rtlist.append("verticality:   " + rtvert)

    rtproj = "exact"
    if 0 < self.params["project"]:
      rtproj = "fail-safe / " + str(self.params["project"])
    rtlist.append("projection:   " + rtproj)

    rtnorm = "none"
    if 0 < self.params["resnorm"]:
      rtnorm = "mean scale "
      rtnorm += str(self.params["resnorm"])
    rtlist.append(" uniscaled:   " + rtnorm)

    rtadjs = str(self.params["atenstart"]) + " / "
    rtadjs += str(self.params["atenstay"]) + " / "
    rtadjs += str(self.params["atencount"])
    rtlist.append(" adjusting:   " + rtadjs)

    rtextend = ""
    if 1 == self.params["extend"]:
      rtextend = "linear"
    if 2 == self.params["extend"]:
      rtextend = "square"
    rtlist.append(" elasticity:   " + rtextend)

    self.varglist = T.Variable()
    self.varglist.set(tuple(rtlist))

    T.Label(fpars, text="Runtime parameters", font=itlF).grid(row=0, column=0)
    self.larglist = T.Listbox(fpars, height=4, width=25, bd=1, listvariable=self.varglist, selectmode="extended")
    self.larglist.grid(row=1, column=0)

    self.larglist.bind("<1>", self.focusArgList)

  def focusArgList(self, event):
    self.larglist.focus()

  def viewHelp(self):
    if self.works:
      return

    self.larglist.select_clear(0, "end")
    if self.fcurr == self.f2help:
      return

    self.fcurr.pack_forget()
    self.fcurr = self.f2help
    self.fcurr.pack(side="top", anchor="n", expand=1, fill="y")

    self.f1file.configure(relief="raised")
    self.f1table.configure(relief="raised")
    self.f1clust.configure(relief="raised")
    self.f1state.configure(relief="raised")
    self.f1disp.configure(relief="raised")
    self.f1help.configure(relief="sunken")

  def prepQuit(self):
    self.f2quit = T.Frame(self.fbot)

    fd = T.Frame(self.f2quit)
    fd.pack(side="top", expand=1, fill="both")
    fe = T.Frame(self.f2quit)
    fe.pack(side="top", expand=0, fill="none")
    ff = T.Frame(self.f2quit)
    ff.pack(side="top", expand=0, fill="none")

    T.Label(fd, text=" ", height=10).pack(side="top")

    ql0 = T.Frame(fd)
    ql0.pack(side="top")
    quitF = ("courier" ,12, "bold")
    self.ql1 = T.Label(ql0, text="  Really to quit the CluDisc?  ", height=2, foreground="red4", font=quitF)
    self.ql2 = T.Label(ql0, text=" ", height=2)
    self.ql1.grid(row=0, column=0, sticky="snew")
    self.ql2.grid(row=1, column=0, sticky="snew")

    self.quitb1 = T.Button(fe, text="  No  ", command=self.viewHelp, bd=1)
    self.quitb2 = T.Button(fe, text=" Yes ", command=self.quitApp, bd=1)
    self.quitb1.pack(side="left")
    self.quitb2.pack(side="left")

    T.Label(ff, text=" ", height=10).pack(side="top")

  def viewQuit(self):
    if self.works:
      return

    if self.fcurr == self.f2quit:
      return

    if not 0 == self.rterror:
      self.ql1.configure(text="Run-time error has occured")
      self.ql2.configure(text="probably insufficient amount of memory")
    else:
      self.ql2.configure(text=" ")
      if not self.disposed:
        self.ql2.configure(text=" Current cluster-tree neither saved nor disposed.")

    self.fcurr.pack_forget()
    self.fcurr = self.f2quit
    self.fcurr.pack(side="top", anchor="n", expand=1, fill="both")
    if 0 == self.rterror:
      self.quitb1.focus()
    else:
      self.quitb2.focus()

    self.f1file.configure(relief="raised")
    self.f1table.configure(relief="raised")
    self.f1clust.configure(relief="raised")
    self.f1state.configure(relief="raised")
    self.f1disp.configure(relief="raised")
    self.f1help.configure(relief="raised")

  def quitApp(self):
    if self.fig:
      try:
        self.fig.close()
        self.fig = None
      except:
        pass
    self.root.quit()

  def setLabel(self, label):
    if None == label:
      self.top.title("CluDisc - clustered scaling")
    else:
      self.top.title("CluDisc - " + str(label))

  def prepWins(self):
    self.top = T.Toplevel(self.root)
    self.top.minsize(900, 600)
    try:
      self.top.tk.call("wm", "iconphoto", self.top._w, "-default", self.app_icon)
    except:
      pass

    self.top.title("CluDisc - clustered scaling")
    self.top.protocol("WM_DELETE_WINDOW", self.viewQuit)

    self.ftop = T.Frame(self.top)
    self.fbot = T.Frame(self.top)
    self.ftop.pack(side="top", anchor="n", expand=0, fill="x")
    self.fbot.pack(side="top", anchor="n", expand=1, fill="both")

    self.f1file = T.Button(self.ftop, text=" File ", command=self.viewFile, relief="raised", bd=1, width=10)
    self.f1table = T.Button(self.ftop, text=" Table ", command=self.viewTable, relief="raised", bd=1, width=10)
    self.f1clust = T.Button(self.ftop, text="Clusters", command=self.viewClust, relief="raised", bd=1, width=10)
    self.f1state = T.Button(self.ftop, text=" State ", command=self.viewState, relief="raised", bd=1, width=10)
    self.f1disp = T.Button(self.ftop, text="Display", command=self.viewDisp, relief="raised", bd=1, width=10)

    f1list = [self.f1file, self.f1table, self.f1clust, self.f1state, self.f1disp]
    for f1l in f1list:
      f1l.pack(side="left")

    self.f1help = T.Button(self.ftop, text=" About ", command=self.viewHelp, relief="raised", bd=1, width=10)
    self.f1help.pack(side="right")

    self.prepFile()
    self.prepTable()
    self.prepClust()
    self.prepState()
    self.prepDisp()
    self.prepHelp()
    self.prepQuit()

    self.fcurr = self.f2quit
    self.viewHelp()

  def runApp(self, initreefile):
    if not "" == initreefile:
      self.datafile = initreefile
      self.saved = False
      self.sname = ""

      self.filestate = False
      try:
        self.loadClustersFile()
      except:
        return

      if not self.filestate:
        self.sto.dcnames = []
        self.sto.prnames = []
        self.stateGroups()
        self.stateAxes()
        self.finfo.configure(text=" load failed ")
        self.updateFileCollsView()

      if self.filestate:
        self.saved = True
        self.sname = os.path.basename(self.datafile)
        self.stateGroups()
        self.stateAxes()
        self.setDispCases()
        self.stateAxialSels(0)

        self.bs1.configure(state="normal")
        self.bs2.configure(state="normal")

        self.fbutt22.configure(state="normal")
        self.fbutt23.configure(state="normal")
        self.fbutt24.configure(state="normal")
        self.finfo.configure(text=" data loaded ")

        flabel = ""
        if "" == self.sto.oname:
          flabel = self.sto.fname + " / " + self.sname
        else:
          flabel = self.sto.oname + " / " + self.sname
        self.setLabel(flabel)

        self.viewClust()

    self.root.update()
    self.root.mainloop()

  def catchExcept(self):
    self.works = False
    self.rterror = 1
    self.disableFileLoad()
    self.viewQuit()

if ('__main__' == __name__):
  try:
    cls = TreeSelect(paramvals)
    sto = StorageData()
    cen = ClusterCenters(paramvals, isclibcd)

    app = CluDiscApp(sto, cls, cen, paramvals, isclibcd)
    app.prepWins()
  except:
    log_err("an unknown error during initialization, please report it")
    exit(1)

  try:
    app.runApp(paramvals["initreefile"])
  except:
    log_err("connection to the window system has broken")
    exit(1)

